#!/usr/bin/env perl

use strict;
use warnings;

use XML::Bare;
use Data::Dumper;
use DateTime;
use Scalar::Util qw(looks_like_number);

my @DELETE_MORE = qw(Dataspace DataType);
my @MILLIS_TO_TIME = (3600000, 60000, 1000);
my $KNOWN_OBJECT_ATTRS = qr/^(?:class)/i;
my $aquarius_file = $ARGV[0] || print_usage();
my $doi_prefix="10.5067/Aquarius/";
my $doi_ver='';
my $versionid='';
my %SNAME =(
	L2 =>      'AQ2_SM',
	L3m_DAY => 'AQ3_DYSM',
	L3m_7D  => 'AQ3_WKSM',
	L3m_MO  => 'AQ3_MOSM',
	L3m_SN  => 'AQ3_SNSM',
	L3m_YR  => 'AQ3_ANSM',
        L3m_SC  => 'AQ3_SCSM',
        L3m_MC  => 'AQ3_MCSM',
        L3m_CU  => 'AQ3_CUSM'
); 

unless (-e $aquarius_file ){ print_usage()}

make_ecs_file($aquarius_file);

exit(0);

sub print_usage {
print STDERR << "--";
Usage: $0 FILE

This script generates an ECS metadata file for a given Aquarius L2 or L3 FILE
The output is <FILE>.met

--
exit(0);
}

sub make_ecs_file {
	my $file = shift;
	my $text = `h5dump -Au $file`;
	my $xml = XML::Bare->new(text => $text, )->parse;
	
	my @nodes = clean_xml($xml->{'HDF5-File'}{'RootGroup'}{'Attribute'});
	@nodes = @{$nodes[0]};
	my %attributes;
	
	foreach (@nodes){
		$attributes{$_->{Name}} = $_->{Data};
	}
	
	my @input_pointer;
	foreach ('Ancillary Files', 'Input Files'){
		if ($attributes{$_}){
			push(@input_pointer, split(',', $attributes{$_}));
		}
	}

	foreach (qw(Start End)){
		if (!$attributes{"$_ Time"}){
			my $milliseconds = $attributes{"$_ Millisec"};
			my ($part, @parts);
			foreach (@MILLIS_TO_TIME){
                $part = int($milliseconds/$_);
                $milliseconds -= $part*$_;
                push(@parts, $part);
			}
			$parts[-1] += $milliseconds/1000;
            $attributes{"$_ Time"} = sprintf('%04i%03i%02i%02i%02i',@attributes{"$_ Year", "$_ Day"},@parts);
		}
	}
	
	foreach (qw(Easternmost Westernmost)){
		if ($attributes{"$_ Latitude"}){
			$attributes{"$_ Longitude"} = $attributes{"$_ Latitude"};
		}
	}
	
	my $doi = '';
        my $sname ='';
        ($versionid = $attributes{'Processing Version'}) =~ s/[V\.0]//g; 
        $doi_ver="00"."$versionid";
	while (my ($k, $v) = each(%SNAME)){
		if ($file =~ /$k/){
                        $sname = $v;
			$doi = "${doi_prefix}${sname}.${doi_ver}";
			last;
		}
	}
#	print Dumper(\%attributes);
	
	# Arrays are groups, the first element is the group name
	# Scalars in groups are mere attributes, printed as-is
	# Hashes are objects, object value rules:
	#   Scalar reference values are printed as-is
	#   Scalars are printed with quotes if !looks_like_number
	#   Any arrayref containing any combination of the above is acceptable
	my @FORMAT = (
	    [   'InventoryMetaData',
	        'GroupType = MasterGroup',
	        [   'ECSDataGranule',
	            {'LocalGranuleID' => $attributes{'Product Name'}},
	            {'ProductionDateTime' => get_datetime_from_julian($attributes{'Processing Time'})},
                {'LocalVersionID' => $attributes{'Processing Version'} || ''},
	        ],
	        ($attributes{'Orbit Number'} ? 
	            [  'OrbitCalculatedSpatialDomain ',
	               ['OBJECT', 'OrbitCalculatedSpatialDomainContainer',
	                   'CLASS = 1',
	                   {OrbitNumber => $attributes{'Orbit Number'}, CLASS => 1,},
	                   {EquatorCrossingLongitude => $attributes{'Orbit Node Longitude'}, CLASS => 1,},
	                   {EquatorCrossingTime => get_time_from_julian($attributes{'Node Crossing Time'}), CLASS => 1,},
	                   {EquatorCrossingDate => get_date_from_julian($attributes{'Node Crossing Time'}), CLASS => 1,},
	               ]
	            ] : ()),
	        [   'CollectionDescriptionClass', 
	            {'ShortName' => "$sname"},
	            {'VersionID' => "$versionid"},
	        ],
	        (@input_pointer ? 
		        [   'InputGranule',
		            {'InputPointer' => [@input_pointer]},
		        ] : ()),
	        [   'SpatialDomainContainer',
	        	(defined($attributes{'Northernmost Latitude'}) ? 
		        	((join(',',@attributes{'Northernmost Latitude', 'Southernmost Latitude', 'Easternmost Longitude', 'Westernmost Longitude'}) eq '90,-90,180,-180' ?
			           ['GranuleLocality',
			               {LocalityValue => "Global"}, 
			           ] : ()),
	                [ 'HorizontalSpatialDomainContainer',
		                ['BoundingRectangle',
	                        {'NorthBoundingCoordinate' => $attributes{'Northernmost Latitude'},},
	                        {'SouthBoundingCoordinate' => $attributes{'Southernmost Latitude'},},
	                        {'EastBoundingCoordinate' => $attributes{'Easternmost Longitude'},},
	                        {'WestBoundingCoordinate' => $attributes{'Westernmost Longitude'},},
		                ],
	                ]) : ()),
	        ],
	        [   'RangeDateTime',
	            {'RangeBeginningDate' => get_date_from_julian($attributes{'Start Time'})},
	            {'RangeBeginningTime' => get_time_from_julian($attributes{'Start Time'})},
	            {'RangeEndingDate' => get_date_from_julian($attributes{'End Time'})},
	            {'RangeEndingTime' => get_time_from_julian($attributes{'End Time'})},
	        ],
	        [   'PGEVersionClass',
	            {'PGEVersion' => \'"07"'},
	        ],
	        [  'AssociatedPlatformInstrumentSensor',
	           ['OBJECT', 'AssociatedPlatformInstrumentSensorContainer', 'CLASS = 1',
           	       {AssociatedPlatformShortName => "SAC-D", CLASS => 1,},
           	       {AssociatedInstrumentShortName => "Aquarius_Radiometer", CLASS => 1,},
           	       {AssociatedSensorShortName => "Aquarius_Radiometer", CLASS => 1,},
	           ]
	        ],
	        ['AdditionalAttributes',
	        	['OBJECT', 'AdditionalAttributesContainer', 'CLASS = 1',
	        		{AdditionalAttributeName => "AscendingDescendingFlag", CLASS => 1,},
	        		['InformationContent', 'CLASS = 1',
	        			{ParameterValue => 'Both', CLASS => 1,},
	        		],
	        	],
	        	['OBJECT', 'AdditionalAttributesContainer', 'CLASS = 2',
	        		{AdditionalAttributeName => "identifier_product_doi", CLASS => 2,},
	        		['InformationContent', 'CLASS = 1',
	        			{ParameterValue => $doi, CLASS => 2,},
	        		],
	        	],
	        	['OBJECT', 'AdditionalAttributesContainer', 'CLASS = 3',
	        		{AdditionalAttributeName => "identifier_product_doi_authority", CLASS => 3,},
	        		['InformationContent', 'CLASS = 3',
	        			{ParameterValue => 'http://dx.doi.org', CLASS => 3,},
	        		],
	        	],
	        ],
	    ]
	);
#	print format_to_dsl(0, @FORMAT);
	open(my $output, '>', "$file.met");
	print {$output} format_to_dsl(0, @FORMAT);
	print {$output} "END";
	close($output);
}


sub format_to_dsl {
	my ($indents, @nodes) = @_;
	my $indent = '   'x$indents;
	my $return = '';
	foreach my $node (@nodes){
		my $ref = ref($node);
		if (!$ref){
			$return .= "$indent$node\n"; 
		} elsif ($ref eq 'ARRAY'){
			my $name = shift(@$node);
			my $type = 'GROUP';
			if ($name =~ /^group|object$/i){
				$type = uc($name);
				$name = shift(@$node);
			}
			my $inner_block = format_to_dsl($indents+1, @$node);
			if ($inner_block){
				$return .= "${indent}$type = $name\n";
				$return .= $inner_block;
				$return .= "${indent}END_$type = $name\n";
			}
		} elsif ($ref eq 'HASH'){
			my $object_name;
			
			my $object_block = '';
			
			while (my ($name, $v) = each(%$node)){
				next unless (defined($v) && $v ne '');
				if ($name =~ $KNOWN_OBJECT_ATTRS){
					my $this_line = "$indent   $name = ";
					if (ref($v) eq 'SCALAR'){
						$this_line .= $$v;
					} elsif (looks_like_number($v)){
                        $this_line .= $v;
                    } elsif (defined($v)) {
                        $this_line .= "\"$v\"";
                    } else {
#                        print "No value for $name\n";
                        $this_line .= '""';
                    }
                    $object_block = "$this_line\n$object_block";
				} else {
					$object_name = $name;
		            if (!ref($v) || ref($v) eq 'SCALAR'){
	                    $object_block .= "$indent   NUM_VAL = 1\n";
	                    $object_block .= "$indent   VALUE = ";
	                    if (ref($v) eq 'SCALAR'){
	                        $object_block .= $$v;
	                    } else {
	                        if (looks_like_number($v)){
	                            $object_block .= $v;
	                        } elsif (defined($v)) {
	                            $object_block .= "\"$v\"";
	                        } else {
#	                        	print "No value for $name\n";
	                            $object_block .= '""';
	                        }
	                    }
	                    $object_block .= "\n";
		            } elsif (ref($v) eq 'ARRAY'){
		            	my @values = @$v;
	                	$object_block .= "$indent   NUM_VAL = " . scalar(@values) . "\n";
		                if (@values){
		                    $object_block .= "$indent   VALUE = (";
		                    my $i = '';
			                while (my $v = shift(@values)){
		                        if ($i){
		                            $object_block .= $i;
		                        } else {
		                            $i = "$indent            ";
		                        }
		                        if (ref($v) eq 'SCALAR'){
		                        	$object_block .= $$v;
		                        } elsif (!ref($v)){
		                        	if (looks_like_number($v)){
		                        		$object_block .= $v;
		                        	} else {
		                        		$object_block .= "\"$v\"";
		                        	}
	                            } else {
		                        	die("Invalid value detected: $v");
		                        }
		                        
		                        $object_block .= (@values ? ',' : ')') . "\n";
			                }
		                } else {
		                    $object_block .= "$indent   VALUE = ()\n";
		                }
		            } elsif (ref($v) eq 'HASH') {
			            $object_block .= format_to_dsl($indents, $v);
		            } else {
                        die("Invalid value detected: $v");
                    }
				}
			}
			if ($object_name){
	            $return .= "${indent}OBJECT = $object_name\n";
	            $return .= $object_block;
	            $return .= "${indent}END_OBJECT = $object_name\n";
			}
		}
	}
	return $return;
}

sub get_datetime_from_julian {
    my $datetime = shift || return '';
    if ($datetime && $datetime =~ /^(?<year>\d{4})(?<day_of_year>\d{3})(?<hour>\d{2})(?<minute>\d{2})(?<second>\d{2})/){
        return DateTime->from_day_of_year(%+)->strftime('%FT%TZ');
    } else {
        die("Invalid datetime string");
    }
}

sub get_date_from_julian {
    my $datetime = shift;
    if ($datetime && $datetime =~ /^(?<year>\d{4})(?<day_of_year>\d{3})(?<hour>\d{2})(?<minute>\d{2})(?<second>\d{2})/){
        return DateTime->from_day_of_year(%+)->strftime('%F');
    }
}

sub get_time_from_julian {
    my $datetime = shift || return '';
    if ($datetime && $datetime && $datetime =~ /^(?<year>\d{4})(?<day_of_year>\d{3})(?<hour>\d{2})(?<minute>\d{2})(?<second>\d{2})/){
        return DateTime->from_day_of_year(%+)->strftime('%TZ');
    }
}


sub clean_xml {
    my @ret;
    foreach (@_){
       my $i = delete_infos($_);
       if (defined($i)){
        push(@ret, $i);
       }
    }
    return @ret;
}


sub delete_infos {
    my $item = shift;
    unless (defined($item)){
        return;
    }
    
    if (ref($item) eq 'HASH'){
        my %item = %$item;
        delete(@item{grep(m/^_/,keys(%$item)), qw(style class rel colspan rowspan width border cellspacing cellpadding href), @DELETE_MORE});
        while (my ($k, $v) = each(%item)){
            $item{$k} = delete_infos($v);
        }
        delete_infos($_) foreach (values(%item));
            if (exists($item{value}) && defined($item{value}) && !ref($item{value})){
                if ($item{value} =~ /^\s*$/){
                    delete($item{value});
                } else {
                    $item{value} =~ s/^\s+|\s+$//g;
                    $item{value} =~ s/^"|"$//g;
                    $item{value} =~ s/^\s+|\s+$//g;
                }
            }
            if (keys(%item) == 1){
                ($item) = values(%item);
            } elsif (keys(%item)){
                $item = \%item;
            } else {
                undef($item);
            }
    } elsif (ref($item) eq 'ARRAY'){
        my @new;
        foreach (@$item) {
            my $i = delete_infos($_);
            push(@new, $i) if ($i);
        }
        if (@new){
          $item = \@new;
        } else {
            undef($item);
        }
    }
    return $item;
}
