#include <stdint.h>
#include "hdf5utils.h"
#include "hdf5.h"
#include "hdf5_hl.h"

using namespace std;

enum { ROOT, HIRESMOM, HOUSEKEEP, METADATA, MOMDATA, SCDATA };

const uint32_t MAXSCANS = 800;
const float CELOFF = 273.15;

// one pulse repetition interval (PRI) ~300 μs
// Number of PRIs per footprint = 32

const uint32_t nREF_PKTs_fp = 2; // Number of reference packets per footprint
const uint32_t nANT_PKTs_fp = 8; // Number of antenna packets per footprint
const uint32_t last_PRIs = 44;
const uint32_t last_PKTs = 11;
const double pi = 3.1415926535897932;

struct fullBandCoeffs_struct {
  double C1;
  double C2;
  double C3[2];
  double C4[2];
  double C5[2];
  double L1[2];
  double L2;
  double L3[2];
  double L4[2];
  double L5[2];
  double T1[2];
  double T2;
  double T3[2];
  double T4[2];
  double T5[2];
  double TDL0[2];
  double TDLoffset[2];
  double Tcoupler0[2];
  double Tdiplexer0[2];
  double Tnd[2];
  double Tnd0[2];
  double Tomt0[2];
  double Txnd[2];
  double Txnd0[2];
  double c1_DLoffset[2];
  double c1_nd[2];
  double c1_xnd[2];
  double c2_DLoffset[2];
  double c2_nd[2];
  double c2_xnd[2];
  double c3_DLoffset[2];
  double c3_nd[2];
  double c3_xnd[2];
  double c4_DLoffset[2];
  double c4_nd[2];
  double c4_xnd[2];
  double channel_phase;
  double dgamma[2];
  double dpsi;
  double lambda[2];
  double nd_phase;
  double rx_phase;
  double tcor_im[2];
  double tcor_re[2];
  double xnd_phase;
};

struct C2_struct {
  double c2_0[2];
  double c2_1[2];
  double c2_2[2];
};

struct C3_struct {
  double c3_0[2];
  double c3_1[2];
  double c3_2[2];
};

struct linearCoeffs_struct {
  C2_struct C2;
  C2_struct C2_16;
  C3_struct C3;
  C3_struct C3_16;

  double Tref[2];
  double Tref16[16][2];
  double h[5295];
  double time_gap;
};
      
struct subBandCoeffs_struct {
  double C1_16[16];
  double C2_16[16];
  double C3_16[16][2];
  double C4_16[16][2];
  double C5_16[16][2];
  double L1_16[16][2];
  double L2_16[16];
  double L3_16[16][2];
  double L4_16[16][2];
  double L5_16[16][2];
  double T1_16[16][2];
  double T2_16[16];
  double T3_16[16][2];
  double T4_16[16][2];
  double T5_16[16][2];
  double TDL0_16[16][2];
  double TDLoffset16[16][2];
  double Tcoupler0_16[16][2];
  double Tdiplexer0_16[16][2];
  double Tnd0_16[16][2];
  double Tnd16[16][2];
  double Tomt0_16[16][2];
  double Txnd0_16[16][2];
  double Txnd16[16][2];
  double c1_DLoffset16[16][2];
  double c1_nd16[16][2];
  double c1_xnd16[16][2];
  double c2_DLoffset16[16][2];
  double c2_nd16[16][2];
  double c2_xnd16[16][2];
  double c3_DLoffset16[16][2];
  double c3_nd16[16][2];
  double c3_xnd16[16][2];
  double c4_DLoffset16[16][2];
  double c4_nd16[16][2];
  double c4_xnd16[16][2];
  double channel_phase16[16];
  double dgamma16[16][2];
  double lambda16[16][2];
  double nd_phase16[16];
  double psi16[16];
  double rx_phase16[16];
  double tcor16_im[16][2];
  double tcor16_re[16][2];
  double xnd_phase16[16];
};

struct fullCalGroup_struct {
  float cal_loss12_radome[MAXSCANS][2];
  float cal_loss1_reflector[MAXSCANS][2];
  float cal_loss2_feed[MAXSCANS];
  float cal_loss3_omt[MAXSCANS][2];
  float cal_loss4_coupler[MAXSCANS][2];
  float cal_loss5_diplexer[MAXSCANS][2];
  float cal_nd_phase[MAXSCANS];
  float cal_rx_phase[MAXSCANS];
  float cal_temp12_radome[MAXSCANS];
  float cal_temp1_reflector[MAXSCANS][2];
  float cal_temp2_feed[MAXSCANS];
  float cal_temp3_omt[MAXSCANS][2];
  float cal_temp4_coupler[MAXSCANS][2];
  float cal_temp5_diplexer[MAXSCANS][2];
  float cal_temp_nd[MAXSCANS];
  float cal_temp_ref[MAXSCANS][2];
  float cal_temp_xnd[MAXSCANS];
  float cal_temp_det[MAXSCANS][2];
  float cal_tempref_offset[MAXSCANS][2];
  float cal_tnd[MAXSCANS][2];
  float cal_tnd34[MAXSCANS];
  float cal_tref[MAXSCANS][2];
  float cal_txnd[MAXSCANS][2];
  float cal_phase[MAXSCANS];
  float cal_xnd_phase[MAXSCANS];
  float cal_psi[MAXSCANS];
  float cal_lambda[MAXSCANS][2];
  float cal_gamma[MAXSCANS][2];
  float cal_tcor_re[MAXSCANS][2];
  float cal_tcor_im[MAXSCANS][2];
};

struct subCalGroup_struct {
  float cal_loss5_diplexer16[MAXSCANS][16][2];
  float cal_loss4_coupler16[MAXSCANS][16][2];
  float cal_loss3_omt16[MAXSCANS][16][2];
  float cal_loss2_feed16[MAXSCANS][16];
  float cal_loss12_radome16[MAXSCANS][2];
  float cal_temp5_diplexer16[MAXSCANS][16][2];
  float cal_temp4_coupler16[MAXSCANS][16][2];
  float cal_temp3_omt16[MAXSCANS][16][2];
  float cal_temp2_feed16[MAXSCANS][16];
  float cal_temp12_radome16[MAXSCANS];
  float cal_temp_ref16[MAXSCANS][16][2];
  float cal_tempref_offset16[MAXSCANS][16][2];
  float cal_temp_nd16[MAXSCANS][16][2];
  float cal_tnd16[MAXSCANS][16][2];
  float cal_tref16[MAXSCANS][16][2];
  float cal_nd_phase16[MAXSCANS][16];
  float cal_temp_xnd16[MAXSCANS][16][2];
  float cal_txnd16[MAXSCANS][16][2];
  float cal_xnd_phase16[MAXSCANS][16];
  float cal_phase16[MAXSCANS][16];
  float cal_tnd34_16[MAXSCANS][16];
  float cal_psi16[MAXSCANS][16];
  float cal_lambda16[MAXSCANS][16][2];
  float cal_temp_det16[MAXSCANS][16][2];
  float cal_rx_phase16[MAXSCANS][16];
  float cal_gamma16[MAXSCANS][16][2];
  float cal_tcor16_re[MAXSCANS][16][2];
  float cal_tcor16_im[MAXSCANS][16][2];
  double calibration_time_seconds[MAXSCANS];
};

struct rfiThresh_k_struct {
  double *RFIthresh;
  double *RFIthresh16;
  double knom16_h[16];
  double knom16_v[16];
  double knom_h;
  double knom_v;
  double kstd16_h[16];
  double kstd16_v[16];
  double kstd_h;
  double kstd_v;
};

/*
struct rfiKurtosisParam_struct {
  rfiThresh_k_struct RFIthresh_k_cal;
  rfiThresh_k_struct RFIthresh_k_data;
};
*/

struct rfiFlagsKurtosis_struct {
  uint8_t *RFIflag_kurt_h;
  uint8_t *RFIflag_kurt_v;
  uint8_t *RFIflag16_kurt_h;
  uint8_t *RFIflag16_kurt_v;
  double *kurtosis_h;
  double *kurtosis_v;
  double *kurtosis16_h;
  double *kurtosis16_v;
};

struct rfiFlagsCalData_struct {
  uint8_t *RFIflag_ref_td1;
  uint8_t *RFIflag_refnd_td1;
  uint8_t *RFIflag16_ref_cf1;
  uint8_t *RFIflag16_refnd_cf1;
};

struct rfiFlags_struct {
  uint8_t *RFIflag_h_td1;
  uint8_t *RFIflag_v_td1;
  uint8_t *RFIflag_t3;
  uint8_t *RFIflag_t4;

  uint8_t *RFIflag16_h_cf1;
  uint8_t *RFIflag16_v_cf1;
  uint8_t *RFIflag16_h_cf11;
  uint8_t *RFIflag16_v_cf11;
  uint8_t *RFIflag16_t3;
  uint8_t *RFIflag16_t4;
};

struct calDataNoRFI_struct {
  float *m2_ref;
  float *m2_ref_nd;
  float *m2_16_ref;
  float *m2_16_ref_nd;
  float *t3_ref;
  float *t3_ref_nd;
  float *t3_16_ref;
  float *t3_16_ref_nd;
  float *t4_ref;
  float *t4_ref_nd;
  float *t4_16_ref;
  float *t4_16_ref_nd;
};

struct TAs_struct {
  float *TA_v;
  float *TA_h;
  float *T3;
  float *T4;
  float *Gainv;
  float *Gainh;
  float *Offsetv;
  float *Offseth;
  float *TA16_v;
  float *TA16_h;
  float *T3_16;
  float *T4_16;
  float *Gainv16;
  float *Gainh16;
  float *Offsetv16;
  float *Offseth16;
  float *ta_v;
  float *ta_h;
  float *ta_3;
  float *ta_4;
  float *ta_filtered_v;
  float *ta_filtered_h;
  float *ta_filtered_3;
  float *ta_filtered_4;
  float *nedt_v;
  float *nedt_h;
  float *nedt_3;
  float *nedt_4;

};

void readShort( hid_t fid, const char* field, short* data);
void readU16( hid_t fid, const char* field, uint16_t* data);
void readU32( hid_t fid, const char* field, uint32_t* data);
void readFloat( hid_t fid, const char* field, float* data);
void readDouble( hid_t fid, const char* field, double* data);

int loadCoeffs( const char* fullBand_filename, const char* subBand_filename,
                fullBandCoeffs_struct *fullBandCoeffs,
                subBandCoeffs_struct *subBandCoeffs);

int loadLinearCoeffs( const char* linearCoeffs_filename, 
                      linearCoeffs_struct *linearCoeffs);

int loadRFIKurtosisParam( const char* rfiKurtosisParameters_filename,
                          const char *groupName,
                          rfiThresh_k_struct *rfiKurtosisParameters);

int deleteRFIFlagsKurtosis( rfiFlagsKurtosis_struct *RFIflagsKurtosis);
      
int gainFilter( hsize_t nscans, size_t i, uint16_t *footprintsPerScan,  
                uint32_t nh, double *h, 
                double *gainv, double *gainh, double *gain, 
                double *offsetv, double *offseth,
                double *T3offset, double *T4offset, 
                double *Gainv, double *Gainh, double *Gain,
                double *Offsetv, double *Offseth,
                double *T3Offset, double *T4Offset);

int gainFilter16( hsize_t nscans, size_t i,
                  uint32_t *highresolution_scan_index,
                  uint16_t *footprintsPerScan,  
                  uint32_t nh, double *h, 
                  double time_gap, double *scan_diff1, double *scan_diff2, 
                  double *gainv16, double *gainh16, double *gain16, 
                  double *offsetv16, double *offseth16,
                  double *T3offset16, double *T4offset16, 
                  double *Gainv16, double *Gainh16, double *Gain16,
                  double *Offsetv16, double *Offseth16,
                  double *T3Offset16, double *T4Offset16);

int replaceNaN( double *refv, double *refh, double *ref_ndv, double *ref_ndh,
                uint32_t npix, uint nelem);

class l1aSMAPFile {
  int32_t h5fid;

 public:
  l1aSMAPFile();
  ~l1aSMAPFile();

  hsize_t nscans;
  hsize_t nhires;

  int32_t ngrps;
  hid_t gid[10];

  hid_t h5fid_ta;
  hid_t gid_ta;

  string halfOrbitStartDateTime;
  string halfOrbitStopDateTime;
  string orbitDirection;
  int32_t orbitPeriod;

  int32_t *rfiThreshidx;
  int32_t *rfiThreshidx16;
  int32_t *rfiThreshidxFP;

  int openl1( const char* l1a_filename, const char* ta_filename,
              string history);

  int lookFB( hsize_t n, uint32_t nPRI, double *sec, uint8_t *FB,
              float *scan_angle);

  int allocateRFIFlagsKurtosis( uint32_t nPRI, 
                                rfiFlagsKurtosis_struct *RFIflagsKurtosis);

  int calCoefficient( fullBandCoeffs_struct *fullBandCoeff,
                      subBandCoeffs_struct *subBandCoeff,
                      fullCalGroup_struct *fullCalGroup,
                      subCalGroup_struct *subCalGroup,
                      uint32_t *subband_scan_counter);

  int kurtosisCalData( const char *refType,
                       double lat1, double lon1,
                       uint32_t *highresolution_scan_index,
                       rfiThresh_k_struct *rfiKurtosisParameters,
                       rfiFlagsKurtosis_struct *rfiFlagsKurt);

  int kurtosis_RFI_Detection( const char *ref_ant,
                              double *lat, double *lon,
                              uint32_t *highresolution_scan_index,
                              float *m1, float *m2, 
                              float *m3, float *m4,
                              float *m1_16, float *m2_16, 
                              float *m3_16, float *m4_16,
                              rfiThresh_k_struct *rfiKurtosisParameters,
                              rfiFlagsKurtosis_struct *rfiFlagsKurt);

  int rfiDetectCalData( const char *rfiParameters_filename,
                        uint32_t *highresolution_scan_index,
                        uint16_t *footprintsPerScan,
                        double lat1, double lon1,
                        uint32_t nPRI_fp,
                        rfiFlagsCalData_struct *RFIflagsCalData);

  int rfiDetect( const char *rfiParameters_filename,
                 uint32_t *highresolution_scan_index,
                 uint16_t *footprintsPerScan,
                 uint32_t nPRI_fp,
                 TAs_struct *TAs,
                 rfiFlags_struct *RFIflags);

  int rfiDetectTimeDomain( const char *ref_ant,
                           uint32_t nPRI_fp, uint32_t nCounts,
                           float M, float *TAtd, float *Trec,
                           double AntMax_td, double BWtau,
                           double *RFIthresh_td, 
                           int32_t *rfiThreshidx,
                           uint8_t *RFIflag_td);

  int rfiDetectCrossFreq( const char *ref_ant,
                          uint32_t nCounts, uint32_t M, 
                          float *TAcf, float *Trec16,
                          double AntMax_cf, double BW, double tau16,
                          double *RFIthresh_cf, 
                          int32_t *rfiThreshidx,
                          uint8_t *RFIflag_cf);

  int rfiRemovalCalData( uint16_t *footprintsPerScan,
                         uint32_t *highresolution_scan_index,
                         rfiFlagsKurtosis_struct *RFIflagsKurt,
                         rfiFlagsKurtosis_struct *RFIflagsKurtND,
                         rfiFlagsCalData_struct *RFIflagsCalData,
                         calDataNoRFI_struct *calDataNoRFI);

  int radomeCharacteristics( const char *reflectorFilename,
                             string surface_temp_selection,
                             double cal_loss12_radome[MAXSCANS][2],
                             double cal_temp12_radome[MAXSCANS]);

  int reflectorCharacteristics( const char *reflectorFilename,
                                double cal_loss1_refl[MAXSCANS][2],
                                double cal_temp1_refl[MAXSCANS][2],
                                double cal_TB1_refl[MAXSCANS][2]);

  int calibrationNew_FPS( linearCoeffs_struct *linearCoeffs, 
                          fullBandCoeffs_struct *fullBandCoeffs, 
                          subBandCoeffs_struct *subBandCoeffs,
                          fullCalGroup_struct *fullCalGroup, 
                          subCalGroup_struct *subCalGroup, 
                          calDataNoRFI_struct *calDataNoRFI,
                          uint32_t *highresolution_scan_index, 
                          uint16_t *footprintsPerScan, TAs_struct *TAs);

  int MAXPD_Algorithm( TAs_struct *TAs, 
                       rfiFlags_struct *RFIflagsAnt, 
                       rfiFlagsKurtosis_struct *rfiFlagsKurt,
                       uint32_t *highresolution_scan_index);
  
  int L1B_bit_flags( float *tah, float *tah_filtered, float *nedt_h,
                     float *tav, float *tav_filtered, float *nedt_v,
                     float *ta3, float *ta3_filtered, float *nedt_3,
                     float *ta4, float *ta4_filtered, float *nedt_4,
                     uint16_t *tb_qual_flag_h,
                     uint16_t *tb_qual_flag_v,
                     uint16_t *tb_qual_flag_3,
                     uint16_t *tb_qual_flag_4);
  
  int closel1( TAs_struct *TAs, double cal_loss1_refl[MAXSCANS][2],
               double cal_temp1_refl[MAXSCANS][2]);

};

