#include <time.h>
#include <math.h>
#include <iostream>
#include <iomanip>
#include <stdlib.h>
#include <string.h>
#include <fstream>
#include "hdf5utils.h"
 
//  Modification history:
//
//  V0.10 16-Jun-11  JMG   Initial version
//  V0.15 20-Jun-11  JMG   Change input to yyyymmdd
//  V0.20 28-Jun-11  JMG   Add "ALL" support for yyyymmdd
//  V0.90 27-Jul-11  JMG   Use launchtime as lower limit for ALL
//  V0.91 01-Aug-11  JMG   Change gps_hkt to cdh_hkt
//  V1.00 10-Aug-11  JMG   Update to V 1.00
//  V1.10 11-Jun-13  JMG   Trap and repair bad HKT times
//                         due to EF8 + 8 carry bug.
//  V1.11 13-Jun-13  JMG   Add -247 to carry bug trap
//                         Write corrected CDH time to
//                         HKT record

#define VOIDP void*
#define LAUNCHTIME 991750828

#define VERSION "1.11"

extern "C" {
  void sedgesort (uint32_t*, uint32_t*, int);
}

using namespace std;

int main (int argc, char* argv[])
{
  uint32_t nrec=0;
  uint32_t nuniq=0;

  uint32_t cdh_hkt0, min_gps, max_gps, prev_cdh=0;
  char *cptr = (char *) &cdh_hkt0;
  int32_t diff_cdh;

  uint8_t hktrec[4000];

  ifstream hktfile;
  ofstream outfile;

  uint32_t recsize=4000;

  cout << "hktmerge_aquarius " << VERSION << " (" 
       <<  __DATE__ << " " << __TIME__ << ")" << endl;

  if ( argc == 1) {
    cout << endl << 
      "hktmerge_aquarius input_hkt_file yyyymmdd output_directory [output filename]" 
	 << endl;
    cout << "Use ALL as 2nd argument to merge all entries" << endl;
    return 0;
  }

  // Open housekeeping file and get # of records
  hktfile.open(argv[1], fstream::in | fstream::binary);
  hktfile.seekg (0, ios::end);
  uint32_t totalbytes = hktfile.tellg();
  hktfile.seekg (0, ios::beg);

  nrec = totalbytes / recsize;
  uint32_t *cdh_hkt = (uint32_t *) calloc(nrec+1, sizeof(uint32_t));
  uint32_t *dup = (uint32_t *) calloc(nrec, sizeof(int32_t));

  // Get DOY from YYYYMMDD
  static int startOfMonth[2][12] = 
    {{0,31,59,90,120,151,181,212,243,273,304,334},
     {0,31,60,91,121,152,182,213,244,274,305,335}};

  istringstream istr;
  uint32_t iyear, imon, iday;
  string str;

  str = argv[2];
  if ( str.compare("ALL") != 0) {
    istr.clear(); istr.str( str.substr(0,4)); istr >> iyear;
    istr.clear(); istr.str( str.substr(4,2)); istr >> imon;
    istr.clear(); istr.str( str.substr(6,2)); istr >> iday;

    int isleap = 0;
    if ( ((iyear % 400) == 0) || (((iyear % 4) == 0) && ((iyear % 100) != 0)) )
      isleap = 1;

    uint32_t doy = startOfMonth[isleap][imon-1] + iday;

    min_gps = get_tai( iyear, doy, 0);
    max_gps = min_gps + 24*3600;
  } else {
    min_gps = LAUNCHTIME;
    max_gps = 1e30;
  }

  for (size_t i=0; i<nrec; i++) {
    hktfile.read( (char *) hktrec, recsize);

    // CDH
    memcpy( cptr+0, &hktrec[8+92+3], 1);
    memcpy( cptr+1, &hktrec[8+92+2], 1);
    memcpy( cptr+2, &hktrec[8+92+1], 1);
    memcpy( cptr+3, &hktrec[8+92+0], 1);

    // OBT
    //    memcpy( cptr+0, &hktrec[580+816+2], 1);
    //memcpy( cptr+1, &hktrec[580+816+3], 1);
    //memcpy( cptr+2, &hktrec[580+816+0], 1);
    //memcpy( cptr+3, &hktrec[580+816+1], 1);

    diff_cdh = cdh_hkt0 - prev_cdh;
    if ( (diff_cdh != -248) && (diff_cdh != -247)) {
      cdh_hkt[i] = cdh_hkt0;
    } else {
      cout << "EF8 + 8 carry bug found in record (0-based): ";
      cout << i << endl;
      cdh_hkt[i] = prev_cdh + 8;
    }

    //    if ( i > 0 && (cdh_hkt0 - prev_gps) != 8) {
    //  cout << "Abnormal time gap: " << prev_gps << " " << cdh_hkt0 << endl;
    // }

    prev_cdh = cdh_hkt[i];
  }

  // Check for dups
  nuniq = 1;
  for (size_t i=0; i<nrec; i++) {

    if ( cdh_hkt[i] < min_gps || cdh_hkt[i] >= max_gps) {
      dup[i] = 0;
      continue;
    }

    // if record not yet dup
    if (dup[i] == 0) {
      // loop through following records
      for (size_t j=i+1; j<nrec; j++) {
	// if not yet dup & same time as inital record then tag with dup number
	if (dup[j] == 0 && cdh_hkt[i] == cdh_hkt[j])
	  dup[j] = nuniq;
      }
      dup[i] = nuniq;
      nuniq++;
    }
  }
  nuniq--;

  cout << left << setw(25) << "Number of Records:";
  cout << right << setw(10) << nrec << endl;

  // Determine first uniq record
  for (size_t i=0; i<nuniq; i++) {
    for (size_t j=0; j<nrec; j++) {
      if (dup[j] == i+1) {
	dup[j] = nuniq + 1;
	break;
      }
    }
  }


  // Sort unique records by CDH
  uint32_t k = 0;
  uint32_t  *idx = 
    (uint32_t *) calloc(nuniq+1, sizeof(uint32_t));
  uint32_t *recidx = 
    (uint32_t *) calloc(nuniq+1, sizeof(uint32_t));
  for (size_t i=0; i<nrec; i++) {
    if (dup[i] == (nuniq + 1)) {
      cdh_hkt[k] = cdh_hkt[i];
      recidx[k] = i;
      idx[k] = k;
      k++;
    }
  }

  cdh_hkt[k] = (uint32_t) (0xffffffff);
  idx[k] = k;
  sedgesort (cdh_hkt, idx, nuniq+1);


  // Output sorted L0 records
  char *outbuf = (char *) calloc(nuniq*recsize, sizeof(char));
  prev_cdh = 0;
  for (size_t i=0; i<nuniq; i++) {
    hktfile.seekg (recidx[idx[i]]*recsize, ios::beg);
    hktfile.read( (char *) hktrec, recsize);

    // CDH
    memcpy( cptr+0, &hktrec[8+92+3], 1);
    memcpy( cptr+1, &hktrec[8+92+2], 1);
    memcpy( cptr+2, &hktrec[8+92+1], 1);
    memcpy( cptr+3, &hktrec[8+92+0], 1);

    diff_cdh = cdh_hkt0 - prev_cdh;
    if ( (diff_cdh == -248) || (diff_cdh == -247)) {
      cdh_hkt0 = prev_cdh + 8;
      memcpy( &hktrec[8+92+3], cptr+0, 1);
      memcpy( &hktrec[8+92+2], cptr+1, 1);
      memcpy( &hktrec[8+92+1], cptr+2, 1);
      memcpy( &hktrec[8+92+0], cptr+3, 1);
    }
    prev_cdh = cdh_hkt0;

    memcpy(&outbuf[i*recsize], hktrec, recsize);
  }

  cout << left << setw(25) << "Number of Unique Records:";
  cout << right << setw(10) << nuniq << endl;

  time_t rawtime;
  tm * ptm;
  char buffer[80];
  rawtime = (time_t) gpstai2unix(min_gps);
  ptm = gmtime ( &rawtime );
  strftime (buffer, 80, "%Y%m%d", ptm);

  // Create output file
  ostringstream outstr;

  // If 4th parameter use for output name
  if ( argc == 5) {
    outstr << argv[3] << "/" << argv[4];
    str = outstr.str();
  } else {
    outstr << argv[3] << "/GSFC_" << buffer;
    str = outstr.str() + "_SACD_HKTMRG.bin";
  }

  cout << "outfile=" << str.c_str() << endl;
  outfile.open(str.c_str(), fstream::out | fstream::binary | fstream::trunc);
  outfile.write( (char *) outbuf, nuniq*recsize);

  free(dup);
  free(cdh_hkt);

  free(recidx);
  free(idx);
  free(outbuf);
}

