#include <stdio.h>
#include <time.h>
#include <math.h>
#include <iostream>
#include <sstream>
#include <iomanip>
#include <libgen.h>
#include "FileStager.h"
#include "hdf5_Aquarius.h"
#include "hdf5utils.h"
 
//  Modification history:
//
//  V0.20 30-Jun-09  JMG   Initial version
//  V0.25 02-Aug-09  JMG   Use gpstai2unix & unix2gpstai functions
//  V0.30 21-Sep-09  JMG   Sort on ICDS block number rather than time
//  V0.35 29-Sep-09  JMG   Perform checksum
//  V0.36 01-Oct-09  JMG   Choose first good checksum record in duplicates
//  V0.37 06-Nov-09  JMG   Add icds_blknum when flagging bad checksums
//  V0.38 17-Nov-09  JMG   Check for bad RAD header value
//  V0.39 11-Dec-09  JMG   Print bad RAD header values, don't exit
//  V0.40 11-Feb-11  JMG   Support CGSS downlink files
//  V0.41 28-Feb-11  JMG   Add support for 3rd parameter (output filename)
//  V0.42 10-Mar-11  JMG   Skip records with bad headers
//  V0.43 30-Mar-11  JMG   Output # of badchecksum records
//  V0.44 14-Apr-11  JMG   Don't modify output name if given
//  V1.00 05-Aug-11  JMG   Don't skip bad rad headers (just notify)
//  V1.01 14-Aug-11  JMG   Step every byte for magic word search
//                         Check for non-zero time for output records
//  V1.02 16-Aug-11  JMG   Report byte offset (0-3)
//  V1.03 16-Aug-11  JMG   Use findSync() to search for magic word for all
//                         records
//  V1.04 22-Dec-11  JMG   Fix "4-sec" time jump (pitch) glitches
//                         and report to text file
//  V1.05 30-Jan-12  JMG   Use icds_blknum to find "4-sec" time jumps
//  V1.06 21-Feb-12  JMG   Subtract 2.0 rather than 2 from nsc
//  V1.07 26-Nov-12  JMG   Move bad radheader check after dup removal and sort
//  V1.08 08-Jan-13  JMG   Update icds_blknum when skipping bad header
//  V1.09 04-Aug-14  JMG   Report first corrected time rather than last for
//                         4-sec glitch


#define VOIDP void*

#define VERSION "1.09"

extern "C" {
  //  void sedgesort (uint64_t*, uint32_t*, int);
  void sedgesort (uint32_t*, uint32_t*, int);
}

using namespace std;

int main (int argc, char* argv[])
{
  uint32_t nrec=0;
  uint32_t nuniq=0;
  int32_t magic=0xefbeadde; // deadbeef rotated

  uint32_t tai0;
  uint32_t nsc0;
  int32_t magic0;
  uint32_t icds0;

  uint32_t recsize=2560;
  char *ptr, **ptrs;

  cout << "l0gen_aquarius " << VERSION << " (" 
       <<  __DATE__ << " " << __TIME__ << ")" << endl;

  if ( argc == 1) {
    cout << endl << "l0gen_aquarius input_downlink_file output_directory [output filename]" 
	 << endl;
    return 0;
  }

  static Aq::Fw::Ws::Io::FileStager infile;
  static Aq::Fw::Ws::Io::FileStager outfile;

  Aq::Fw::Ws::Io::FileStager::NameType ntype =
    Aq::Fw::Ws::Io::FileStager::NAME_IS_FILENAME;

  // Stage input file (argv[1])
  Aq::Fw::Ws::Io::NumBytesAvail n = infile.stageFile(argv[1], ntype, magic);
  if ( n == 0) {
    cout << "No records found in: " << argv[1] << endl;
    exit(1);
  }

  nrec = infile.numRecs(recsize);
  //  std::cout << "bytes left: " << infile.bytesLeft() << std::endl;

  // Read TAI & 62.5 nanosec values (seconds from Jan 6, 1980)
  uint32_t *tai = (uint32_t *) calloc(nrec, sizeof(int32_t));
  uint32_t *nsc = (uint32_t *) calloc(nrec, sizeof(int32_t));
  uint32_t *dup = (uint32_t *) calloc(nrec, sizeof(int32_t));

  uint32_t *icds_blknum = (uint32_t *) calloc(nrec+1, sizeof(int32_t));

  ptrs = (char **) calloc(nrec, sizeof(char *));

  ptr = infile.getNext(4);  // offset by magic word length
  for (size_t i=0; i<nrec; i++) {
    ptrs[i] = ptr - 4;
    memcpy(&magic0, ptrs[i], 4);

    ptr = infile.getNext(sizeof(int32_t), (char *) &tai0);
    ptr = infile.getNext(sizeof(int32_t), (char *) &nsc0);
    ptr = infile.getNext(5+8);  // Skip deploy and ICDS status
    ptr = infile.getNext(sizeof(int32_t), (char *) &icds0);

    tai[i] = SWAP_4(tai0);
    nsc[i] = SWAP_4(nsc0);
    icds_blknum[i] = SWAP_4(icds0);

    ptr = infile.findSync( magic, 4);
  }
  ptr = infile.rewind(magic);


  // Compute checksums
  uint16_t ui16;
  uint16_t checksum;
  uint16_t computed_checksum;
  uint8_t *l0buf = (uint8_t *) calloc(recsize, sizeof(uint8_t));
  bool *badchecksum = (bool *) calloc(nrec, sizeof(bool));
  ptr = infile.getNext(0);
  for (size_t i=0; i<nrec; i++) {
    badchecksum[i] = false;

    memcpy(l0buf, ptr, recsize);

    memcpy(&computed_checksum, l0buf, sizeof(uint16_t));
    for (size_t j=1; j<recsize/2-1; j++) {
      memcpy(&ui16, &l0buf[j*sizeof(uint16_t)], sizeof(uint16_t));
      computed_checksum ^= ui16;
    }
    memcpy(&checksum, &l0buf[recsize-2], sizeof(uint16_t));

    if ( computed_checksum != checksum) {
      badchecksum[i] = true;
      cout.precision(5);

      double tmedbl = tai[i] - ((nsc[i]-2.0) / 
				((double) 16000000.)) + 0.0000001;
      cout << "Checksum failure: " << setiosflags(ios::fixed)
	   << right << setw(7) << i << " " << right << setw(7)
	   << computed_checksum << " " << right << setw(7) << checksum << " "
	   << setprecision(3) << setw(15)
	   << tmedbl << " " << icds_blknum[i] << endl;
    }
    ptr = infile.findSync( magic, 0);
  }
  ptr = infile.rewind(magic);


  // Check for dups (skipping bad checksum records)
  nuniq = 1;
  for (size_t i=0; i<nrec; i++) {
    if ( badchecksum[i]) {
      dup[i] = 0;
      continue;
    }

    // if record not yet dup
    if (dup[i] == 0) {
      // loop through following records
      for (size_t j=i+1; j<nrec; j++) {
	// if not yet dup & same time as inital record then tag with dup number
	//	if (dup[j] == 0 && tme64[i] == tme64[j])
	if (dup[j] == 0 && icds_blknum[i] == icds_blknum[j])
	  dup[j] = nuniq;
      }
      dup[i] = nuniq;
      nuniq++;
    }
  }
  nuniq--;

  //  for (size_t i=0; i<nrec; i++) {
  //  if ( dup[i] == 0) cout << i << endl;
  //}


  cout << left << setw(25) << "Number of Records:";
  cout << right << setw(10) << nrec << endl;


  // Determine first uniq record
  for (size_t i=0; i<nuniq; i++) {
    for (size_t j=0; j<nrec; j++) {
      if (dup[j] == i+1) {
	dup[j] = nuniq + 1;
	break;
      }
    }
  }


  // Sort unique records by ICDS blknum
  uint32_t k = 0;
  uint32_t  *idx = 
    (uint32_t *) calloc(nuniq+1, sizeof(uint32_t));
  uint32_t *recidx = 
    (uint32_t *) calloc(nuniq+1, sizeof(uint32_t));
  for (size_t i=0; i<nrec; i++) {
    if (dup[i] == (nuniq + 1)) {
      //      tme64[k] = tme64[i];
      icds_blknum[k] = icds_blknum[i];
      recidx[k] = i;
      idx[k] = k;
      k++;
    }
  }
  //  tme64[k] = (uint64_t) (0xffffffff) * 0xffffffff;
  icds_blknum[k] = (uint32_t) (0xffffffff);
  idx[k] = k;
  //  sedgesort (tme64, idx, nuniq+1);
  sedgesort (icds_blknum, idx, nuniq+1);


  // Output sorted L0 records
  char *outbuf = (char *) calloc(nuniq*recsize, sizeof(char));
  ptr = infile.rewind(magic);
  k = 0;
  for (size_t i=0; i<nuniq; i++) {
    //    memcpy(&outbuf[k*recsize], &ptr[recidx[idx[i]]*recsize], recsize);
    memcpy(&outbuf[k*recsize], ptrs[recidx[idx[i]]], recsize);

    // Check for magic word at beginning of record
    memcpy(&magic0, &outbuf[k*recsize], 4);


    // Check for correct RAD header
    // (64080   64081   64082   64083) Normal
    // ( 6064    6065    6066    6067) Memory Dump
    uint16_t rad_header;
    memcpy(&rad_header, &outbuf[k*recsize+RADIOMETER_SCI_OFFSET], 2);
    rad_header = SWAP_2(rad_header);
    if ( rad_header != 64080 && rad_header != 64081 &&
	 rad_header != 64082 && rad_header != 64083 &&
	 rad_header !=  6064 && rad_header !=  6065 &&
	 rad_header !=  6066 && rad_header !=  6067) {

      cout << "*** Improper RAD header value: " << setiosflags(ios::fixed) 
	   << right << setw(7) << rad_header 
      	   << " " << right << setw(7) << i << setw(15) << icds_blknum[i]
	   << endl;
    }

    // Check for non-zero tai (gps) time
    // Note tai0 should be byte-swapped to give correct tai value
    memcpy(&tai0, &outbuf[k*recsize+4], 4);
    if ( magic0 == magic && tai0 > 0) {
      icds_blknum[k] = icds_blknum[i];
      k++;
    }
  }

  nuniq = k;

  // Find first "non-bump" entry by checking for 2.88 sec
  // difference in adjacent blocks
  // Note: This will not find bump at first data point
  uint32_t tai_1, tai_2;
  uint32_t nsc_1, nsc_2;
  double tme1, tme2;
  for (size_t i=1; i<nuniq; i++) {
    memcpy(&tai_1, &outbuf[(i-1)*recsize+4], sizeof(uint32_t));
    memcpy(&nsc_1, &outbuf[(i-1)*recsize+8], sizeof(uint32_t));
    tme1 = SWAP_4(tai_1)-(SWAP_4(nsc_1)-2.0)*62.5e-9;

    memcpy(&tai_2, &outbuf[(i+1)*recsize+4], sizeof(uint32_t));
    memcpy(&nsc_2, &outbuf[(i+1)*recsize+8], sizeof(uint32_t));
    tme2 = SWAP_4(tai_2)-(SWAP_4(nsc_2)-2.0)*62.5e-9;

    if ( fabs((tme2-tme1) - 2.88) < 1e-6) {
      k = i;
      break;
    }
  }
  // cout << "first good: " << k << endl;

  // Compare current block time with previous one.
  // If not 1.44 sec then set time to 1.44 seconds after previous one
  double tme2_new = 0, tme2_new_first = 0;
  int32_t first = 1;
  for (size_t i=k; i<nuniq; i++) {
    memcpy(&tai_1, &outbuf[(i-1)*recsize+4], sizeof(uint32_t));
    memcpy(&nsc_1, &outbuf[(i-1)*recsize+8], sizeof(uint32_t));
    tme1 = SWAP_4(tai_1)-(SWAP_4(nsc_1)-2.0)*62.5e-9;

    memcpy(&tai_2, &outbuf[(i)*recsize+4], sizeof(uint32_t));
    memcpy(&nsc_2, &outbuf[(i)*recsize+8], sizeof(uint32_t));
    tme2 = SWAP_4(tai_2)-(SWAP_4(nsc_2)-2.0)*62.5e-9;

    double diff = fabs(tme2-tme1)/1.44;
    if ( floor(diff) != (icds_blknum[i]-icds_blknum[i-1])) {
    //    if ( (diff - floor(diff)) > 1e-6) {
      cout << endl << "Bad time jump at: " 
	   << setprecision(2) << setw(15) << fixed
	   << tme2 << " in " << argv[1] << endl;

      tme2_new = tme1 + (double) 1.44;
      if ( first) {
        tme2_new_first = tme2_new;
        first = 0;
      }

      cout << "Corrected time:   " << setprecision(2) << setw(15) << fixed
	   << tme2_new << endl << endl;

      uint32_t tai_2_new;
      if ( (SWAP_4(nsc_1)-2.0)*62.5e-9 < 0.44)
	tai_2_new = SWAP_4(tai_1) + 2;
      else
	tai_2_new = SWAP_4(tai_1) + 1;

      uint32_t nsc_2_new = (tai_2_new - tme2_new) / 62.5e-9 + 2;

      //      cout << "fixed: " <<
      //tai_2_new-(nsc_2_new-2)*62.5e-9 - tme1 << endl;

      tai_2 = SWAP_4(tai_2_new);
      nsc_2 = SWAP_4(nsc_2_new);
      memcpy(&outbuf[(i)*recsize+4], &tai_2, sizeof(uint32_t));
      memcpy(&outbuf[(i)*recsize+8], &nsc_2, sizeof(uint32_t));
    }
  }

  cout << left << setw(25) << "Number of Unique Records:";
  cout << right << setw(10) << nuniq << endl;

  time_t rawtime;
  tm * ptm;
  char buffer1[80];
  char buffer2[80];
  char buffer3[80];

  memcpy(&tai0, &outbuf[0*recsize+4], 4);
  memcpy(&nsc0, &outbuf[0*recsize+8], 4);

  rawtime = (time_t) gpstai2unix(SWAP_4(tai0)-(SWAP_4(nsc0)-2.0)*62.5e-9);
  ptm = gmtime ( &rawtime );
  strftime (buffer1, 80, "%Y-%m-%d %H:%M:%S", ptm);
  cout << endl;
  cout << "start_time=" << buffer1 << endl;

  memcpy(&tai0, &outbuf[(nuniq-1)*recsize+4], 4);
  memcpy(&nsc0, &outbuf[(nuniq-1)*recsize+8], 4);

  rawtime = (time_t) gpstai2unix(SWAP_4(tai0)-(SWAP_4(nsc0)-2.0)*62.5e-9);
  ptm = gmtime ( &rawtime );
  strftime (buffer2, 80, "%Y-%m-%d %H:%M:%S", ptm);
  strftime (buffer3, 80, "%Y%m%d_%H%M%S", ptm);
  cout << "stop_time= " << buffer2 << endl;


  // Create output file (from buffer3)
  ostringstream outstr;
  string str;

  // If 3rd parameter use for output name
  if ( argc == 4) {
    outstr << argv[2] << "/" << argv[3];
    str = outstr.str();
  } else {
    outstr << argv[2] << "/" << buffer3;
    str = outstr.str() + ".AQ_L0";
  }

  cout << "outfile=" << str.c_str() << endl;
  Aq::Fw::Ws::Io::NumBytesAvail nout = 
    outfile.createFile(str.c_str(), ntype, nuniq*recsize, outbuf);
  if ( nout == 0) {
    cout << "Error creating: " << str.c_str() << endl;
    exit(1);
  }

  // Create output text file
  ofstream fout;
  str = str + ".txt";
  fout.open(str.c_str());
  fout << "Number of Records: ";
  fout << nuniq << endl;
  fout << "Start Time: " << buffer1 << endl;
  fout << "Stop Time:  " << buffer2 << endl;
  uint32_t n_badchecksum = 0;
  for (size_t i=0; i<nrec; i++) 
    if ( badchecksum[i] == true) n_badchecksum++;
  fout << "Number of records with bad checksum:  " << n_badchecksum << endl;
  fout << "Byte Offset (0-3): " << (n - nrec*recsize) % 4 << endl;

  // Report 4-sec time glitch if any to text file.
  if ( tme2_new_first != 0) {
    rawtime = (time_t) gpstai2unix(tme2_new_first);
    ptm = gmtime ( &rawtime );
    strftime (buffer2, 80, "%Y-%m-%d %H:%M:%S", ptm);
    fout << "4-sec time glitch at: " << buffer2 << ".";

    // Fractional second
    fout << setprecision(0) << setw(2) << fixed << setfill ('0') 
	 << 100 * (tme2_new - floor(tme2_new_first)) << endl;
  }


  fout.close();

  free(tai);
  free(nsc);
  free(dup);
  free(icds_blknum);
  //  free(tme64);
  free(recidx);
  free(idx);
  free(outbuf);
  free(badchecksum);
  free(ptrs);
}

