#include <stdio.h>
#include <math.h>
#include <time.h>

#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <list>
#include <sys/wait.h>
#include <libgen.h>
#include <timeutils.h>
#include "FileStager.h"
#include "hdf5_Aquarius.h"
#include "hdf5utils.h"

#define MAXOPENFILES 50

// Wing size in minutes (Additional records before & after orbit)
#define WINGLEN 10

#define VERSION "1.11"

//    Modification history:
//          Programmer     Organization      Date      Description of change
//        --------------   ------------    --------    ---------------------
//        Joel Gales       FutureTech      04/02/08    Original development
//        Joel Gales       FutureTech      08/02/09    Use gpstai2unix &
//                                                     unix2gpstai functions
//        Joel Gales       FutureTech      08/06/09    Add converted tlm
//        Joel Gales       FutureTech      09/02/09    Add support for non-
//                                                     eph/att processing.
//        Joel Gales       FutureTech      09/25/09    Incorporate radiometer
//                                                     fine temp telemetry cnv.
//        Joel Gales       FutureTech      09/29/09    Incorporate ICDS fine 
//                                                     temp telemetry conv.
//        Joel Gales       FutureTech      10/27/09    Allow 2 orb/att recs 
//                                                     after last blk_sec for 
//                                                     scat interp
//        Joel Gales       FutureTech      11/10/09    Tag I&T files with "InT"
//        Joel Gales       FutureTech      04/15/10    Check for existance of
//                                                     eph & att files.
//        Joel Gales       FutureTech      06/20/10    Add "Input Files" &
//                                                     "Processing Control"
//                                                     metadata.
//        Joel Gales       FutureTech      11/03/10    Trap bad ephem files
//        Joel Gales       FutureTech      02/23/11    Add SAC-D HKT record
//                                                     write
//        Joel Gales       FutureTech      03/28/11    Fix to run without
//                                                     eph & att (hkt required)
//        Joel Gales       FutureTech      04/20/11    Add cold sky cal to
//                                                     att_flags
//        Joel Gales       FutureTech      06/01/11    Fix to run without
//                                                     eph & att & hkt
//        Joel Gales       FutureTech      06/29/11    Add outfile tag option
//        Joel Gales       FutureTech      08/10/11    Update to V 1.00
//        Joel Gales       FutureTech      08/14/11    Skip L0 records after
//                                                     last att record
//        Joel Gales       FutureTech      08/15/11    Fix HKT write problem
//                                                     (Error in argv index)
//        Joel Gales       FutureTech      08/21/11    Reformat print output
//                                                     to make comparison of
//                                                     orbit limits with L1A
//                                                     filenames easier
//        Joel Gales       FutureTech      08/24/11    Fix overwriting of
//                                                     blknum, rad/anc frmnum
//                                                     metadata in closel1()
//        Joel Gales       FutureTech      09/09/11    Update metadata
//                                                     from Liang's comments.
//        Joel Gales       FutureTech      10/28/11    Pass version number to
//                                                     createl1
//        Joel Gales       FutureTech      10/28/11    Improve diagnostic
//                                                     error messages
//        Joel Gales       FutureTech      02/14/12    Fix problem with delta
//                                                     time != 1.44 sec
//                                                     (Fix in ydhmsf().) 
//        Joel Gales       FutureTech      02/21/12    Subtract 2.0 rather
//                                                     than 2 from nsc
//        Joel Gales       FutureTech      06/05/12    Fix bad rad_frmnum bug
//        Joel Gales       FutureTech      07/04/12    Remove error for
//                                                     delta(ephtme) != 60


using namespace std;

int get_tme_z( string *, double *, double *);

int main (int argc, char* argv[])
{
  cout << "l1agen_aquarius " << VERSION << " (" 
       <<  __DATE__ << " " << __TIME__ << ")" << endl;

  if ( argc == 1) {
    cout << endl << 
      "l1agen_aquarius l0_file output_dir ephemfile attfile hktfile [outfile_tag]"
	 << endl;
    return 0;
  }

  int32_t magic=0xefbeadde; // deadbeef rotated

  uint32_t tai0;
  uint32_t nsc0;
  double granStart, granStop;

  const int recsize=2560;
  char *ptrL0, *cptr;
  char namebuf[512];
  char timebuf[80];
  char outfile_ext[32];
  strcpy( outfile_ext, ".AQ_L1A");

  const double oneth = 1000.0000000000;
  struct tm *tm0; 
 
  string ydhmsf_str;

  string line;

  std::ostringstream ostr;

  int annotate_eph(ifstream *, unsigned char*, double *, double *, uint32_t);

  static Aq::Fw::Ws::Io::FileStager infile;

  Aq::Fw::Ws::Io::FileStager::NameType ntype =
    Aq::Fw::Ws::Io::FileStager::NAME_IS_FILENAME;

  string processControl = "";
  for (int i=0; i<argc; i++) {
    processControl += argv[i];
    processControl += " ";
  }
  processControl = processControl.substr( 0, processControl.length()-1);


  // Stage input file (argv[1])
  Aq::Fw::Ws::Io::NumBytesAvail n = infile.stageFile(argv[1], ntype, magic);
  if ( n == 0) {
    cout << "Error opening: " << argv[1] << endl;
    exit(1);
  }

  uint32_t nrecL0 = infile.numRecs(recsize);

  uint32_t n_files_open = 0;
  uint32_t n_files_close = 0;
  uint32_t tai_start[MAXOPENFILES];
  uint32_t tai_stop[MAXOPENFILES];

  uint32_t nrec_eph = 0;
  uint32_t nrec_att = 0;
  uint32_t nrec_hkt = 0;
  unsigned char *orb=NULL;
  double *ephtme=NULL;
  double *ephz=NULL;
  double *atttme=NULL;
  uint32_t *hkttme=NULL;

  ifstream ephfile;
  ifstream attfile;
  ifstream hktfile;
  uint8_t hktrec[4000];

  int system_retstat;

  // Add outfile_ext tag if present
  if ( argc == 7) strcat( outfile_ext, argv[6]);

  if ( argc >= 6) {
    // Stage & annotate ephemeris file
    cout << endl << 
      "Stage/Annotate Ephemeris file and Stage Attitude file." << endl;
    char *ephname = argv[3];
    ephfile.open( ephname, ifstream::in);
    if (ephfile.is_open()) {
      getline( ephfile, line);
      istringstream iStream(line);
      iStream >> nrec_eph;
    } else {
      cout << ephname << " not found." << endl;
      exit(1);
    }
    orb = (unsigned char *) calloc(nrec_eph, sizeof(char));
    ephtme = (double *) calloc(nrec_eph, sizeof(double));
    ephz   = (double *) calloc(nrec_eph, sizeof(double));
    annotate_eph( &ephfile, orb, ephtme, ephz, nrec_eph);

    // Check if no "<" (N min before orbit start)
    size_t i0;
    for (i0=0; i0<nrec_eph; i0++) if (orb[i0] == '<') break;
    if (i0 == nrec_eph && nrec_eph > 0) {
      cout << "No '<' symbols in orb array." << endl;
      exit(110);
    }

      
    // Stage attitude file
    char *attname = argv[4];
    attfile.open( attname, ifstream::in);
    if (attfile.is_open()) {
      getline( attfile, line);
      istringstream iStream(line);
      iStream >> nrec_att;
    } else {
      cout << attname << " not found." << endl;
      exit(1);
    }

    // Get attitude times
    atttme = (double *) calloc(nrec_att, sizeof(double));
    for (size_t j=0; j<nrec_att; j++) {
      getline( attfile, line);
      get_tme_z( &line, &atttme[j], NULL);
      //    printf("%d %s %f\n", j, line.c_str(), atttme[j]);
    }

    // Stage HKT file
    hktfile.open(argv[5], fstream::in | fstream::binary);
    hktfile.seekg (0, ios::end);
    int32_t totalbytes = hktfile.tellg();
    hktfile.seekg (0, ios::beg);
    nrec_hkt = totalbytes / 4000;
  } // if ( argc >= 6)

  uint32_t gps_hkt;
  cptr = (char *) &gps_hkt;
  
  hkttme = (uint32_t *) calloc(nrec_hkt, sizeof(uint32_t));
  for (size_t i=0; i<nrec_hkt; i++) {
    hktfile.read( (char *) hktrec, 4000);
    //t = byte([obt[2], obt[3], obt[0], obt[1]])
    memcpy( cptr+0, &hktrec[580+816+2], 1);
    memcpy( cptr+1, &hktrec[580+816+3], 1);
    memcpy( cptr+2, &hktrec[580+816+0], 1);
    memcpy( cptr+3, &hktrec[580+816+1], 1);
    hkttme[i] = gps_hkt;
  }

  // Read TAI & nanoSec values
  cout << endl << "nrecL0: " << nrecL0 << endl << endl;
  uint32_t *tai = (uint32_t *) calloc(nrecL0, sizeof(int32_t));
  uint32_t *nsc = (uint32_t *) calloc(nrecL0, sizeof(int32_t));
  ptrL0 = infile.getNext(4);  // offset by magic word length
  for (size_t i=0; i<nrecL0; i++) {
    ptrL0 = infile.getNext(sizeof(int32_t), (char *) &tai0);
    ptrL0 = infile.getNext(sizeof(int32_t), (char *) &nsc0);
    ptrL0 = infile.getNext(recsize-8);

    tai[i] = SWAP_4(tai0) - 1; // switch to previous GPS sec
    nsc[i] = SWAP_4(nsc0);
  }
  ptrL0 =  infile.rewind(magic);

  // Find first relevant record in eph file
  double tme, tme_bfr=0.0, tme_aft=0.0;

  size_t ephrecnum = 0;
  size_t prev_ephrecnum = 0;
  size_t first_rel_eph;

  // Display orbit times
  if ( nrec_eph > 0) {
    // Note: Both tai and ephtme time arrays are in seconds from 01/06/1980
    while (1) {
      tme = ephtme[ephrecnum];

      if (tai[0] <= tme) {
	first_rel_eph = ephrecnum-1;
	if (first_rel_eph < 0) {
	  cout << "First ephemeris record after first L0 record" << endl;
	  exit(1);
	}
	break;
      }
      ephrecnum++;
    }

    // Determine next open file time (N minutes before orbit start)
    while (1) {
      tme_bfr = ephtme[ephrecnum];

      if (orb[ephrecnum] == '<') {
	ydhmsf_str = ydhmsf( gpstai2unix( tme_bfr), 'G');
	prev_ephrecnum = ephrecnum;
	break;
      }
      ephrecnum++;
    }

    // Determine next close file time (N minutes after orbit start)
    while (1) {
      tme_aft = ephtme[ephrecnum];

      if (orb[ephrecnum] == '>') {
	ydhmsf_str = ydhmsf( gpstai2unix( tme_aft), 'G');
	break;
      }
      ephrecnum++;
    }
  }

  uint32_t nopen = 0; 
  static Hdf::hdf5_Aquarius outfile[2];
  ofstream fout[2];

  int32_t openid = 0;
  // Create new orbit file (L1A)

  ostr.str("");
  string str;
  str.assign(basename(argv[1]));
  if ( nrec_eph > 0) str = "/Q" + str.substr(0, 15); else str = "/" + str;


  time_t rawtime = (time_t) gpstai2unix( tai[0]);
  tm *ptm = gmtime ( &rawtime );
  strftime (timebuf, 80, "%Y%m%d_%H%M%S", ptm);

  ostr << argv[2] << str.c_str() << "_T" << timebuf << outfile_ext;
  if ( nrec_eph == 0) ostr << "_InT";

  // If no eph then set # L1 recs to # L0recs
  int32_t maxL1Arecs = MAXSCAN;
  if ( nrec_eph == 0) maxL1Arecs = 
    (int32_t) ((tai[nrecL0-1] - tai[0]) / 1.44) + 100;

  uint32_t icds_blknum[2];
  uint32_t icds_blknum0;
  memcpy(&icds_blknum[openid], &ptrL0[ICDS_STS_OFFSET+8], 4);
  icds_blknum[openid] = SWAP_4(icds_blknum[openid]);


  // Note: Granule Stop Time set in writel1()
  string softwareId = VERSION;
  outfile[openid].createl1((char *) ostr.rdbuf()->str().c_str(), 
			   maxL1Arecs, maxL1Arecs/5, 
			   (double) (tai[0] + 1 - ((nsc[0]-2.0) * 62.5e-9)),
			   basename( argv[1]), 
			   (char *) processControl.c_str(),
			   (char *) softwareId.c_str());

  cout << "create partial orbit file:  "
       << (char *) ostr.rdbuf()->str().c_str();
  cout << " " << outfile[openid].getH5fid() << " " << openid << endl << endl;

  cout << "----------------------------------------" << endl;
  ydhmsf_str = ydhmsf( gpstai2unix( tme_bfr), 'G');
  cout << "<: " << ydhmsf_str.c_str() << " " 
       << (uint32_t) tme_bfr << " " << ephrecnum << endl;

  // Open temp L1A text file
  ostr << ".txt";
  fout[openid].open(ostr.rdbuf()->str().c_str());
  fout[openid] << "Start Time: " << timebuf << endl;

  tai_start[n_files_open++] = tai[0];
  nopen++;

  /////////////////////////////////////////////////////
  //////// Loop through L0 records (MAIN LOOP) ////////
  /////////////////////////////////////////////////////
  for (size_t irec=0; irec<nrecL0; irec++) {

    // Note: tai is integer
    //    ydhmsf_str = ydhmsf( gpstai2unix( tai[irec]), 'G');
    //cout << irec << " " << ydhmsf_str.c_str() << endl;

    if ( irec == 0) goto write_packet;

    // Necessary if attitude file ends before L0 file
    if ( nrec_att != 0)
      if ( tai[irec] > atttme[nrec_att-1]) continue;

    if ( nrec_eph == 0 && (irec % 5000) == 0) 
      cout << "irec: " << irec << " out of " << nrecL0 << endl;

    // Create new orbit file (L1A)
    if (tai[irec-1] < tme_bfr && tai[irec] >= tme_bfr) {

      time_t rawtime = (time_t) gpstai2unix( tai[irec]);
      tm *ptm = gmtime ( &rawtime );
      strftime (timebuf, 80, "%Y%m%d_%H%M%S", ptm);

      ostr.str("");
      ostr << argv[2] << str.c_str() << "_T" << 
	timebuf << outfile_ext;

      memcpy(&icds_blknum[1-openid], &ptrL0[ICDS_STS_OFFSET+8], 4);
      icds_blknum[1-openid] = SWAP_4(icds_blknum[1-openid]);

      // Note: Granule Stop Time set in writel1()
      outfile[1-openid].createl1((char *) ostr.rdbuf()->str().c_str(), 
				 MAXSCAN, MAXSCAN/5, 
				 (double) (tai[irec] + 1 - 
					   ((nsc[irec]-2.0) * 62.5e-9)),
				 basename( argv[0]), 
				 (char *) processControl.c_str(),
				 (char *) softwareId.c_str());

      cout << "create file:  " << (char *) ostr.rdbuf()->str().c_str();
      cout << " " << outfile[1-openid].getH5fid() << " " << 1-openid;
      cout << setw(7) << irec << endl << endl;


      ostr << ".txt";
      fout[1-openid].open(ostr.rdbuf()->str().c_str());
      fout[1-openid] << "Start Time: " << timebuf << endl;

      tai_start[n_files_open++] = tai[irec];
      openid = 1 - openid;
      nopen++;
    } // endif

    if (tai[irec-1] < tme_aft && tai[irec] >= tme_aft) {

      outfile[1-openid].getGranuleTimes( &granStart, &granStop);

      H5Fget_name( outfile[1-openid].getH5fid(), namebuf, sizeof(namebuf));
      //      cout << "namebuf: " << namebuf << endl;
      ostr.str("");
      ostr << "mv " << namebuf << " ";
      cptr = strstr( namebuf, outfile_ext); 
      *cptr = 0;

      time_t rawtime = (time_t) gpstai2unix( granStop);
      tm *ptm = gmtime ( &rawtime );
      strftime (timebuf, 80, "%Y%m%d_%H%M%S", ptm);

      ostr << namebuf << "_" << timebuf << outfile_ext;
      // cout << ostr.rdbuf()->str().c_str() << endl;
      system_retstat = system( ostr.rdbuf()->str().c_str());
      if ( system_retstat != 0) {
	cout << "system call: " << ostr.rdbuf()->str().c_str()
	     << " returned error." << endl;
	exit(1);
      }

      ydhmsf_str = ydhmsf( gpstai2unix( tme_aft), 'G');
      cout << ">: " << ydhmsf_str.c_str() << " " << (uint32_t) tme_aft;
      cout << " " << ephrecnum << " " << setw(7) << irec << endl;
      cout << "close file:   " << namebuf << "_" << 
	timebuf << outfile_ext 
	   << " (" << (granStop - granStart) / 60 << ")" << endl;
      cout << "----------------------------------------" << endl;

      outfile[1-openid].closel1( granStop);
      tai_stop[n_files_close++] = (uint32_t) granStop;
      nopen--;

      // Close & Rename temp L1A text file
      fout[1-openid] << "Stop Time:  " << timebuf << endl;
      fout[1-openid] << "Number of Records: " << 
	outfile[1-openid].nBlks() << endl;
      fout[1-openid].close();

      *cptr = '.';
      ostr.str("");
      ostr << "mv " << namebuf << ".txt ";
      cptr = strstr( namebuf, outfile_ext); 
      *cptr = 0;
      ostr << namebuf << "_" << timebuf << outfile_ext << ".txt";
      // cout << ostr.rdbuf()->str().c_str() << endl;
      system_retstat = system( ostr.rdbuf()->str().c_str());
      if ( system_retstat != 0) {
	cout << "system call: " << ostr.rdbuf()->str().c_str()
	     << " returned error." << endl;
	exit(1);
      }

      // Find next "N minutes before orbit start"
      while (1) {
	if ( ephrecnum == nrec_eph) break;
 
	tme_bfr = ephtme[ephrecnum];

	if (orb[ephrecnum] == '<' && tai[irec] < tme_bfr) {
	  ydhmsf_str = ydhmsf( gpstai2unix( tme_bfr), 'G');
	  cout << endl << "----------------------------------------" << endl;
	  cout << "<: " << ydhmsf_str.c_str() 
	       << " " << (uint32_t) tme_bfr;
	  cout << " " << ephrecnum << " " << setw(7) << irec << endl;
	  prev_ephrecnum = ephrecnum;
	  break;
	}
	ephrecnum++;
      }

      // Find next "N minutes after orbit start"
      while (1) {
	if ( ephrecnum == nrec_eph) {
	  cout << "Insufficient ephermeris records to complete processing." 
	       << endl;
	  ydhmsf_str = ydhmsf( gpstai2unix( tai[irec]), 'G');
	  cout << ydhmsf_str.c_str() << endl;
	  exit(110);
	}

	tme_aft = ephtme[ephrecnum];

	if (orb[ephrecnum] == '>') {
	  ydhmsf_str = ydhmsf( gpstai2unix( tme_aft), 'G');
	  //	  cout << ">: " << ydhmsf_str.c_str() << " " << (uint32_t) tme_aft;
	  //cout << " " << ephrecnum << " " << setw(7) << irec << endl;
	  break;
	}
	ephrecnum++;
      }
    } // endif

  write_packet:
    // Write record
    memcpy(&tai0, &ptrL0[4], 4);
    memcpy(&nsc0, &ptrL0[8], 4);

    if (SWAP_4(tai0) != (tai[irec] + 1) || SWAP_4(nsc0) != nsc[irec]) {
      cout << "L0 record misalignment" << endl;
      cout << SWAP_4(tai0) << " " << tai[irec-1] << endl;
      cout << SWAP_4(nsc0) << " " << nsc[irec-1] << endl;
      exit(1);
    }

    if (nopen == 1) {
      memcpy(&icds_blknum0, &ptrL0[ICDS_STS_OFFSET+8], 4);
      icds_blknum0 = SWAP_4(icds_blknum0);
      if ( (icds_blknum0 - icds_blknum[openid]) >= (maxL1Arecs-3)) {
	H5Fget_name( outfile[openid].getH5fid(), namebuf, sizeof(namebuf)); 
	cout << "Warning: Approaching allocated file size at ICDS block: " << 
	  icds_blknum0 << endl;
	cout << namebuf << endl;
      } 
      outfile[openid].writel1(ptrL0);
    } else {
      for (size_t i=1; i<=nopen; i++) {
	memcpy(&icds_blknum0, &ptrL0[ICDS_STS_OFFSET+8], 4);
	icds_blknum0 = SWAP_4(icds_blknum0);
	if ( (icds_blknum0 - icds_blknum[i-1]) >= (maxL1Arecs-3)) {
	  cout << "Warning: Approaching allocated file size at ICDS block: " << 
	  icds_blknum0 << endl;
	  cout << namebuf << endl;
	} 
	outfile[i-1].writel1(ptrL0); 
      }
    }

    // Get next L0 blk
    ptrL0 = infile.getNext(recsize);

    // Check for jump over orbit boundary
    if ( irec < (nrecL0-1)) {
      double diff = tai[irec+1] - tai[irec];
      if ( (tai[irec] < (tme_bfr + WINGLEN*60)) &&
	   (tai[irec+1]  > (tme_bfr + WINGLEN*60)) &&
	   (diff > 2)) {

	ydhmsf_str = ydhmsf( gpstai2unix( tai[irec]), 'G');
	cout << "Gap of " << diff << " seconds between: " 
	     << ydhmsf_str.c_str() << " and ";

	ydhmsf_str = ydhmsf( gpstai2unix( tai[irec+1]), 'G');
	cout << ydhmsf_str.c_str() << endl;
      }
    }


  } // L0 record loop (irec)


  // Close remaining open files
  if (n_files_open-n_files_close == 1) {

    outfile[openid].getGranuleTimes( &granStart, &granStop);

    H5Fget_name( outfile[openid].getH5fid(), namebuf, sizeof(namebuf));
    ostr.str("");
    ostr << "mv " << namebuf << " ";
    cptr = strstr( namebuf, outfile_ext); 
    *cptr = 0;

    time_t rawtime = (time_t) gpstai2unix( granStop);
    tm *ptm = gmtime ( &rawtime );
    strftime (timebuf, 80, "%Y%m%d_%H%M%S", ptm);

    ostr << namebuf << "_" << timebuf << outfile_ext;
    if ( nrec_eph == 0) ostr << "_InT";
    // cout << ostr.rdbuf()->str().c_str() << endl;
    system_retstat = system( ostr.rdbuf()->str().c_str());
    if ( system_retstat != 0) {
      cout << "system call: " << ostr.rdbuf()->str().c_str()
	   << " returned error." << endl;
      exit(1);
    }

    cout << "close partial orbit file:   " << namebuf << "_";
    if ( nrec_eph == 0) {
      cout << timebuf << ".AQ_L1A_InT" << endl;
    } else {
      cout << timebuf << outfile_ext
	   << " (" << (granStop - granStart) / 60 << ")" << endl;
    }

    outfile[openid].closel1( granStop);
    tai_stop[n_files_close++] = (uint32_t) granStop;

    // Close & Rename temp L1A text file
    fout[openid] << "Stop Time:  " << timebuf << endl;
    fout[openid] << "Number of Records: " << 
      outfile[openid].nBlks() << endl;
    fout[openid].close();

    *cptr = '.';
    ostr.str("");
    ostr << "mv " << namebuf << ".txt ";
    cptr = strstr( namebuf, outfile_ext); 
    *cptr = 0;

    if ( nrec_eph == 0) 
      ostr << namebuf << "_" << timebuf << ".AQ_L1A_InT.txt";
    else
      ostr << namebuf << "_" << timebuf << outfile_ext << ".txt";

    system_retstat = system( ostr.rdbuf()->str().c_str());
    if ( system_retstat != 0) {
      cout << "system call: " << ostr.rdbuf()->str().c_str()
	   << " returned error." << endl;
      exit(1);
    }


  } else if (n_files_open-n_files_close == 2) {

    outfile[1-openid].getGranuleTimes( &granStart, &granStop);

    H5Fget_name( outfile[1-openid].getH5fid(), namebuf, sizeof(namebuf));
    ostr.str("");
    ostr << "mv " << namebuf << " ";
    cptr = strstr( namebuf, outfile_ext); 
    *cptr = 0;

    time_t rawtime = (time_t) gpstai2unix( granStop);
    tm *ptm = gmtime ( &rawtime );
    strftime (timebuf, 80, "%Y%m%d_%H%M%S", ptm);

    ostr << namebuf << "_" << timebuf << outfile_ext;
    //    // cout << ostr.rdbuf()->str().c_str() << endl;
    system_retstat = system( ostr.rdbuf()->str().c_str());
    if ( system_retstat != 0) {
      cout << "system call: " << ostr.rdbuf()->str().c_str()
	   << " returned error.";
      exit(1);
    }

    cout << "close partial orbit file:   " << namebuf << "_" << 
      timebuf << outfile_ext
	 << " (" << (granStop - granStart) / 60 << ")" << endl;
    cout << endl;


    outfile[1-openid].closel1( granStop);
    tai_stop[n_files_close++] = (uint32_t) granStop;

    // Close & Rename temp L1A text file
    fout[1-openid] << "Stop Time:  " << timebuf << endl;
    fout[1-openid] << "Number of Records: " << 
      outfile[1-openid].nBlks() << endl;
    fout[1-openid].close();

    *cptr = '.';
    ostr.str("");
    ostr << "mv " << namebuf << ".txt ";
    cptr = strstr( namebuf, outfile_ext); 
    *cptr = 0;
    ostr << namebuf << "_" << timebuf << outfile_ext << ".txt";
    system_retstat = system( ostr.rdbuf()->str().c_str());
    if ( system_retstat != 0) {
      cout << "system call: " << ostr.rdbuf()->str().c_str()
	   << " returned error.";
      exit(1);
    }

    outfile[openid].getGranuleTimes( &granStart, &granStop);

    H5Fget_name( outfile[openid].getH5fid(), namebuf, sizeof(namebuf));
    ostr.str("");
    ostr << "mv " << namebuf << " ";
    cptr = strstr( namebuf, outfile_ext); 
    *cptr = 0;

    rawtime = (time_t) gpstai2unix( granStop);
    ptm = gmtime ( &rawtime );
    strftime (timebuf, 80, "%Y%m%d_%H%M%S", ptm);

    ostr << namebuf << "_" << timebuf << outfile_ext;
    // cout << ostr.rdbuf()->str().c_str() << endl;
    system_retstat = system( ostr.rdbuf()->str().c_str());
    if ( system_retstat != 0) {
      cout << "system call: " << ostr.rdbuf()->str().c_str()
	   << " returned error.";
      exit(1);
    }

    cout << "close partial orbit file:   " << namebuf << "_" << 
      timebuf << outfile_ext
	 << " (" << (granStop - granStart) / 60 << ")" << endl;
    cout << endl;

    outfile[openid].closel1( granStop);
    tai_stop[n_files_close++] = (uint32_t) granStop;

    // Close & Rename temp L1A text file
    fout[openid] << "Stop Time:  " << timebuf << endl;
    fout[openid] << "Number of Records: " << 
      outfile[openid].nBlks() << endl;
    fout[openid].close();

    *cptr = '.';
    ostr.str("");
    ostr << "mv " << namebuf << ".txt ";
    cptr = strstr( namebuf, outfile_ext); 
    *cptr = 0;
    ostr << namebuf << "_" << timebuf << outfile_ext << ".txt";
    system_retstat = system( ostr.rdbuf()->str().c_str());
    if ( system_retstat != 0) {
      cout << "system call: " << ostr.rdbuf()->str().c_str()
	   << " returned error.";
      exit(1);
    }

  } else {
    cout << "More that 2 files remain open" << endl;
    exit(1);
  }

  if (n_files_open != n_files_close) {
    cout << "Number of files opened: " << n_files_open;
    cout << "not equal to number of files closed: " << n_files_close << endl;
    exit(1);
  }


  // Write Navigation
  if ( nrec_eph > 0) {
    // Write EPH (orbit)
    cout << endl << "Writing Ephemeris" << endl;
    for (size_t i=0; i<n_files_open; i++) {
      int32_t jstart=-1, jstop=-1;
      for (size_t j=0; j<nrec_eph; j++) {
	tme = ephtme[j];

	if (tai_start[i] <= tme && jstart == -1) {
	  jstart = j-1;
	}
	if (tai_stop[i] <= tme && jstop == -1) {
	  jstop = j;
	  break;
	} 
      }

      if ( jstart == -1 || jstop == -1) {
	cout << "Error determining proper orbital records to write." << endl;
	exit(110);
      }

      // Need 2 recs after tai_stop for scat interp
      if ( jstop != (int32_t) nrec_eph-1) jstop++;

      int32_t numOrbVec = jstop-jstart+1;
      double *orb_time = (double *) calloc(numOrbVec, sizeof(double));
      double *orb_pos = (double *)  calloc(numOrbVec, 3*sizeof(double));
      double *orb_vel = (double *)  calloc(numOrbVec, 3*sizeof(double));

      // Skip to first relevant line
      ephfile.seekg (0, ios::beg);
      getline( ephfile, line); // skip #recs line
      for (int j=0; j<jstart; j++) getline( ephfile, line);

      for (int j=jstart; j<=jstop; j++) {
	getline( ephfile, line);
	
	// Convert ephtme to 1970 base
	// Note: Make sure arg of type "time_t" is used in gmtime()
	// Get hour, min, sec to compute seconds since start of day
	time_t bintime = (time_t) gpstai2unix( ephtme[j]);
	tm0 = gmtime( &bintime);
	double daysec = tm0->tm_hour*3600 + tm0->tm_min*60 + tm0->tm_sec;

	// JMG
	// orb_time 
	orb_time[j-jstart] = daysec;
	//orb_time[j-jstart] = ephtme[j];
	//

	orb_pos[3*(j-jstart)]   = atof(line.substr(28, 15).c_str()) * oneth;
	orb_pos[3*(j-jstart)+1] = atof(line.substr(43, 15).c_str()) * oneth;
	orb_pos[3*(j-jstart)+2] = atof(line.substr(58, 15).c_str()) * oneth;
	
	orb_vel[3*(j-jstart)]   = atof(line.substr(73, 15).c_str()) * oneth;
	orb_vel[3*(j-jstart)+1] = atof(line.substr(88, 15).c_str()) * oneth;
	orb_vel[3*(j-jstart)+2] = atof(line.substr(103, 15).c_str()) * oneth;
      }

      // Open L1A granule
      ostr.str("");

      rawtime = (time_t) gpstai2unix( tai_start[i]);
      ptm = gmtime ( &rawtime );
      strftime (timebuf, 80, "%Y%m%d_%H%M%S", ptm);

      ostr << argv[2] << str.c_str() << "_T" << timebuf;

      rawtime = (time_t) gpstai2unix( tai_stop[i]);
      ptm = gmtime ( &rawtime );
      strftime (timebuf, 80, "%Y%m%d_%H%M%S", ptm);

      ostr << "_" << timebuf << outfile_ext;

      outfile[0].openl1((char *) ostr.rdbuf()->str().c_str(), H5F_ACC_RDWR);
      outfile[0].writel1_eph(numOrbVec, orb_time, orb_pos, orb_vel);
      outfile[0].closel1();

      free(orb_time);
      free(orb_pos);
      free(orb_vel);
    }
    ephfile.close();


    // Write ATT & QUATERNIONS
    cout << "Writing Attitude" << endl;
    for (size_t i=0; i<n_files_open; i++) {
      int32_t jstart=-1, jstop=-1;
      for (size_t j=0; j<nrec_att; j++) {
	tme = atttme[j];

	if (tai_start[i] <= tme && jstart == -1) {
	  jstart = j-1;
	}
	if (tai_stop[i] <= tme && jstop == -1) {
	  jstop = j;
	  break;
	} 
      }
      //      if ( jstop == -1) jstop = atttme[nrec_att-1];

      if ( jstart == -1 || jstop == -1) {
	cout << "Error determining proper attitude records to write." << endl;
	exit(110);
      }

      // Need 2 recs after tai_stop for scat interp
      if ( jstop != (int32_t) nrec_att-1) jstop++;

      int32_t numAttSamp = jstop-jstart+1;
      cout << "Number of Attitude Samples: " << numAttSamp << endl;
      cout << "Tai Start: " << tai_start[i] << endl;
      cout << "Tai Stop:  " << tai_stop[i] << endl;

      double *att_time = (double *) calloc(numAttSamp, sizeof(double));
      double *att_ang = (double *) calloc(numAttSamp, 3*sizeof(double));
      double *quaternion = (double *) calloc(numAttSamp, 4*sizeof(double));
      uint32_t *att_flags = (uint32_t *) calloc(numAttSamp, sizeof(uint32_t));

      // Skip to first relevant line
      attfile.seekg (0, ios::beg);
      getline( attfile, line); // skip #recs line
      for (int j=0; j<jstart; j++) getline( attfile, line);

      for (int j=jstart; j<=jstop; j++) {
	getline( attfile, line);
	get_tme_z( &line, &att_time[j-jstart], NULL);

	att_ang[3*(j-jstart)]   = atof(line.substr(28, 15).c_str());
	att_ang[3*(j-jstart)+1] = atof(line.substr(43, 15).c_str());
	att_ang[3*(j-jstart)+2] = atof(line.substr(58, 15).c_str());

	quaternion[4*(j-jstart)]   = atof(line.substr(73, 15).c_str());
	quaternion[4*(j-jstart)+1] = atof(line.substr(88, 15).c_str());
	quaternion[4*(j-jstart)+2] = atof(line.substr(103, 15).c_str());
	quaternion[4*(j-jstart)+3] = atof(line.substr(118, 15).c_str());

	// Cold Sky Calibration
	if ( fabs(att_ang[3*(j-jstart)+1]) > 5.0) att_flags[j-jstart] += 1;

	// Convert att_time to 1970 base
	// Get hour, min, sec to compute daysec
	time_t bintime = (time_t) gpstai2unix( atttme[j]);
	tm0 = gmtime( &bintime);
	double daysec = tm0->tm_hour*3600 + tm0->tm_min*60 + tm0->tm_sec; 
	att_time[j-jstart] = daysec;
      }

      ostr.str("");

      time_t rawtime = (time_t) gpstai2unix( tai_start[i]);
      tm *ptm = gmtime ( &rawtime );
      strftime (timebuf, 80, "%Y%m%d_%H%M%S", ptm);

      ostr << argv[2] << str.c_str() << "_T" << timebuf;

      rawtime = (time_t) gpstai2unix( tai_stop[i]);
      ptm = gmtime ( &rawtime );
      strftime (timebuf, 80, "%Y%m%d_%H%M%S", ptm);

      ostr << "_" << timebuf << outfile_ext;

      outfile[0].openl1((char *) ostr.rdbuf()->str().c_str(), H5F_ACC_RDWR);
      outfile[0].writel1_att(numAttSamp, att_time, att_ang, quaternion, 
			     att_flags);
      outfile[0].closel1();

      free(att_time);
      free(att_ang);
      free(quaternion);
      free(att_flags);
    }
    attfile.close();
  }

  // Write HKT records
  if ( nrec_hkt > 0) {
    cout << "Writing HKT" << endl;
    for (size_t i=0; i<n_files_open; i++) {
      int32_t jstart=-1, jstop=-1;
      for (size_t j=0; j<nrec_hkt; j++) {
	tme = hkttme[j];

	if (tai_start[i] <= tme && jstart == -1) {
	  jstart = j-1;
	}
	if (tai_stop[i] <= tme && jstop == -1) {
	  jstop = j;
	  break;
	} 
      }

      if ( jstart == -1 || jstop == -1) {
	cout << "Error determining proper HKT records to write." << endl;
	exit(110);
      }

      // Need 2 recs after tai_stop for scat interp
      if ( jstop != (int32_t) nrec_hkt-1) jstop++;

       // Skip to first relevant line
      hktfile.seekg (0, ios::beg);
      for (int j=0; j<jstart; j++) hktfile.read( (char *) hktrec, 4000);


      ostr.str("");

      time_t rawtime = (time_t) gpstai2unix( tai_start[i]);
      tm *ptm = gmtime ( &rawtime );
      strftime (timebuf, 80, "%Y%m%d_%H%M%S", ptm);

      ostr << argv[2] << str.c_str() << "_T" << timebuf;

      rawtime = (time_t) gpstai2unix( tai_stop[i]);
      ptm = gmtime ( &rawtime );
      strftime (timebuf, 80, "%Y%m%d_%H%M%S", ptm);

      if ( nrec_eph == 0)
	ostr << "_" << timebuf << ".AQ_L1A_InT";
      else
	ostr << "_" << timebuf << outfile_ext;

      outfile[0].openl1((char *) ostr.rdbuf()->str().c_str(), H5F_ACC_RDWR);

      uint32_t k=0;
      for (int j=jstart; j<=jstop; j++) {
	hktfile.read( (char *) hktrec, 4000);
	outfile[0].writel1_hkt( k, hktrec);
	k++;
      }
      outfile[0].closel1();
    }
    hktfile.close();
  }

  free(tai);
  free(nsc);
  if ( orb != NULL) free(orb);
  if ( ephtme != NULL) free(ephtme);
  if ( ephz != NULL) free(ephz);
  if ( atttme != NULL) free(atttme);
}


int annotate_eph(ifstream *ephfile, unsigned char *orb, 
		 double *ephtme, double *ephz, uint32_t nrec_eph)
{
  std::list<double> pretme;

  // tme is seconds from 01/06/1980

  double tme_pre=-1;
  double tme_pst=-1;
  double tme0, tme1, tme2;
  double  z0, z1, z2;

  string line;
  string line_prev;

  if ( nrec_eph == 0) return 0;

  getline( *ephfile, line);
  get_tme_z( &line, &tme0, &z0);
  line_prev = line;

  getline( *ephfile, line);
  get_tme_z( &line, &tme1, &z1);
  
  if ((tme1 - tme0) != 60) {
    cout << endl << "Ephemeris records difference: " << tme1 - tme0 << endl;
    cout << line_prev << endl;
    cout << line << endl;
    //    exit(110);
  }

  ephtme[0] = tme0;
  ephtme[1] = tme1;
  ephz[0] = z0;
  ephz[1] = z1;

  line_prev = line;

  // Mark orbit start and N minutes after
  for (size_t i=2; i<nrec_eph-1; i++) {

    getline( *ephfile, line);
    get_tme_z( &line, &tme2, &z2);

    if ((tme2 - tme1) != 60) {
      cout << endl << "Ephemeris records difference: " << tme1 - tme0 << endl;
      cout << line_prev << endl;
      cout << line << endl;
      //    exit(110);
    }

    ephtme[i] = tme2;
    ephz[i] = z2;

    // Mark @ N minutes after orbit start
    if (tme1 < tme_pst && tme2 >= tme_pst) {
      orb[i] = '>';
      //cout << "N min after: " << (int32_t) tme1 << endl;
    }

    // Find beginning of new orbit
    // Mark orbit start
    // Determine N minutes previous and N minutes past
    // Store previous time
    if (z0 > z1 && z1 <= z2) {
      orb[i] = 'O';
      if ( tme_pre != -1) {
	cout << "orbit start (approx): " << setw(5) << i << ": ";
	cout << (int32_t) tme1;
	cout << " " << (tme1 - tme_pre - WINGLEN*60);
	cout << " (" << line.substr(0, 19).c_str() << ")" << endl;
      }
      tme_pre = tme1 - WINGLEN*60;
      tme_pst = tme1 + WINGLEN*60;
      
      pretme.push_back(tme_pre);
    }

    tme0 = tme1;
    tme1 = tme2;
    z0 = z1;
    z1 = z2;
    line_prev = line;
  }
  //  cout << pretme.size() << endl;


  // Mark N minutes before orbit start

  tme0 = ephtme[0];
  tme1 = ephtme[1];

  // Pop to beginning of pretme vector & find 1st tmp_pre after tme1
  tme_pre = pretme.front();
  pretme.pop_front();
  while (tme_pre < tme1) {
    tme_pre = pretme.front();
    pretme.pop_front();
  }

  for (size_t i=2; i<nrec_eph-1; i++) {
    tme2 = ephtme[i];

    // Mark @ N minutes before orbit start
    if (tme1 < tme_pre && tme2 >= tme_pre) {
      orb[i] = '<';
      tme_pre = pretme.front();
      if (pretme.size() == 0) break;
      pretme.pop_front();
      //cout << "N min before: " << (int32_t) tme1 << endl;
      //      cout << pretme.size() << endl;
      tme1 = tme2;
    }
  }

  return 0;
}


int get_tme_z( string *line, double *tme, double *z)
{
  // tme is seconds from 01/06/1980

  if (z != NULL) *z = atof(line->substr(58, 15).c_str());

  /*
  // Uses system call to date  
  tstr = "date -d '" + line->substr(0, 19) + " GMT' +%s";
  if ((fptr = popen(tstr.c_str(), "r")) != NULL)
    fgets(buf, bufSize, fptr);
  pclose(fptr);
  */

  struct tm tt;
  tt.tm_sec = atoi(line->substr(17, 2).c_str());
  tt.tm_min = atoi(line->substr(14, 2).c_str());
  tt.tm_hour = atoi(line->substr(11, 2).c_str());
  tt.tm_mday = atoi(line->substr(8, 2).c_str());
  tt.tm_mon = atoi(line->substr(5, 2).c_str()) - 1;
  tt.tm_year = atoi(line->substr(0, 4).c_str()) - 1900;
  tt.tm_isdst = 0;

  time_t sec = (time_t) unix2gpstai(timegm( &tt) + 
				    atof(line->substr(19, 9).c_str()));
  *tme = sec;

  /*
  if (sec != *tme) {
    printf("%f %f\n", sec, *tme);
    exit(-1);
  }
  */
  return 0;
}

