#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <libgen.h>
#include <iostream>
#include <sstream>
#include <fstream>
#include <iomanip>
#include <list>
#include "hdf5_Aquarius.h"
#include "hdf5utils.h"

#define MAXINPUT 20
 
#define NAMELEN 256

#define BLKCHK 500
#define PI      3.141592653589793

#define VERSION "1.07"

extern "C" void cdata_();
extern "C" void asap_rots_(int *, int *, double *, 
			   double *, int *, double *);

//    Modification history:
//
//   Programmer     Organization      Date      Description of change
//   ----------     ------------    --------    ---------------------
//   Joel Gales       FutureTech      04/02/08    Original development
//   Joel Gales       FutureTech      08/03/09    Use gpstai2unix &
//                                                unix2gpstai functions
//   Joel Gales       FutureTech      09/16/09    Add support for RAD &
//                                                ATC subframes.
//   Joel Gales       FutureTech      09/21/09    Add subframe check
//   Joel Gales       FutureTech      09/22/09    Base output block number 
//                                                on icds block number.
//   Joel Gales       FutureTech      10/27/09    Fix bug so first orb/att 
//                                                before first blk_sec
//                                                Allow 2 orb/att recs after 
//                                                last blk_sec for scat interp
//   Joel Gales       FutureTech      11/06/09    Add output of number of 
//                                                input files
//   Joel Gales       FutureTech      12/23/09    Fix bug with output rad frame
//                                                # for "radiom_" and 
//                                                "dpu_status_tlm"
//   Joel Gales       FutureTech      04/15/10    Check for l1a files 
//                                                (in l1open)
//   Joel Gales       FutureTech      05/02/10    Cleanup inconsistant 
//                                                subframe reporting
//   Joel Gales       FutureTech      05/05/10    Use orbit start time for
//                                                merged filename
//   Joel Gales       FutureTech      06/20/10    Add "Input Files" &
//                                                "Processing Control" metadata
//   Joel Gales       FutureTech      12/11/10    Extend and initialize
//                                                Time, OrbTime, AttTime
//   Joel Gales       FutureTech      02/16/11    Fix node longitude calc
//                                                atan->atan2
//   Joel Gales       FutureTech      02/23/11    Add SAC-D HKT record copy
//   Joel Gales       FutureTech      03/14/11    Check for empty SAC-D HKT 
//                                                field
//   Joel Gales       FutureTech      08/10/11    Update to V 1.00
//   Joel Gales       FutureTech      09/09/11    Setup Cycle/Pass number
//                                                to start Aug 25
//   Joel Gales       FutureTech      09/28/11    Fix orbit number bug
//                                                (change in closel1())
//   Joel Gales       FutureTech      10/28/11    Pass version number to
//                                                createl1
//   Joel Gales       FutureTech      11/04/11    Add Missing Blocks attr
//   Joel Gales       FutureTech      07/05/12    Fix leap second bug
//                                                (calls to readl1_eph() &
//                                                 readl1_att() in l1_io.cpp)
//   Joel Gales       FutureTech      08/22/12    Fix orbit number bug
//                                                (Put intermediate orbit
//                                                 values in closel1())
//   Joel Gales       FutureTech      01/26/15    If temp L1A file starts
//                                                after orbit start then
//                                                skip first frame to avoid
//                                                possible empty telemetry
//                                                subframes
//   Joel Gales       FutureTech      06/04/15    Continue if block pointer 
//                                                is at EOF

using namespace std;

int main (int argc, char* argv[])
{
 cout << "l1amerge_aquarius " << VERSION << " (" 
       <<  __DATE__ << " " << __TIME__ << ")" << endl;

 if ( argc == 1) {
   cout << endl << "l1amerge_aquarius input_filelist output_directory" 
	<< endl;
   return 0;
 }

  static Hdf::hdf5_Aquarius l1afile[MAXINPUT];
  static Hdf::hdf5_Aquarius l1aoutfile;

  string processControl;
  processControl = argv[0];
  processControl += " ";
  processControl += argv[1];
  processControl += " ";
  processControl += argv[2];
  char namebuf[NAMELEN];

  // Read input file list & open files
  ifstream inFile;  
  inFile.open(argv[1]);
  if (!inFile) {
    cout << "Unable to open filelist: " << argv[1];
    exit(1); // terminate with error
  }

  // Convert ZULU orbit start/stop times to seconds from 01/06/1980
  double orbitStart, orbitStop;
  char timeString[64], orbitStartString[64];
  inFile >> timeString;
  strcpy( orbitStartString, timeString);
  orbitStart = get_tai( timeString);
  inFile >> timeString;
  orbitStop = get_tai( timeString);

  size_t nInput = 0;
  string inputFiles = "";
  while (inFile >> namebuf) {
    cout << "Open: " << namebuf << endl; 
    l1afile[nInput].openl1(namebuf, H5F_ACC_RDONLY);
    nInput++;
    inputFiles += basename(namebuf);
    inputFiles += ",";
  }
  inFile.close();
  cout << "Number of input files: " << nInput << endl << endl;
  inputFiles = inputFiles.substr( 0, inputFiles.length()-1);

  if (nInput > MAXINPUT) {
    cout << " No more than " << MAXINPUT << " input files allowed." << endl;
    exit(110);
  }

  // Read block times (seconds from 01/06/1980) & radiometer header
  double *Time[MAXINPUT];
  int32_t *atc_frmnum[MAXINPUT];
  uint8_t *atc_subframe[MAXINPUT];
  int32_t *rad_frmnum[MAXINPUT];
  uint8_t *rad_subframe[MAXINPUT];
  int32_t *icds_blknum[MAXINPUT];
  uint16_t *rad_header[MAXINPUT];

  for (size_t i=0; i<nInput; i++) {
    Time[i] = new double[l1afile[i].nBlks()+1];
    Time[i][l1afile[i].nBlks()] = 0;

    atc_frmnum[i] = new int32_t[l1afile[i].nBlks()];
    atc_subframe[i] = new uint8_t[l1afile[i].nBlks()];
    rad_frmnum[i] = new int32_t[l1afile[i].nBlks()];
    rad_subframe[i] = new uint8_t[l1afile[i].nBlks()];
    icds_blknum[i] = new int32_t[l1afile[i].nBlks()];
    rad_header[i] = new uint16_t[l1afile[i].nBlks()];
    for (size_t j=0; j<l1afile[i].nBlks(); j++) {
      l1afile[i].readl1_frame_info( j, 
				    &atc_frmnum[i][j], &atc_subframe[i][j],
				    &rad_frmnum[i][j], &rad_subframe[i][j]);

      l1afile[i].readl1_radiometer( j, rad_frmnum[i][j], rad_subframe[i][j],
				    &Time[i][j], &rad_header[i][j], 
				    NULL, NULL, NULL);
      //      printf("%d %f15.2 %d\n", j, Time[i][j], rad_subframe[i][j]);

      l1afile[i].readl1_icds_blknum( j, &icds_blknum[i][j]);
    }
  }

  // Initialize blkPtr (skip to first record at beginning of granule)
  int32_t blkPtr[MAXINPUT];
  for (size_t i=0; i<nInput; i++) {
    blkPtr[i] = -1;
    for (size_t j=0; j<l1afile[i].nBlks(); j++) {
      if (Time[i][j] > (orbitStart-10*60)) {
        blkPtr[i] = j;
	break;
      }
    }
    if ( blkPtr[i] == -1) {
      cout << "blkPtr not initialized for file number: ";
      cout << i << endl;
      cout << Time[i][l1afile[i].nBlks()-1] << " " << orbitStart-10*60 << endl;
 
      // Close input L1A files
      for (size_t i=0; i<nInput; i++) l1afile[i].closel1();

      // Delete dynamic arrays
      for (size_t i=0; i<nInput; i++) {
	delete[] Time[i];
	delete[] atc_frmnum[i];
	delete[] atc_subframe[i];
	delete[] rad_frmnum[i];
	delete[] rad_subframe[i];
	delete[] icds_blknum[i];
	delete[] rad_header[i];
      }
      return 1;
    }
  }

  // Check if any relevant records
  bool noRecs = true;
  for (size_t i=0; i<nInput; i++) {
    if (blkPtr[i] != -1) {
      noRecs = false;
      break;
    }
  }

  // Exit if no relevant records
  if (noRecs == true) {
    cout << "No relevant records found in input files." << endl;

    // Close input L1A files
    for (size_t i=0; i<nInput; i++) l1afile[i].closel1();

    // Delete dynamic arrays
    for (size_t i=0; i<nInput; i++) {
      delete[] Time[i];
      delete[] atc_frmnum[i];
      delete[] atc_subframe[i];
      delete[] rad_frmnum[i];
      delete[] rad_subframe[i];
      delete[] icds_blknum[i];
      delete[] rad_header[i];
    }
    return 1;
  }

  // for (size_t i=0; i<nInput; i++) cout << (int32_t) Time[i][blkPtr[i]] << endl;

  // Determine minimum granule time
  double minTime = 1e30;
  for (size_t i=0; i<nInput; i++) {
    if (blkPtr[i] != -1)
      if (minTime > Time[i][blkPtr[i]]) minTime = Time[i][blkPtr[i]];
  }

  // Create merged L1A file
  time_t tme = (time_t) gpstai2unix( orbitStart);
  struct tm *tm0;
  tm0 = gmtime(&tme);
  //  cout << asctime(tm0) << endl;

  // Construct name string (Orbit Start Time)
  std::ostringstream ostr;
  ostr.fill ('0');

  // Note: tm_yday is days since Jan 1st
  //ostr << argv[2] << "/" << 1900+tm0->tm_year << "/" << tm0->tm_yday+1 << "/";
  //ostr << argv[2] << "/";
  //ostr << "Q" << 1900 + tm0->tm_year;
  //ostr << setw(3) << tm0->tm_yday + 1;
  //ostr << setw(2) << tm0->tm_hour;
  //ostr << setw(2) << tm0->tm_min;
  //ostr << setw(2) << tm0->tm_sec;

  ostr << argv[2] << "/";
  ostr << "Q";
  ostr << orbitStartString;
  ostr << ".L1A_SCI";
  cout << "Merged Filename: " << ostr.rdbuf()->str().c_str() << endl;

  string outfilename = ostr.rdbuf()->str(); 
  double granuleStart = minTime;
  string softwareId = VERSION;
  l1aoutfile.createl1((char *) outfilename.c_str(), MAXSCAN, MAXSCAN/5,
		      granuleStart,
		      (char *) inputFiles.c_str(), 
		      (char *) processControl.c_str(),
		      (char *) softwareId.c_str());

  l1aoutfile.createl1_eph( (int32_t) (MAXSCAN*1.44/60) + 1);
  l1aoutfile.createl1_att( (int32_t) (MAXSCAN*1.44/8) + 1);

  bool write = true;

  hid_t grp_in[MAXINPUT][6], grp_out[6], grp[2];
  for (size_t i=0; i<nInput; i++) {
    l1afile[i].getH5gid(l1afile[i].getH5fid(), grp_in[i]);
  }
  l1aoutfile.getH5gid(l1aoutfile.getH5fid(), grp_out);

  // Write Orbit Start/Stop Times in merged L1A file
  l1aoutfile.writeOrbitTimes( grp_out[0], orbitStart, orbitStop);


  hsize_t numMembs[6];
  for (size_t i=0; i<6; i++) {
    H5Gget_num_objs(grp_in[0][i], &numMembs[i]);
    //    cout << i << " " << numMembs[i] << endl;
  }


  // Check subframe consistancy
  uint8_t subframeATC[MAXINPUT];
  uint8_t subframeRAD[MAXINPUT];
  double time_min[MAXINPUT];
  while (1) {
    for (size_t i=0; i<nInput; i++) {

      subframeATC[i] = 255;
      subframeRAD[i] = 255;

      // Skip out-of-range granules
      if (blkPtr[i] == -1)
	continue;

      // If blank block in frame then increment block pointer and continue
      if (Time[i][blkPtr[i]] == -1) {
	blkPtr[i]++;
	continue;
      }

      if (Time[i][blkPtr[i]] == minTime) {
	subframeATC[i] = atc_subframe[i][blkPtr[i]];
	subframeRAD[i] = rad_subframe[i][blkPtr[i]];
	time_min[i] = Time[i][blkPtr[i]];
	blkPtr[i]++;
      } // if Time
    } // input file loop

    for (size_t i=1; i<nInput; i++) {
      if ( subframeATC[i-1] != subframeATC[i] && 
	   subframeATC[i-1] != 255 && subframeATC[i]!= 255) {
	cout << "*** Inconsistant ATC subframes" << setprecision(2) << endl;
	cout << fixed << time_min[i-1] << endl;
	cout << i-1 << " " << blkPtr[i-1] << endl; 
	cout << i   << " " << blkPtr[i]   << endl;
	cout << "subframes: " << (long) subframeATC[i-1] << " " << 
	  (long) subframeATC[i] << endl;

	exit(1);
      }
      if ( subframeRAD[i-1] != subframeRAD[i] && 
	   subframeRAD[i-1] != 255 && subframeRAD[i]!= 255) {
	cout << "*** Inconsistant RAD subframes" << setprecision(2) << endl;
	cout << fixed << time_min[i-1] << endl;
	cout << i-1 << " " << blkPtr[i-1] << endl;
	cout << i   << " " << blkPtr[i]   << endl;
	cout << "subframes: " << (long) subframeRAD[i-1] << " " << 
	  (long) subframeRAD[i] << endl;

	exit(1);
      }
    } // input file loop

    minTime = 1e30;
    for (size_t i=0; i<nInput; i++) {
      if (blkPtr[i] == -1) continue;
      if (blkPtr[i] < (int32_t) l1afile[i].nBlks()) {
	if (minTime > Time[i][blkPtr[i]]) minTime = Time[i][blkPtr[i]];
      }
    } // input file loop

    // Check if EOF for all input files
    uint32_t eoF=0;
    for (size_t i=0; i<nInput; i++) {
      if (blkPtr[i] == -1) continue;
      if ((blkPtr[i] == (int32_t) l1afile[i].nBlks()) || 
	  (Time[i][blkPtr[i]] >= (orbitStop+10*60))) eoF++;
    }
    if (eoF == nInput) break;
  } // while


  // Initialize blkPtr (skip to first record at beginning of granule)
  for (size_t i=0; i<nInput; i++) {
    blkPtr[i] = -1;
    for (size_t j=0; j<l1afile[i].nBlks(); j++) {
      if (Time[i][j] > (orbitStart-10*60) &&
          Time[i][j] < orbitStart) {
        blkPtr[i] = j;
	break;
      } else if (Time[i][j] > orbitStart) {
        // if temp granule starts after orbit start then skip first frame
        blkPtr[i] = 4;
        break;
      }

    }
  }

  for (size_t i=0; i<nInput; i++) cout << i << " " << blkPtr[i] << endl;

  // Determine minimum granule time
  minTime = 1e30;
  for (size_t i=0; i<nInput; i++) {
    if (blkPtr[i] != -1)
      if (minTime > Time[i][blkPtr[i]]) minTime = Time[i][blkPtr[i]];
  }


  // Main Loop
  bool first_write = true;
  while (1) {
    for (size_t i=0; i<nInput; i++) {

      // Skip out-of-range granules
      if (blkPtr[i] == -1)
	continue;

      // If blank block in frame then increment block pointer and continue
      if (Time[i][blkPtr[i]] == -1) {
	blkPtr[i]++;
	continue;
      }

      // Continue if block pointer is at EOF
      if (blkPtr[i] >= l1afile[i].nBlks()) continue;

      if (Time[i][blkPtr[i]] == minTime) {
	if (write) {	
	  write = false;

	  ///////////////////////////////////////////////////////////
	  uint32_t prev_icds_blknum;
	  uint16_t prev_rad_header;
	  uint8_t offset_corr;
	  uint32_t first_icds_blknum;
	  uint8_t first_atc_subframe;

	  if ( first_write) {
	    first_icds_blknum = icds_blknum[i][blkPtr[i]];
	    first_atc_subframe = atc_subframe[i][blkPtr[i]];

	    prev_icds_blknum = icds_blknum[i][blkPtr[i]] - 1;
	    prev_rad_header = rad_header[i][blkPtr[i]];

	    offset_corr = rad_subframe[i][blkPtr[i]];
	  }

	  if ( (icds_blknum[i][blkPtr[i]] - prev_icds_blknum) > 1)
	    cout << "Jump in icds_blknum from " << prev_icds_blknum << " to "
		 << icds_blknum[i][blkPtr[i]] << " file # (0-based) " 
		 << i << endl;

	  int32_t out_blknum = icds_blknum[i][blkPtr[i]] - first_icds_blknum;

	  // Compute radiometer frame number
	  if (prev_rad_header == 0 && rad_header[i][blkPtr[i]] != 0) 
	    offset_corr = out_blknum % 4;
	  int32_t out_rad_frmnum = 
	    (out_blknum + offset_corr - rad_subframe[i][blkPtr[i]]) / 4;

	  // Compute ATC frame number
	  int32_t out_atc_frmnum = (out_blknum + first_atc_subframe) / 4;

	  // Set frame numbers variables for output file
	  l1aoutfile.setFrameNums( out_rad_frmnum, out_atc_frmnum);
	  ///////////////////////////////////////////////////////////

	  // Loop through all groups
	  for (size_t jgrp=1; jgrp<6; jgrp++) {
	    // grp0 = "/"
	    // grp1 = "/Block Attributes"
	    // grp2 = "/Raw Aquarius Data"
	    // grp3 = "/SAC-D Telemetry"
	    // grp4 = "/Navigation"
	    // grp5 = "/Converted Telemetry"

	    // Skip if SAC-D TELEMETRY group
	    if (jgrp == 3) continue;

	    // Skip if NAVIGATION group
	    if (jgrp == 4) continue;

	    // Loop through all group members
	    for (size_t k=0; k<numMembs[jgrp]; k++) {
	      H5Gget_objname_by_idx(grp_in[i][jgrp], k, namebuf, NAMELEN);
	      H5G_obj_t objType = H5Gget_objtype_by_idx(grp_in[i][jgrp], k);
	      if (objType == H5G_DATASET) {
		grp[0] = grp_in[i][jgrp];
		grp[1] = grp_out[jgrp];

		if (strcmp(namebuf, "atc_frmnum") == 0 ||
		    strcmp(namebuf, "atc_subframe") == 0 ||
		    strcmp(namebuf, "rad_frmnum") == 0 ||
		    strcmp(namebuf, "rad_subframe") == 0)
		  continue;

		// Copy record
		if (strncmp(namebuf, "radiom_", 7) == 0 ||
		    strcmp(namebuf, "dpu_status_tlm") == 0) {
		  l1afile[i].copyl1(rad_frmnum[i][blkPtr[i]], 
				    rad_subframe[i][blkPtr[i]], 
				    out_rad_frmnum,
				    grp, namebuf);
		} else if (strncmp(namebuf, "rfe", 3) == 0 ||
			   strncmp(namebuf, "rbe", 3) == 0 ||
			   strcmp(namebuf, "dpu_analog_tlm") == 0) {
		  l1afile[i].copyl1(rad_frmnum[i][blkPtr[i]], -1,
				    out_rad_frmnum,
				    grp, namebuf);
		} else if (strncmp(namebuf, "atc_omt", 7) == 0 ||
			   strncmp(namebuf, "atc_rbe", 7) == 0) {
		  l1afile[i].copyl1(atc_frmnum[i][blkPtr[i]], -1,
				    out_atc_frmnum,
				    grp, namebuf);
		} else {
		  l1afile[i].copyl1(blkPtr[i], -1, l1aoutfile.nBlks(), 
				    grp, namebuf);
		}

	      } // if dataset
	    } // group member loop
	  } // grp loop

	  l1aoutfile.writel1_frame_info( out_blknum, 
					 &out_atc_frmnum, 
					 &atc_subframe[i][blkPtr[i]],
					 &out_rad_frmnum, 
					 &rad_subframe[i][blkPtr[i]]);

	  if ((l1aoutfile.nBlks() % BLKCHK) == 0) {
	    cout << endl << "write: " << l1aoutfile.nBlks() << endl;
	    for (size_t j=0; j<nInput; j++) {
	      cout << setw (2);
	      cout << j;
	      cout << setw (6);
	      cout << blkPtr[j];
	      cout << setw (6);
	      cout << l1afile[j].nBlks() << endl;
	    }
	  }

	  // Increment block count
	  for (size_t j=0; j<icds_blknum[i][blkPtr[i]]-prev_icds_blknum; j++)
	    l1aoutfile.incBlks();
	  if (l1aoutfile.nBlks() >= MAXSCAN) {
	    cout << "*** MAXSCAN parameter: " << MAXSCAN << 
	      " is too small." << endl;
	    exit (110);
	  }

	  prev_icds_blknum = icds_blknum[i][blkPtr[i]];
	  prev_rad_header = rad_header[i][blkPtr[i]];
	  first_write = false;

	} // if write
	blkPtr[i]++;
      } // if Time
    } // input file loop

    minTime = 1e30;
    for (size_t i=0; i<nInput; i++) {
      if (blkPtr[i] == -1) continue;
      if (blkPtr[i] < (int32_t) l1afile[i].nBlks()) {
	if (minTime > Time[i][blkPtr[i]]) minTime = Time[i][blkPtr[i]];
      }
    } // input file loop
    write = true;

    // Check if EOF for all input files
    uint32_t eoF=0;
    for (size_t i=0; i<nInput; i++) {
      if (blkPtr[i] == -1) continue;
      //      if ((blkPtr[i] == l1afile[i].nBlks()) || 
      //	  (Time[i][blkPtr[i]] > (orbitStop+10*60))) eoF++;

      if ((blkPtr[i] == (int32_t) l1afile[i].nBlks()) || 
	  (Time[i][blkPtr[i]] >= (orbitStop+10*60))) eoF++;
    }
    if (eoF == nInput) break;
  } // while


  // ====================== Copy Orbit records ======================
  cout << "Copying Orbit Records" << endl;

  // Read orbit times
  double *OrbTime[MAXINPUT];
  uint32_t numOrbVec;
  for (size_t i=0; i<nInput; i++) {
    l1afile[i].readl1_eph( &numOrbVec);
    OrbTime[i]= new double[numOrbVec+1];
    OrbTime[i][numOrbVec] = 0;
    l1afile[i].readl1_eph( OrbTime[i], NULL, NULL);
  }

  // Initialize orbPtr
  uint32_t orbPtr[MAXINPUT];
  for (size_t i=0; i<nInput; i++) {
    orbPtr[i] = 0;
    l1afile[i].readl1_eph( &numOrbVec);
    for (size_t j=0; j<numOrbVec; j++) {
      if (OrbTime[i][j] > (orbitStart-10*60)) {
	if (j > 0) orbPtr[i] = j-1; else orbPtr[i] = j;
	break;
      }
    }
  }

  // Determine minimum orbit time
  double minorbTime = OrbTime[0][orbPtr[0]];
  for (size_t i=1; i<nInput; i++) {
    if (minorbTime > OrbTime[i][orbPtr[i]]) minorbTime = OrbTime[i][orbPtr[i]];
  }

  int32_t numOrbWrite=0;

  while (1) {
    for (size_t i=0; i<nInput; i++) {
      l1afile[i].readl1_eph( &numOrbVec);
      // if (OrbTime[i][orbPtr[i]] == minorbTime && (orbPtr[i] < numOrbVec)) {
      if (OrbTime[i][orbPtr[i]] == minorbTime) {
	if (write) {	
	  write = false;

	  size_t jgrp = 4;

	  // Loop through all group members
	  for (size_t k=0; k<numMembs[jgrp]; k++) {
	    H5Gget_objname_by_idx(grp_in[i][jgrp], k, namebuf, NAMELEN);
	    H5G_obj_t objType = H5Gget_objtype_by_idx(grp_in[i][jgrp], k);
	    if (objType == H5G_DATASET) {
	      grp[0] = grp_in[i][jgrp];
	      grp[1] = grp_out[jgrp];

	      // Copy record
	      if (strncmp(namebuf, "orb_", 4) == 0)
		l1afile[i].copyl1(orbPtr[i], -1, numOrbWrite, grp, namebuf);

	    } // if dataset
	  } // group member loop
	  /*
	  for (size_t j=0; j<nInput; j++) {
	    l1afile[j].readl1_eph( &numOrbVec);
	    cout << j << " " <<orbPtr[j] << " " << numOrbVec << endl;
	    cout << (int32_t) OrbTime[j][orbPtr[j]] << " " << (int32_t) minorbTime << endl << endl;
	  }
	  */

	  numOrbWrite++;

	} // if write
	orbPtr[i]++;
	//	cout << i << " " << orbPtr[i] << " " << (long) OrbTime[i][orbPtr[i]] << " " << numOrbWrite << " " << endl;
      } // if OrbTime
    } // i loop

    minorbTime = 1e30;
    for (size_t i=0; i<nInput; i++) {
      l1afile[i].readl1_eph( &numOrbVec);
      if (orbPtr[i] < numOrbVec) {
	if (minorbTime > OrbTime[i][orbPtr[i]])
	  minorbTime = OrbTime[i][orbPtr[i]];
      }
    } // input file loop
    write = true;

    // Check if EOF for all input files
    uint32_t eoF=0;
    for (size_t i=0; i<nInput; i++) {
      l1afile[i].readl1_eph( &numOrbVec);
      // Need 2 orb records after orbitstop + wing for scat spline interp
      if ((orbPtr[i] == numOrbVec) || 
	  (OrbTime[i][orbPtr[i]] > (orbitStop+10*60+60))) 
	eoF++;
    }

    // if eof for all files that exit while loop
    if (eoF == nInput) break;
  }

  l1aoutfile.setOrbVec( numOrbWrite);


  // ====================== Copy Attitude records ======================
  cout << "Copying Attitude Records" << endl;

  // Read attitude times
  double *AttTime[MAXINPUT];
  uint32_t numAttSamp;
  for (size_t i=0; i<nInput; i++) {
    l1afile[i].readl1_att( &numAttSamp);
    AttTime[i]= new double[numAttSamp+1];
    AttTime[i][numAttSamp] = 0;
    l1afile[i].readl1_att( AttTime[i], NULL, NULL);
  }

  // Initialize attPtr
  uint32_t attPtr[MAXINPUT];
  for (size_t i=0; i<nInput; i++) {
    attPtr[i] = 0;
    l1afile[i].readl1_att( &numAttSamp);
    for (size_t j=0; j<numAttSamp; j++) {
      if (AttTime[i][j] > (orbitStart-10*60)) {
	if (j > 0) attPtr[i] = j-1; else attPtr[i] = j;
	break;
      }
    }
  }


  // Determine minimum attitude time
  double minattTime = AttTime[0][attPtr[0]];
  for (size_t i=1; i<nInput; i++) {
    if (minattTime > AttTime[i][attPtr[i]]) minattTime = AttTime[i][attPtr[i]];
  }

  int32_t numAttWrite=0;

  while (1) {
    for (size_t i=0; i<nInput; i++) {
      if (AttTime[i][attPtr[i]] == minattTime) {
	if (write) {	
	  write = false;

	  size_t jgrp = 4;

	  // Loop through all group members
	  for (size_t k=0; k<numMembs[jgrp]; k++) {
	    H5Gget_objname_by_idx(grp_in[i][jgrp], k, namebuf, NAMELEN);
	    H5G_obj_t objType = H5Gget_objtype_by_idx(grp_in[i][jgrp], k);
	    if (objType == H5G_DATASET) {
	      grp[0] = grp_in[i][jgrp];
	      grp[1] = grp_out[jgrp];

	      // Copy record
	      if (strncmp(namebuf, "att_", 4) == 0)
		l1afile[i].copyl1(attPtr[i], -1, numAttWrite, grp, namebuf);
	      if (strncmp(namebuf, "quat", 4) == 0)
		l1afile[i].copyl1(attPtr[i], -1, numAttWrite, grp, namebuf);

	    } // if dataset
	  } // group member loop

	  numAttWrite++;
 
	} // if write
	attPtr[i]++;
      } // if AttTime
    } // i loop

    minattTime = 1e30;
    for (size_t i=0; i<nInput; i++) {
      l1afile[i].readl1_att( &numAttSamp);
      if (attPtr[i] < numAttSamp) {
	if (minattTime > AttTime[i][attPtr[i]])
	  minattTime = AttTime[i][attPtr[i]];
      }
    } // input file loop
    write = true;

    // Check if EOF for all input files
    uint32_t eoF=0;
    for (size_t i=0; i<nInput; i++) {
      l1afile[i].readl1_att( &numAttSamp);
      // Need 2 att records after orbitstop + wing for scat spline interp
      if ((attPtr[i] == numAttSamp) || 
	  (AttTime[i][attPtr[i]] > (orbitStop+10*60+8))) eoF++;
    }
    if (eoF == nInput) break;
  }

  l1aoutfile.setAttSamp( numAttWrite);



  // ====================== Copy SAC-D HKT records ======================
  cout << "Copying HKT Records" << endl;

  // Read HKT times
  uint32_t *HktTime[MAXINPUT];
  uint32_t numHktRec[MAXINPUT];
  for (size_t i=0; i<nInput; i++) {
    l1afile[i].readl1_hkt( &numHktRec[i]);
    if ( numHktRec[i] == 0) {
      cout << "No valid SAC-D HKT records in input file: " << i << endl;
      exit(1);
    }
    HktTime[i]= new uint32_t[numHktRec[i]+1];
    HktTime[i][numHktRec[i]] = 0;
    l1afile[i].readl1_hkttme( numHktRec[i], HktTime[i]);
  }



  // Initialize hktPtr
  uint32_t hktPtr[MAXINPUT];
  for (size_t i=0; i<nInput; i++) {
    hktPtr[i] = 0;
    for (size_t j=0; j<numHktRec[i]; j++) {
      if (HktTime[i][j] > (orbitStart-10*60)) {
	if (j > 0) hktPtr[i] = j-1; else hktPtr[i] = j;
	break;
      }
    }
  }

  // Determine minimum HKT time
  double minhktTime = HktTime[0][hktPtr[0]];
  for (size_t i=1; i<nInput; i++) {
    if (minhktTime > HktTime[i][hktPtr[i]]) minhktTime = HktTime[i][hktPtr[i]];
  }

  int32_t numHktWrite=0;

  while (1) {
    for (size_t i=0; i<nInput; i++) {
      if (HktTime[i][hktPtr[i]] == minhktTime) {
	if (write) {	
	  write = false;

	  size_t jgrp = 3;

	  // Loop through all group members
	  for (size_t k=0; k<numMembs[jgrp]; k++) {
	    H5Gget_objname_by_idx(grp_in[i][jgrp], k, namebuf, NAMELEN);
	    H5G_obj_t objType = H5Gget_objtype_by_idx(grp_in[i][jgrp], k);
	    if (objType == H5G_DATASET) {
	      grp[0] = grp_in[i][jgrp];
	      grp[1] = grp_out[jgrp];

	      // Copy record
	      l1afile[i].copyl1(hktPtr[i], -1, numHktWrite, grp, namebuf);

	    } // if dataset
	  } // group member loop

	  numHktWrite++;
 
	} // if write
	hktPtr[i]++;
      } // if HktTime
    } // i loop

    minhktTime = 1e30;
    for (size_t i=0; i<nInput; i++) {
      if (hktPtr[i] < numHktRec[i]) {
	if (minhktTime > HktTime[i][hktPtr[i]])
	  minhktTime = HktTime[i][hktPtr[i]];
      }
    } // input file loop
    write = true;


    // Check if EOF for all input files
    uint32_t eoF=0;
    for (size_t i=0; i<nInput; i++) {
      // Need 2 hkt records after orbitstop + wing for scat spline interp
      if ((hktPtr[i] == numHktRec[i]) || 
	  (HktTime[i][hktPtr[i]] > (orbitStop+10*60+8))) eoF++;
    }
    if (eoF == nInput) break;
  } // while loop



  // Determine maximum granule time
  double maxTime = -1e30;
  for (size_t i=0; i<nInput; i++) {
    if (blkPtr[i] != -1)
      if (maxTime < Time[i][blkPtr[i]-1]) {
	maxTime = Time[i][blkPtr[i]-1];
	//	cout << i << " " << blkPtr[i]-1 << " " << maxTime << endl;
      }
  }

 

  // Close input L1A files
  for (size_t i=0; i<nInput; i++) l1afile[i].closel1();

  // Delete dynamic arrays
  for (size_t i=0; i<nInput; i++) {
    delete[] Time[i];
    delete[] atc_frmnum[i];
    delete[] atc_subframe[i];
    delete[] rad_frmnum[i];
    delete[] rad_subframe[i];
    delete[] icds_blknum[i];
    delete[] rad_header[i];
    delete[] OrbTime[i];
    delete[] AttTime[i];
    delete[] HktTime[i];
  }

  // Determine Orbit Node Longitude
  l1aoutfile.readl1_eph( &numOrbVec);
  OrbTime[0] = new double[numOrbVec];
  double *OrbPos = new double[numOrbVec*3];
  l1aoutfile.readl1_eph( OrbTime[0], OrbPos, NULL);
  //for (size_t i=0; i<numOrbVec; i++) cout << i << " " << OrbPos[3*i+2] << endl;
  float nodeLongitude=-999;
  double nodeCrossingTime=-999, vec_in[6], vec_out[6];
  cdata_();
  for (size_t i=0; i<numOrbVec-1; i++) {
    if ( OrbPos[3*i+2] < 0 && OrbPos[3*(i+1)+2] > 0) {
      nodeCrossingTime = 0.5*(OrbTime[0][i+1] + OrbTime[0][i]);
      vec_in[0] = OrbPos[3*i];
      vec_in[1] = OrbPos[3*i+1];
      vec_in[2] = OrbPos[3*i+2];
      vec_in[3] = 0;
      vec_in[4] = 0;
      vec_in[5] = 0;

      tme = (time_t) gpstai2unix( nodeCrossingTime);
      tm0 = gmtime(&tme);

      double daysec = tm0->tm_hour*3600 + tm0->tm_min*60 + tm0->tm_sec; 
      int year = 1900 + tm0->tm_year;
      int doy = tm0->tm_yday;
      int nvec = 1;
      asap_rots_( &year, &doy, &daysec, vec_in, &nvec, vec_out);
      nodeLongitude = atan2( vec_out[1], vec_out[0]) * (180 / PI);
      break;
    }
  }

  // Close output L1A file
  cout << "granStart: " << granuleStart << endl;
  cout << "maxTime:   " << maxTime << endl;

  uint8_t rd_sts_oplut;
  int32_t oplut_accum[8] = {0,0,0,0,0,0,0,0};
  int32_t n_non_default = 0;
  int32_t iblk;
  int32_t nBlks = l1aoutfile.nBlks();
  int32_t atc_frmnum0, rad_frmnum0;
  uint8_t atc_subframe0, rad_subframe0;
  for (iblk=0; iblk<nBlks; iblk++) {
    l1aoutfile.readl1_frame_info( iblk,
				  &atc_frmnum0, &atc_subframe0,
				  &rad_frmnum0, &rad_subframe0);

    l1aoutfile.readl1_dpu_status_tlm( rad_frmnum0, rad_subframe0, 10, 
				      &rd_sts_oplut);
    oplut_accum[rd_sts_oplut]++;
    if ( rd_sts_oplut != 0) n_non_default++;
  }
  float percent_non_default = ((float) n_non_default) / nBlks;

  int32_t missingBlks = lrint((maxTime - granuleStart)/1.44) + 1 - nBlks;

  l1aoutfile.closel1( granuleStart, maxTime, 
		      nodeCrossingTime, nodeLongitude,
		      oplut_accum, percent_non_default, missingBlks);

  delete[] OrbTime[0];
  delete[] OrbPos;

  return 0;
}


