C--$Header: /mnt/caretr2/scatsim/CVSROOT/L1A_to_L2/L1B_compute_kpc.f,v 1.6 2009/10/05 14:30:08 fore Exp $
****************************************************************

      subroutine L1B_compute_kpc( r_rx_power, r_Vnoise, r_Hnoise, i_beam, sc_position, 
     &   sc_velocity, yaw_deg, pitch_deg, roll_deg, r_kpc_out, r_TBd, r_snr_dB_out )
      
c*******************************************************************
c**
c**   FUNCTIONAL DESCRIPTION: This routine computes theoretical
c**   estimates of Kpc for the measured data, and outputs Kpc, SNR,
c**   and the dechirped time-bandwidth product TBd. Kpc represents relative error
c**   (delta sig-0/sig-0) for ONE measurement (one PRI), not one block of 8 measurements.
c**   Kpc formula based on M. Spencer's SeaWinds "egg" formula
c**   (see EM 334G.2004.01(rev A), 10 Dec 2004: Aquarius Scatterometer Measurement Errors: I)
c**
c**    DATE WRITTEN:   25-JUNE-2009
c**
c**    PROGRAMMER:  Adam P. Freedman
c**
c**
c**    ROUTINES CALLED: 
c**
c**    INCLUDE FILES:  
c**
c**    NOTES: 
c**
c**    UPDATE LOG: Integrated with L1A_to_L2 on 9/3/2009
c**
c**    Date Changed        Reason Changed                  CR # and Version #
c**    ------------       ----------------                 -----------------  
c**
c** 
c*******************************************************************

      implicit none

c     PARAMETER AND STRUCTURE DEFINITION FILES: 

      include 'L1B_instrument_module_params.inc'
c--Contains a common block and declarations of the folowing variables...
C--and declares a common block called /instrument_module_params/ with all of these variables in it.
C
C      real*8 r_SSPA_PowerW  
C      real*8 r_SFE_tx_loss(6)
C      real*8 r_SFE_rx_loss(6)
C      real*8 r_SFE_LB_loss(6)
C      real*8 r_SBE_rx_loss(2)
C      real*8 r_SBE_gain      
C      real*8 r_OMT_loss(2,6)
C      real*8 r_tx_reflect
C      real*8 r_SBE_noise_fig
C      real*8 r_rx_BW
C      real*8 r_tx_BW
C      real*8 r_tx_frequency
C      real*8 r_rx_length(3)
C      real*8 r_chirp_rate
C      real*8 r_tx_pulse_length
C      real*8 r_B1HV
C      real*8 r_B1NV
C      real*8 r_T_cnd(3)
C      real*8 r_gain_bias(3,0:2,2)  ! (beam,TX_pol,RX_pol) (TX_pol==0 means noise, 1 means H, 2 is V )
C      real*8 r_averaging_time(3,2) ! (beam #, RX-pol) (RX-pol==1 means H RX-pol==2 means V)
C      real*8 r_averaging_time_LB
       
c     INPUT VARIABLES:
                     
      real*8 r_rx_power(10)     ! power in linear units (only use 1-6 -- HV nV VV VH nH HH )
      real*8 r_Vnoise           ! noise power (linear units) for V-pol to use for SNR, KPC computations
      real*8 r_Hnoise           ! noise power (linear units) for H-pol to use for SNR, KPC computations
                                ! Note that r_Vnoise and r_Hnoise may be averaged noise values!
      
      integer i_beam                    ! beam number
      real*8 sc_position(3)     ! spacecraft position
      real*8 sc_velocity(3)     ! spacecraft veloctiy
      real*8 yaw_deg                    ! must be in degrees
      real*8 pitch_deg                  ! must be in degrees
      real*8 roll_deg                   ! must be in degrees

c     OUTPUT VARIABLES:
      real*8 r_kpc_out(4)               ! Kpc order: HV,VV,VH,HH [??]
      real*8 r_snr_dB_out(4)    ! SNR order: HV,VV,VH,HH [dB]
                                ! These variables are just used to swap around the order 
                                ! of the output variables for L1A_to_L2 software. AGF 9/3/2009

      real*8 r_TBd      ! dechirped bandwidth-time product for beam i_beam

c     LOCAL VARIABLES:

      real*8 r_kpc(4)           ! Kpc for each polarization, order: VV, HV, VH, HH, for beam i_beam
      real*8 r_snr_dB(4)        ! SNR in dB for each polarization for beam i_beam

      real*8 r_look_reduction
      real*8 r_first_term
      real*8 r_second_term
      real*8 r_snr_mod
      
      integer i, ii, i_pol
      
      real*8 r_beam_power_dB(4), r_beam_power_lin(4)    ! measured echo power for 4 polarizations, order VV, HV, VH, HH
      real*8 r_noise_power_dB(2), r_noise_power_lin(2)  ! measured noise-only power for nV, nH
      real*8 r_echo_power(4)    ! noise-free echo power detected, 4 polarizations
      real*8 r_snr_lin(4)               ! signal to noise ration, linear, for 4 polarizations
      
      real*8 r_TBn              ! the noise time-bandwidth product
      real*8 r_Kp_SNR(4)                ! The SNR-dependent portion of the Kpc equation, for 4 polarizations
      
      real*8 r_theta_B(4), r_phi_B(4)   ! 3-dB gain pointing directions for 4 corner points of beam
                                                                        ! order of points: phi = 0, +90, +180, +270
      real*8 r_vector_beam(3)   ! unit vector for corner points of beam
      real*8 earth_position(3)  ! geolocation of r_vector_beam in XYZ
      real*8 slant_range                ! slant range from S/P to earth_position, in meters
      logical good_solution             ! .true. indicates good position, .false. indicates error      
      real*8 r_2way_time(4)             ! round-trip travel time for 4 corner points of beam
      real*8 r_mintime, r_maxtime       ! min and max round-trip travel times
      real*8 r_fill_time                ! footprint 3-dB fill time
      real*8 r_Kp_Nlooks                ! number of effective looks for single radar echo measurement
      real*8 r_Nrxwin(4)        ! number of receive window lengths used in noise averaging
      
c     BLOCK DATA:

      real*8 theta_B_edge(3,3,4)  ! first index i_beam, second index i_pol, third index 4 corner pts
      real*8 phi_B_edge(3,3,4)
      common /beam_edge_points/ theta_B_edge, phi_B_edge
      
c     DATA STATEMENTS:

      real*8 R_SOL              ! speed of light (put in constants.f file with include statement?)
      data R_SOL/299792458.0d0/ ! m/s
      real*8 pi
      data pi /3.141592653589793238d0/

C     common blocks:

      integer i_dbg_level
      common /debug/ i_dbg_level

c     PROCESSING STEPS:

c       Compute SNR: Spencer formulation

c       SNR = total returned energy/total noise energy = power integrated over beam, weighted by Rx window/noise power


C--Swap around the indicies since this subroutine expects things
C--to be in the (VV,HV,VH,HH) order. We will swap them back before
C--returning variables to main software.
C--real*8 array r_rx_power has 6 non-zero entries from L1A_to_L2: order (HV, nV, VV, VH, nH, HH)
      r_beam_power_dB( 1 ) = 10.0d0*log10( r_rx_power( 3 ) ) ! VV power in dB
      r_beam_power_dB( 2 ) = 10.0d0*log10( r_rx_power( 1 ) ) ! HV power in dB
      r_beam_power_dB( 3 ) = 10.0d0*log10( r_rx_power( 4 ) ) ! VH power in dB
      r_beam_power_dB( 4 ) = 10.0d0*log10( r_rx_power( 6 ) ) ! HH power in dB
      
      do i=1, 4
         r_beam_power_lin(i) = 10.0d0**(r_beam_power_dB(i)/10.d0)
      end do
      
      r_noise_power_lin(1) = r_Vnoise ! nV power with no CND firing (may be an averaged noise power!)
      r_noise_power_lin(2) = r_Hnoise ! nH power (may be an averaged noise power!)
                                      ! L1B_compute_params.inc contains the averaging windows
                                      ! for the noise.

      do i = 1, 2
         r_noise_power_dB(i) = 10.0d0*log10( r_noise_power_lin(i) )  ! order nV, nH
      end do


      do i = 1, 2
         r_echo_power(i)   = r_beam_power_lin(i)   - r_noise_power_lin(1)       ! V channels
         r_echo_power(i+2) = r_beam_power_lin(i+2) - r_noise_power_lin(2)       ! H channels
         r_snr_lin(i)      = r_echo_power(i)       / r_noise_power_lin(1)
         r_snr_lin(i+2)    = r_echo_power(i+2)     / r_noise_power_lin(2)
      end do
      
      do i = 1, 4
         r_snr_dB(i) = 10.0d0*log10( abs( r_snr_lin(i) ) ) ! use absolute value  
      end do
      
            
c       noise bandwidth-time coefficient
          r_TBn = r_rx_BW * r_rx_length(i_beam)

c   Compute noise integration time NEW**
      ! times assumed in seconds....
      r_Nrxwin(1) = max(r_averaging_time(i_beam,2) / 0.18, 1.0) / 2.0 ! V channels
      r_Nrxwin(2) = r_Nrxwin(1)
      r_Nrxwin(3) = max(r_averaging_time(i_beam,1) / 0.18, 1.0) ! H channels
      r_Nrxwin(4) = r_Nrxwin(3)

c       Second half (SNR portion) of Kpc calculation

      do i = 1, 4
         r_Kp_SNR(i) = 1.0d0 / (r_TBn * r_snr_lin(i)) * (2.0d0 + (1.0d0 + 1.0d0/r_Nrxwin(i)) / r_snr_lin(i))  
      end do

c       Compute first half (geometric looks portion) of Kpc calculation

c       Compute fill time

      do ii = 1, 4
         r_theta_B(ii) = 0.0d0
         r_phi_B(ii) = 0.0d0
      end do
      do i_pol=1,3 ! 3 polarizations: HV, VV, HH
         do ii = 1, 4   ! 4 corner points: phi = 0, +90, +180, +270
            r_theta_B(ii) = r_theta_B(ii) + theta_B_edge(i_beam,i_pol,ii) / 3.0d0 
            r_phi_B(ii)   = r_phi_B(ii) + phi_B_edge(i_beam,i_pol,ii) / 3.0d0
         enddo
      end do

      
      do ii = 1, 4      
        r_vector_beam(1) = sin( r_theta_B(ii) * PI / 180.0d0 ) * cos( r_phi_B(ii) * PI / 180.0d0 )
        r_vector_beam(2) = sin( r_theta_B(ii) * PI / 180.0d0 ) * sin( r_phi_B(ii) * PI / 180.0d0 )
        r_vector_beam(3) = cos( r_theta_B(ii) * PI / 180.0d0 )
        
        call geolocate( r_vector_beam, sc_position, sc_velocity,
     $ yaw_deg, pitch_deg, roll_deg, i_beam, earth_position,
     $ slant_range, good_solution )
     
        if (good_solution) then
           r_2way_time(ii) = 2.0d0 * slant_range / R_SOL
        else
           r_2way_time(ii) = -1.0d0
           write(6,*) 'L1B_compute_kpc: WARNING: Imaginary solution found!'
        end if
      end do
      
      !! min travel time should be inner 3-dB point, max travel time should be outer 3-dB point
      
      r_mintime = 10000.d0
      r_maxtime = -10000.d0
      do ii = 1, 4
         r_mintime = min(r_mintime,r_2way_time(ii))
         r_maxtime = max(r_maxtime,r_2way_time(ii))
      end do
      
      r_fill_time = r_maxtime - r_mintime
      if (r_mintime.lt.0) then
         r_fill_time = 1.0d-6 ! small fill time => large Kpc for problem points
      end if
      
      if( i_dbg_level.ge.1) write(6,*) 'L1B_compute_kpc: fp_fill_time: ',r_fill_time
      
c       Compute tx bandwidth * fill time = tx time * reduced band-width ("dechirped") 

      r_TBd = r_fill_time * r_chirp_rate * r_tx_pulse_length
      
      r_Kp_Nlooks = 1 / r_TBd
      
c       Total Kpc

      do i = 1, 4
         r_kpc(i) = sqrt( r_Kp_Nlooks + r_Kp_SNR(i) )
      end do
      
C--Swap order of output arrays before returning them to L1A_to_L2 main software
      r_snr_dB_out( 1 ) = r_snr_dB( 2 ) ! HV
      r_snr_dB_out( 2 ) = r_snr_dB( 1 ) ! VV
      r_snr_dB_out( 3 ) = r_snr_dB( 3 ) ! VH
      r_snr_dB_out( 4 ) = r_snr_dB( 4 ) ! HH

      r_kpc_out( 1 ) = r_kpc( 2 ) ! HV
      r_kpc_out( 2 ) = r_kpc( 1 ) ! VV
      r_kpc_out( 3 ) = r_kpc( 3 ) ! VH
      r_kpc_out( 4 ) = r_kpc( 4 ) ! HH

      if( i_dbg_level.ge.2 )then
        write(6,'(a50)') 'L1B KPC dbg output:'
        write(6,'((a50),2f12.6)') 'VV, echo_pwr, noise_pwr [linear]: ', r_echo_power(1), r_noise_power_lin(1)
        write(6,'((a50),2f12.6)') 'HV, echo_pwr, noise_pwr [linear]: ', r_echo_power(2), r_noise_power_lin(1)
        write(6,'((a50),2f12.6)') 'VH, echo_pwr, noise_pwr [linear]: ', r_echo_power(3), r_noise_power_lin(2)
        write(6,'((a50),2f12.6)') 'HH, echo_pwr, noise_pwr [linear]: ', r_echo_power(4), r_noise_power_lin(2)

        write(6,'((a50),4f12.6)') 'VV, echo_pwr, noise_pwr, SNR, KPC [dB]: ',
     1                             r_beam_power_dB(1), r_noise_power_dB(1), r_snr_dB(1), r_kpc(1)
        write(6,'((a50),4f12.6)') 'HV, echo_pwr, noise_pwr, SNR, KPC [dB]: ',
     1                             r_beam_power_dB(2), r_noise_power_dB(1), r_snr_dB(2), r_kpc(2)
        write(6,'((a50),4f12.6)') 'VH, echo_pwr, noise_pwr, SNR, KPC [dB]: ', 
     1                             r_beam_power_dB(3), r_noise_power_dB(2), r_snr_dB(3), r_kpc(3)  
        write(6,'((a50),4f12.6)') 'HH, echo_pwr, noise_pwr, SNR, KPC [dB]: ', 
     1                             r_beam_power_dB(4), r_noise_power_dB(2), r_snr_dB(4), r_kpc(4)
      endif             


      
      return
      end


      
