C--$Header: /u/caretr2/scatsim/CVSROOT/L1A_to_L2/L1B_gen_quality_flag.f,v 1.16 2012/04/27 16:31:43 fore Exp $
c*****************************************************************      
      subroutine L1B_gen_quality_flag( i_seq,
     $                                 calibration_level, 
     $                                 L1B_land_fraction,
     $                                 L1B_ice_fraction,
     $                                 antlook,
     $                                 bad_time_tag_flag,
     $                                 bad_block_id_flag,
     $                                 unable_to_geolocate_flag,
     $                                 beam_off_earth_flag,
     $                                 bad_kfactor_flag,
     $                                 RFI_FLAGS_VPOL,
     $                                 RFI_FLAGS_HPOL,
     $                                 RX_power,
     $                                 LB_power,
     $                                 scat_telemetry_eu,
     $                                 external_temperatures_eu,
     $                                 icds_engineering_telemetry_eu,
     $                                 flag_suppress_tlm_warnings,
     $                                 L1B_quality_flag     )
c*****************************************************************      
c**   
c**   FILE NAME: L1B_gen_quality_flag.f
c**   
c**   DATE WRITTEN: 9/15/2009
c**   
c**   PROGRAMMER: Alex Fore
c**   
c**   FUNCTIONAL DESCRIPTION:  
c**
c**   ROUTINES CALLED: 
c**   
c**   NOTES: Meant to be called from inside the i_seq, 
c**    i_block loops.  i.e. this subroutine should be called once
c**    for every (block,sequence) combination.
c**
c**    Based on Limits_table_v2c.xls from APF
c**
c**    Note that there are three other place in L1A_to_L2.f where
c**    the quality flag is modified:
c**     -- It is intialized to zero at the start of every block in the driver (L1A_to_L2.f)
c**     -- The negative sigma0 bit is set when we compute the echo-noise
c**     -- we check the CND gain after this subroutine call, 
c**        after the i_seq loop in the main driver.
c**
c**
c**   UPDATE LOG: Updated 9/22/2009, replaced inc angle with look angle limits.
c**    Added flag to suppress telemetry warnings if user requests.
c**
c**   Date Changed        Reason Changed        CR # and Version #
c**   ------------       ----------------       ------------------
c**     
c*****************************************************************   
      implicit none
      include 'L1B.inc' ! contains the flagging thresholds, read in 
                        ! already by main software (L1A_to_L2 calls 
                        ! subroutine L1B_read_limits) from 
                        ! text file "L1B_limits_YYYY-MM-DD.txt".
                        
c     PARAMETER STATEMENTS
      integer stdout, stderr, stdin
      parameter( stdout = 6, stderr = 0, stdin = 5 )
      
c     INPUT VARIABLES
      integer i_seq                   ! which sequence (1-8) -- so we know if CND is firing or not.
      integer calibration_level       ! integer, specified by user at run-time
      
      real*8  L1B_land_fraction(3)    ! land fraction for beams 1-3
      real*8  L1B_ice_fraction(3)     ! ice fraction for beams 1-3
      real*8  antlook(3)              ! antenna look angle in deg for beams 1-3
      
      integer*1 RFI_FLAGS_VPOL(3,8)   ! set to 1 if RFI, 0 if no RFI, from L0 data
      integer*1 RFI_FLAGS_HPOL(3,8)   ! set to 1 if RFI, 0 if no RFI, from L0 data

      
      real*8 RX_power(6,3,8)          ! [dB] indicies are: i_chan, i_beam, i_seq 
      real*8 LB_power(6,3)            ! [dB] indicies are: i_chan, i_beam
      
      real*4 scat_telemetry_eu( 38 )
      real*4 external_temperatures_eu( 38 )
      real*4 icds_engineering_telemetry_eu( 24 )
      
      logical bad_time_tag_flag        ! generated earlier in L1B processing
      logical bad_block_id_flag        ! generated earlier in L1B processing
      logical beam_off_earth_flag(3)   ! generated earlier in L1B processing
      logical bad_kfactor_flag(3)      ! generated earlier in L1B processing      
      logical unable_to_geolocate_flag ! generated earlier in L1B processing
      logical flag_suppress_tlm_warnings ! suppress warnings for out-of-range telemetry values
                                         

c     OUTPUT VARIABLES 
      integer*2 L1B_quality_flag(4,3)  ! one for each (channel,beam) 
                                       ! channel order is (HV,VV,VH,HH)
                                       ! beam order is (1,2,3)

c     LOCAL VARIABLES
      integer i_chan, i_beam
      
      integer four_chan_to_six_map(4) ! six channels including noise only RX, 4 without.

      integer i_bad_data_flag         ! integer temporaries for bit(s) that
      integer i_bad_calibration_flag  ! may be set by many different conditions.
      integer i_hardware_status_flag  ! I can pass these around and set the bit
      integer i_anomalous_temps_flag  ! in L1B_quality_flag at the end.
      
      real*8 tmp_lower_bound, tmp_upper_bound
      real*8 tmp_pwr
      
      real*8 r_kBoltz /1.38d-23/      ! Boltzmann's constant
      real*8 r_kB_BW                  ! r_kBoltz * rx bandwidth
      logical has_rfi
      integer tmp_dbg_level
      
c     DATA STATEMENTS
c     COMMON BLOCKS
      integer i_dbg_level
      common /debug/ i_dbg_level
c     FUNCTION STATEMENTS
c     PROCESSING STEPS
      
      tmp_dbg_level = -10 ! Always write out warnings
      
C--This is just an index map to make things easier below.      
      four_chan_to_six_map(1) = 1 ! HV
      four_chan_to_six_map(2) = 3 ! VV
      four_chan_to_six_map(3) = 4 ! VH
      four_chan_to_six_map(4) = 6 ! HH
      
      do i_beam = 1, 3
        do i_chan = 1, 4
C--Check RFI flags---Changed 9/26/2011 for new L1B RFI flagging scheme
          if( i_chan.le.2 )then                                 ! V-RX polarization
            if( mod(RFI_FLAGS_VPOL(i_beam,i_seq),2).eq.0 ) then ! not instrument flagged
              if( RFI_FLAGS_VPOL(i_beam,i_seq).eq.0) then       ! not outlier flagged
                call L1B_set_flag( L1B_quality_flag(i_chan,i_beam), L1B_QF_RFI, L1B_QF_RFI_none )
              else                                              ! outlier flagged
                call L1B_set_flag( L1B_quality_flag(i_chan,i_beam), L1B_QF_RFI, L1B_QF_RFI_outlier_only )
              endif
            else                                                ! instrument flagged
              if( RFI_FLAGS_VPOL(i_beam,i_seq).eq.1) then       ! not outlier flagged
                call L1B_set_flag( L1B_quality_flag(i_chan,i_beam), L1B_QF_RFI, L1B_QF_RFI_instr_only )
              else                                              ! outlier flagged
                call L1B_set_flag( L1B_quality_flag(i_chan,i_beam), L1B_QF_RFI, L1B_QF_RFI_instr_and_outlier )
              endif              
            endif
            
          else                                                  ! H-RX polarization
            if( mod(RFI_FLAGS_HPOL(i_beam,i_seq),2).eq.0 ) then ! not instrument flagged
              if( RFI_FLAGS_HPOL(i_beam,i_seq).eq.0) then       ! not outlier flagged
                call L1B_set_flag( L1B_quality_flag(i_chan,i_beam), L1B_QF_RFI, L1B_QF_RFI_none )
              else                                              ! outlier flagged
                call L1B_set_flag( L1B_quality_flag(i_chan,i_beam), L1B_QF_RFI, L1B_QF_RFI_outlier_only )
              endif
            else                                                ! instrument flagged
              if( RFI_FLAGS_HPOL(i_beam,i_seq).eq.1) then       ! not outlier flagged
                call L1B_set_flag( L1B_quality_flag(i_chan,i_beam), L1B_QF_RFI, L1B_QF_RFI_instr_only )
              else                                              ! outlier flagged
                call L1B_set_flag( L1B_quality_flag(i_chan,i_beam), L1B_QF_RFI, L1B_QF_RFI_instr_and_outlier )
              endif              
            endif            
          endif
          
          if( iand( L1B_quality_flag(i_chan,i_beam), L1B_QF_RFI ).gt.0 )then
            has_rfi = .true.
          else
            has_rfi = .false.
          endif

C--Check all of the telemetry values and return integers for the flags.
          call L1B_check_all_telem( scat_telemetry_eu,
     $                              external_temperatures_eu,
     $                              icds_engineering_telemetry_eu,
     $                              calibration_level,
     $                              flag_suppress_tlm_warnings,
     $                              i_bad_data_flag,          ! output
     $                              i_bad_calibration_flag,   ! output
     $                              i_hardware_status_flag,   ! output
     $                              i_anomalous_temps_flag )  ! output
         
C--Set the high backscatter flag if echo power is out-of-range,
C--also elevate the hardware status flag if needed.
          tmp_pwr = RX_power(four_chan_to_six_map(i_chan),i_beam,i_seq)
          tmp_lower_bound = ECHO_PWR_limits(1,i_chan,i_beam)
          tmp_upper_bound = ECHO_PWR_limits(2,i_chan,i_beam)
          if( has_rfi ) tmp_upper_bound = 100.0
          if( .not. (tmp_pwr.ge.tmp_lower_bound .and.
     $               tmp_pwr.le.tmp_upper_bound ) ) then
            
            if( i_hardware_status_flag.eq.0) i_hardware_status_flag=1 ! Add scat hdr error
            if( i_hardware_status_flag.eq.2) i_hardware_status_flag=3 ! Add scat hdr error            
            call L1B_set_flag( L1B_quality_flag(i_chan,i_beam), 
     $                         L1B_QF_HIGH_BACKSCATTER, L1B_QF_HIGH_BACKSCATTER )

            if( i_dbg_level .ge. tmp_dbg_level ) then
              write(stdout,'((a),2I5,3f12.6)') 
     $  'WARNING: RX_power out of range i_beam, i_chan: ', i_beam, i_chan, tmp_pwr, tmp_lower_bound, tmp_upper_bound
            endif
          endif

C--Check the loopback power and elevate the hardware status flag if needed
          tmp_pwr = LB_power(four_chan_to_six_map(i_chan),i_beam)
          if( .not. (tmp_pwr.ge.LB_PWR_limits(1,i_chan,i_beam) .and.
     $               tmp_pwr.le.LB_PWR_limits(2,i_chan,i_beam) ) ) then

            i_bad_calibration_flag = 1
            if( i_hardware_status_flag.eq.0) i_hardware_status_flag=1 ! Add scat hdr error
            if( i_hardware_status_flag.eq.2) i_hardware_status_flag=3 ! Add scat hdr error
            
            if( i_dbg_level .ge. tmp_dbg_level ) then
              write(stdout,'((a),2I5,3f12.6)') 
     $  'WARNING: LB_power out of range i_beam, i_chan: ', i_beam, i_chan, tmp_pwr, tmp_lower_bound, tmp_upper_bound
            endif
          endif


C--Check the noise power and elevate the hardware status flag if needed
C--Make sure to check nearest nV channel WITHOUT CND if RX polarization is V.
          if( i_chan.le.2 ) then
            if( ( mod(i_seq,2).eq.1 .and. i_beam.eq.2 ) .or. ! CND fires on beam 2 nV for odd n_seq
     $          ( mod(i_seq,2).eq.0 .and. 
     $          (i_beam.eq.1 .or. i_beam.eq.3) ) ) then      ! CND fires on beams 1,3 nV for even n_seq

              tmp_pwr         = RX_power(2,i_beam,i_seq)
              tmp_lower_bound = CND_PWR_limits(1,i_beam)
              tmp_upper_bound = CND_PWR_limits(2,i_beam)

              if( has_rfi ) tmp_upper_bound = 100.0
              if( .not. (tmp_pwr.ge.tmp_lower_bound .and.
     $                   tmp_pwr.le.tmp_upper_bound ) ) then

                i_bad_calibration_flag = 1
                if( i_hardware_status_flag.eq.0) i_hardware_status_flag=2 ! add other hdr error
                if( i_hardware_status_flag.eq.1) i_hardware_status_flag=3 ! add other hdr error

                if( i_dbg_level .ge. tmp_dbg_level ) then
                  write(stdout,'((a),2I5,3f12.6)') 
     $  'WARNING: NOISE_CND_pwr out of range i_beam, i_chan: ', i_beam,  i_chan, tmp_pwr, tmp_lower_bound, tmp_upper_bound
                endif      
              endif
              
            else ! no CND for this (i_beam,i_seq) combination            
              tmp_pwr = RX_power(2,i_beam,i_seq)
              tmp_lower_bound = NOISE_PWR_limits(1,1,i_beam)
              tmp_upper_bound = NOISE_PWR_limits(2,1,i_beam)  
              if( has_rfi ) tmp_upper_bound = 100.0
              if( .not. (tmp_pwr.ge.tmp_lower_bound .and.
     $                   tmp_pwr.le.tmp_upper_bound ) ) then

                i_bad_calibration_flag = 1
                if( i_hardware_status_flag.eq.0) i_hardware_status_flag=1 ! Add scat hdr error
                if( i_hardware_status_flag.eq.2) i_hardware_status_flag=3 ! Add scat hdr error
            
                if( i_dbg_level .ge. tmp_dbg_level ) then
                  write(stdout,'((a),2I5,3f12.6)') 
     $  'WARNING: NOISE_pwr out of range i_beam, i_chan: ', i_beam, i_chan, tmp_pwr, tmp_lower_bound, tmp_upper_bound
                endif            
              endif
            endif            
          else ! nH channel, no CND
            tmp_pwr         = RX_power(5,i_beam,i_seq)
            tmp_lower_bound = NOISE_PWR_limits(1,2,i_beam)
            tmp_upper_bound = NOISE_PWR_limits(2,2,i_beam)
            if( has_rfi ) tmp_upper_bound = 100.0
            if( .not. (tmp_pwr.ge.tmp_lower_bound .and.
     $                 tmp_pwr.le.tmp_upper_bound ) ) then

              i_bad_calibration_flag = 1
              if( i_hardware_status_flag.eq.0) i_hardware_status_flag=1 ! Add scat hdr error
              if( i_hardware_status_flag.eq.2) i_hardware_status_flag=3 ! Add scat hdr error
            
              if( i_dbg_level .ge. tmp_dbg_level ) then
                write(stdout,'((a),2I5,3f12.6)') 
     $  'WARNING: NOISE_pwr out of range i_beam, i_chan: ', i_beam, i_chan, tmp_pwr, tmp_lower_bound, tmp_upper_bound
              endif            
            endif
          endif
                    
C--Set the calibration level flags in the quality flag depending
C--on the specifield -cal_level command line option.
          if( calibration_level .eq. 1 ) then
            call L1B_set_flag( L1B_quality_flag(i_chan,i_beam), L1B_QF_CAL_LEVEL, L1B_QF_CAL_LEVEL_1 )
          elseif( calibration_level .eq. 2 ) then
            call L1B_set_flag( L1B_quality_flag(i_chan,i_beam), L1B_QF_CAL_LEVEL, L1B_QF_CAL_LEVEL_2 )
          elseif( calibration_level .eq. 3 ) then
            call L1B_set_flag( L1B_quality_flag(i_chan,i_beam), L1B_QF_CAL_LEVEL, L1B_QF_CAL_LEVEL_3 )
          endif

C--Set the beam-off-earth flag
          if( beam_off_earth_flag(i_beam) )
     $      call L1B_set_flag( L1B_quality_flag(i_chan,i_beam), L1B_QF_BEAM_OFF_EARTH, L1B_QF_BEAM_OFF_EARTH )

C--Set the pointing error flag is antenna look angle is out-of-range.
          if( .not.
     $        ( antlook(i_beam).ge.LOOK_ANGLE_limits(1,i_beam) .and.
     $          antlook(i_beam).le.LOOK_ANGLE_limits(2,i_beam) ) )
     $        call L1B_set_flag( L1B_quality_flag(i_chan,i_beam), L1B_QF_POINTING_ERROR, L1B_QF_POINTING_ERROR )

C--Set the land/ice flag bit if the land fraction or ice fraction is larger
C--than their respective upper limits. 
          if( L1B_ice_fraction(i_beam)  .ge. ICE_FRACTION_limits(2) .or.
     $        L1B_land_fraction(i_beam) .ge. LAND_FRACTION_limits(2))
     $        call L1B_set_flag( L1B_quality_flag(i_chan,i_beam), L1B_QF_LAND_ICE, L1B_QF_LAND_ICE )
          
          if( unable_to_geolocate_flag )
     $        call L1B_set_flag( L1B_quality_flag(i_chan,i_beam), L1B_QF_GEOLOCATION_ERROR, L1B_QF_GEOLOCATION_ERROR )

          if( bad_time_tag_flag )
     $        call L1B_set_flag( L1B_quality_flag(i_chan,i_beam), L1B_QF_BAD_TIME_TAG, L1B_QF_BAD_TIME_TAG )

          
C--Set BAD_DATA and BAD_CAL if the kfactor is not computable (then sigma0 is not
C--computable. AGF 11/21/2011
          if( bad_kfactor_flag(i_beam) ) then
            i_bad_data_flag        = 1
            i_bad_calibration_flag = 1
          endif  
          
          ! Set L1B_QF_HARDWARE_PROBLEM_OTHER if bad_block_id_flag == .true.
          if( bad_block_id_flag ) then
            if( i_hardware_status_flag.eq.0 ) i_hardware_status_flag = 2
            if( i_hardware_status_flag.eq.1 ) i_hardware_status_flag = 3            
          endif
          
C--Set the bits in L1B_quality_flag depending on the already computed 
C--integers (i_bad_data_flag,i_bad_calibration_flag,i_anomalous_temps_flag,i_hardware_status_flag)
          if( i_bad_data_flag.eq.1) 
     $      call L1B_set_flag( L1B_quality_flag(i_chan,i_beam), L1B_QF_BAD_DATA, L1B_QF_BAD_DATA )

          if( i_bad_calibration_flag.eq.1 ) 
     $      call L1B_set_flag( L1B_quality_flag(i_chan,i_beam), L1B_QF_BAD_CAL, L1B_QF_BAD_CAL )

          if( i_anomalous_temps_flag.eq.1 ) 
     $      call L1B_set_flag( L1B_quality_flag(i_chan,i_beam), L1B_QF_ANOMALOUS_TEMPS, L1B_QF_ANOMALOUS_TEMPS )

          if( i_hardware_status_flag.eq.1) then
            call L1B_set_flag( L1B_quality_flag(i_chan,i_beam), L1B_QF_HARDWARE_PROBLEM, L1B_QF_HARDWARE_PROBLEM_SCAT )

          elseif( i_hardware_status_flag.eq.2) then
            call L1B_set_flag( L1B_quality_flag(i_chan,i_beam), L1B_QF_HARDWARE_PROBLEM, L1B_QF_HARDWARE_PROBLEM_OTHER )
          elseif( i_hardware_status_flag.eq.3) then
            call L1B_set_flag( L1B_quality_flag(i_chan,i_beam), L1B_QF_HARDWARE_PROBLEM, L1B_QF_HARDWARE_PROBLEM_BOTH )
          endif
        enddo
      enddo

      end
      
c*****************************************************************      
      subroutine L1B_check_all_telem( scat_telemetry_eu,
     $                                external_temperatures_eu,
     $                                icds_engineering_telemetry_eu,
     $                                calibration_level,
     $                                flag_suppress_tlm_warnings,
     $                                i_bad_data_flag,
     $                                i_bad_calibration_flag,
     $                                i_hardware_status_flag,
     $                                i_anomalous_temps_flag )
c*****************************************************************      
c**   
c**   FILE NAME: L1B_gen_quality_flag.f
c**   
c**   DATE WRITTEN: 9/15/2009
c**   
c**   PROGRAMMER: Alex Fore
c**   
c**   FUNCTIONAL DESCRIPTION:  
c**
c**   ROUTINES CALLED: 
c**   
c**   NOTES: Based on Limits_table_v2b.xls from APF
c**
c**   UPDATE LOG:
c**
c**   Date Changed        Reason Changed        CR # and Version #
c**   ------------       ----------------       ------------------
c**     
c*****************************************************************   
      implicit none
      include 'L1B.inc' ! contains the flagging thresholds, read in 
                        ! already by main software (L1A_to_L2 calls 
                        ! subroutine L1B_read_limits) from 
                        ! text file "L1B_limits.txt".
c     PARAMETER STATEMENTS
c     INPUT VARIABLES
      real*4 scat_telemetry_eu( 38 )
      real*4 external_temperatures_eu( 38 )
      real*4 icds_engineering_telemetry_eu( 24 )
      integer calibration_level
      logical flag_suppress_tlm_warnings
c     OUTPUT VARIABLES
      integer i_bad_data_flag
      integer i_bad_calibration_flag
      integer i_hardware_status_flag
      integer i_anomalous_temps_flag
c     LOCAL VARIABLES
      real*4 tmp_telem_val
      integer my_dbg_level
c     DATA STATEMENTS
c     COMMON BLOCKS
      integer i_dbg_level
      common /debug/ i_dbg_level
c     FUNCTION STATEMENTS
c     PROCESSING STEPS
      
      
      my_dbg_level = -10 ! default is to write all out of range telemetry values
      if( flag_suppress_tlm_warnings ) my_dbg_level = 10 ! suppress warnings if desired
      
      i_bad_data_flag        = 0
      i_bad_calibration_flag = 0
      i_hardware_status_flag = 0
      i_anomalous_temps_flag = 0
      
      
C--Scat telemetry checks      
      tmp_telem_val = scat_telemetry_eu(10) ! FIFO error
      if( .not. ( tmp_telem_val.ge.SC_STS_FIFO_ERR_limits(1) .and.
     $            tmp_telem_val.le.SC_STS_FIFO_ERR_limits(2) ) ) then
        
        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: FIFO error out of range: ',tmp_telem_val,SC_STS_FIFO_ERR_limits
        
        i_bad_data_flag        = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error
      endif

C--Removed per discussions with APF on 1/19/2012-------------------------------
C      tmp_telem_val = scat_telemetry_eu(8) ! SC_STS_SCAT_LVPS_OVR_CUR
C      if( .not. ( tmp_telem_val.ge.SC_STS_SCAT_LVPS_OVR_CUR_limits(1) .and.
C     $            tmp_telem_val.le.SC_STS_SCAT_LVPS_OVR_CUR_limits(2) ) ) then
C        
C        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
c     $  'WARNING: SC_STS_SCAT_LVPS_OVR_CUR out of range:', 
c     $   tmp_telem_val,SC_STS_SCAT_LVPS_OVR_CUR_limits
C        
C        i_bad_data_flag        = 1
C        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
C        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error
C      endif
C
C      tmp_telem_val = scat_telemetry_eu(7) ! SC_STS_SCAT_LVPS_BUS_VLT
C      if( .not. ( tmp_telem_val.ge.SC_STS_SCAT_LVPS_BUS_VLT_limits(1) .and.
C     $            tmp_telem_val.le.SC_STS_SCAT_LVPS_BUS_VLT_limits(2) ) ) then
C
C        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
c     $  'WARNING: SC_STS_SCAT_LVPS_BUS_VLT out of range: ',
c     $   tmp_telem_val,SC_STS_SCAT_LVPS_BUS_VLT_limits
C
C        i_bad_data_flag        = 1
C        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
C        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error
C
C      endif
C--End--Removed per discussions with APF on 1/19/2012--------------------------

      tmp_telem_val = scat_telemetry_eu(4) ! SC_STS_CG_LTCH_PRM
      if( .not. ( tmp_telem_val.ge.SC_STS_CG_LTCH_PRM_limits(1) .and.
     $            tmp_telem_val.le.SC_STS_CG_LTCH_PRM_limits(2) ) ) then
        
        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_STS_CG_LTCH_PRM out of range: ',
     $   tmp_telem_val,SC_STS_CG_LTCH_PRM_limits
        
        i_bad_data_flag        = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif

      tmp_telem_val = scat_telemetry_eu(3) ! SC_STS_CG_LTCH_RED
      if( .not. ( tmp_telem_val.ge.SC_STS_CG_LTCH_RED_limits(1) .and.
     $            tmp_telem_val.le.SC_STS_CG_LTCH_RED_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_STS_CG_LTCH_RED out of range: ',
     $   tmp_telem_val,SC_STS_CG_LTCH_RED_limits

        i_bad_data_flag        = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif
      
      tmp_telem_val = scat_telemetry_eu(11) ! SC_LVPS_BOX_TEMP
      if( .not. ( tmp_telem_val.ge.SC_LVPS_BOX_TEMP_limits(1) .and.
     $            tmp_telem_val.le.SC_LVPS_BOX_TEMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_LVPS_BOX_TEMP out of range: ',
     $   tmp_telem_val,SC_LVPS_BOX_TEMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 2    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif
      
      tmp_telem_val = scat_telemetry_eu(12) ! SC_SSPA_RF_DCK_TMP
      if( .not. ( tmp_telem_val.ge.SC_SSPA_RF_DCK_TMP_limits(1) .and.
     $            tmp_telem_val.le.SC_SSPA_RF_DCK_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_SSPA_RF_DCK_TMP out of range: ',tmp_telem_val,SC_SSPA_RF_DCK_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 2    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif

      tmp_telem_val = scat_telemetry_eu(13) ! SC_SCG_TMP
      if( .not. ( tmp_telem_val.ge.SC_SCG_TMP_limits(1) .and.
     $            tmp_telem_val.le.SC_SCG_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_SCG_TMP out of range: ',tmp_telem_val,SC_SCG_TMP_limits
        
        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 2    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif

      tmp_telem_val = scat_telemetry_eu(14) ! SC_SBE_LNA_TMP
      if( .not. ( tmp_telem_val.ge.SC_SBE_LNA_TMP_limits(1) .and.
     $            tmp_telem_val.le.SC_SBE_LNA_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_SBE_LNA_TMP out of range: ',tmp_telem_val,SC_SBE_LNA_TMP_limits

        i_anomalous_temps_flag = 1        
        if( calibration_level .ge. 2    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif

      tmp_telem_val = scat_telemetry_eu(15) ! SC_SBE_TX_CHN_TMP
      if( .not. ( tmp_telem_val.ge.SC_SBE_TX_CHN_TMP_limits(1) .and.
     $            tmp_telem_val.le.SC_SBE_TX_CHN_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_SBE_TX_CHN_TMP out of range: ',tmp_telem_val,SC_SBE_TX_CHN_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 2    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif

      tmp_telem_val = scat_telemetry_eu(16) ! SC_SBE_RX_CHN_TMP
      if( .not. ( tmp_telem_val.ge.SC_SBE_RX_CHN_TMP_limits(1) .and.
     $            tmp_telem_val.le.SC_SBE_RX_CHN_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_SBE_RX_CHN_TMP out of range: ',tmp_telem_val,SC_SBE_RX_CHN_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 2    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif

      tmp_telem_val = scat_telemetry_eu(17) ! SC_SFE_TX_LOAD_TMP
      if( .not. ( tmp_telem_val.ge.SC_SFE_TX_LOAD_TMP_limits(1) .and.
     $            tmp_telem_val.le.SC_SFE_TX_LOAD_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_SFE_TX_LOAD_TMP out of range: ',tmp_telem_val,SC_SFE_TX_LOAD_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 2    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif



      tmp_telem_val = scat_telemetry_eu(18) ! SC_SBE_STP_ATTEN_TMP
      if( .not. ( tmp_telem_val.ge.SC_SBE_STP_ATTEN_TMP_limits(1) .and.
     $            tmp_telem_val.le.SC_SBE_STP_ATTEN_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_SBE_STP_ATTEN_TMP out of range: ',tmp_telem_val,SC_SBE_STP_ATTEN_TMP_limits
        
        i_bad_data_flag        = 1
        i_anomalous_temps_flag = 1        
        if( calibration_level .ge. 1    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif

      tmp_telem_val = scat_telemetry_eu(19) ! SC_SFE_LBK_ATTEN_TMP
      if( .not. ( tmp_telem_val.ge.SC_SFE_LBK_ATTEN_TMP_limits(1) .and.
     $            tmp_telem_val.le.SC_SFE_LBK_ATTEN_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_SFE_LBK_ATTEN_TMP out of range: ',tmp_telem_val,SC_SFE_LBK_ATTEN_TMP_limits

        i_bad_data_flag        = 1
        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 1    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif
      
      tmp_telem_val = scat_telemetry_eu(20) ! SC_SFE_LBK_SW_TMP
      if( .not. ( tmp_telem_val.ge.SC_SFE_LBK_SW_TMP_limits(1) .and.
     $            tmp_telem_val.le.SC_SFE_LBK_SW_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_SFE_LBK_SW_TMP out of range: ',tmp_telem_val,SC_SFE_LBK_SW_TMP_limits

        i_bad_data_flag        = 1
        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 1    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif

      tmp_telem_val = scat_telemetry_eu(21) ! SC_SFE_BM_SW_TMP
      if( .not. ( tmp_telem_val.ge.SC_SFE_BM_SW_TMP_limits(1) .and.
     $            tmp_telem_val.le.SC_SFE_BM_SW_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_SFE_BM_SW_TMP out of range: ',tmp_telem_val,SC_SFE_BM_SW_TMP_limits

        i_bad_data_flag        = 1
        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 1    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif


      tmp_telem_val = scat_telemetry_eu(22) ! SC_SFE_GND_MON
      if( .not. ( tmp_telem_val.ge.SC_SFE_GND_MON_limits(1) .and.
     $            tmp_telem_val.le.SC_SFE_GND_MON_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_SFE_GND_MON out of range: ',tmp_telem_val,SC_SFE_GND_MON_limits

        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif

      tmp_telem_val = scat_telemetry_eu(23) ! SC_SFE_N15_VLT_MON
      if( .not. ( tmp_telem_val.ge.SC_SFE_N15_VLT_MON_limits(1) .and.
     $            tmp_telem_val.le.SC_SFE_N15_VLT_MON_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_SFE_N15_VLT_MON out of range: ',tmp_telem_val,SC_SFE_N15_VLT_MON_limits

        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif
      
      tmp_telem_val = scat_telemetry_eu(24) ! SC_SBE_PLM1_TUNG_VLT
      if( .not. ( tmp_telem_val.ge.SC_SBE_PLM1_TUNG_VLT_limits(1) .and.
     $            tmp_telem_val.le.SC_SBE_PLM1_TUNG_VLT_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_SBE_PLM1_TUNG_VLT out of range: ',tmp_telem_val,SC_SBE_PLM1_TUNG_VLT_limits

        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif

      tmp_telem_val = scat_telemetry_eu(25) ! SC_SBE_PLM2_TUNG_VLT
      if( .not. ( tmp_telem_val.ge.SC_SBE_PLM2_TUNG_VLT_limits(1) .and.
     $            tmp_telem_val.le.SC_SBE_PLM2_TUNG_VLT_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_SBE_PLM2_TUNG_VLT out of range: ',tmp_telem_val,SC_SBE_PLM2_TUNG_VLT_limits

        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif

      tmp_telem_val = scat_telemetry_eu(26) ! SC_8MHZ_STLO_OUT_PWR_LVL
      if( .not. ( tmp_telem_val.ge.SC_8MHZ_STLO_OUT_PWR_LVL_limits(1) .and.
     $            tmp_telem_val.le.SC_8MHZ_STLO_OUT_PWR_LVL_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_8MHZ_STLO_OUT_PWR_LVL out of range: ',tmp_telem_val,SC_8MHZ_STLO_OUT_PWR_LVL_limits

        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif

      tmp_telem_val = scat_telemetry_eu(27) ! SC_16MHZ_PWR_LVL
      if( .not. ( tmp_telem_val.ge.SC_16MHZ_PWR_LVL_limits(1) .and.
     $            tmp_telem_val.le.SC_16MHZ_PWR_LVL_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_16MHZ_PWR_LVL out of range: ',tmp_telem_val,SC_16MHZ_PWR_LVL_limits

        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif

      tmp_telem_val = scat_telemetry_eu(28) ! SC_SBE_PLM1_PWR_LVL
      if( .not. ( tmp_telem_val.ge.SC_SBE_PLM1_PWR_LVL_limits(1) .and.
     $            tmp_telem_val.le.SC_SBE_PLM1_PWR_LVL_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_SBE_PLM1_PWR_LVL out of range: ',tmp_telem_val,SC_SBE_PLM1_PWR_LVL_limits

        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif

      tmp_telem_val = scat_telemetry_eu(29) ! SC_SBE_PLM2_PWR_LVL
      if( .not. ( tmp_telem_val.ge.SC_SBE_PLM2_PWR_LVL_limits(1) .and.
     $            tmp_telem_val.le.SC_SBE_PLM2_PWR_LVL_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_SBE_PLM2_PWR_LVL out of range: ',tmp_telem_val,SC_SBE_PLM2_PWR_LVL_limits

        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif

      tmp_telem_val = scat_telemetry_eu(30) ! SC_SBE_TX_EXCTR_PWR_MON
      if( .not. ( tmp_telem_val.ge.SC_SBE_TX_EXCTR_PWR_MON_limits(1) .and.
     $            tmp_telem_val.le.SC_SBE_TX_EXCTR_PWR_MON_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_SBE_TX_EXCTR_PWR_MON out of range: ',tmp_telem_val,SC_SBE_TX_EXCTR_PWR_MON_limits

        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif

      tmp_telem_val = scat_telemetry_eu(31) ! SC_LVPS_CNVRTR_CURR
      if( .not. ( tmp_telem_val.ge.SC_LVPS_CNVRTR_CURR_limits(1) .and.
     $            tmp_telem_val.le.SC_LVPS_CNVRTR_CURR_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_LVPS_CNVRTR_CURR out of range: ',tmp_telem_val,SC_LVPS_CNVRTR_CURR_limits

        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif

      tmp_telem_val = scat_telemetry_eu(32) ! SC_SSPA_OUTPT_STG_VLT
      if( .not. ( tmp_telem_val.ge.SC_SSPA_OUTPT_STG_VLT_limits(1) .and.
     $            tmp_telem_val.le.SC_SSPA_OUTPT_STG_VLT_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_SSPA_OUTPT_STG_VLT out of range: ',tmp_telem_val,SC_SSPA_OUTPT_STG_VLT_limits

        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif

      tmp_telem_val = scat_telemetry_eu(33) ! SC_SSPA_INTRM_STG_VLT
      if( .not. ( tmp_telem_val.ge.SC_SSPA_INTRM_STG_VLT_limits(1) .and.
     $            tmp_telem_val.le.SC_SSPA_INTRM_STG_VLT_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_SSPA_INTRM_STG_VLT out of range: ',tmp_telem_val,SC_SSPA_INTRM_STG_VLT_limits

        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif

      tmp_telem_val = scat_telemetry_eu(34) ! SC_SSPA_INPT_STG_VLT
      if( .not. ( tmp_telem_val.ge.SC_SSPA_INPT_STG_VLT_limits(1) .and.
     $            tmp_telem_val.le.SC_SSPA_INPT_STG_VLT_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_SSPA_INPT_STG_VLT out of range: ',tmp_telem_val,SC_SSPA_INPT_STG_VLT_limits

        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif

      tmp_telem_val = scat_telemetry_eu(35) ! SC_SFE_P5V_MON
      if( .not. ( tmp_telem_val.ge.SC_SFE_P5V_MON_limits(1) .and.
     $            tmp_telem_val.le.SC_SFE_P5V_MON_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_SFE_P5V_MON out of range: ',tmp_telem_val,SC_SFE_P5V_MON_limits

        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif

      tmp_telem_val = scat_telemetry_eu(36) ! SC_SCG_P5V_MON
      if( .not. ( tmp_telem_val.ge.SC_SCG_P5V_MON_limits(1) .and.
     $            tmp_telem_val.le.SC_SCG_P5V_MON_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_SCG_P5V_MON out of range: ',tmp_telem_val,SC_SCG_P5V_MON_limits

        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif

      tmp_telem_val = scat_telemetry_eu(37) ! SC_SBE_P5V_MON
      if( .not. ( tmp_telem_val.ge.SC_SBE_P5V_MON_limits(1) .and.
     $            tmp_telem_val.le.SC_SBE_P5V_MON_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_SBE_P5V_MON out of range: ',tmp_telem_val,SC_SBE_P5V_MON_limits

        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif

      tmp_telem_val = scat_telemetry_eu(38) ! SC_SBE_P12V_NSW_MON
      if( .not. ( tmp_telem_val.ge.SC_SBE_P12V_NSW_MON_limits(1) .and.
     $            tmp_telem_val.le.SC_SBE_P12V_NSW_MON_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: SC_SBE_P12V_NSW_MON out of range: ',tmp_telem_val,SC_SBE_P12V_NSW_MON_limits

        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 1 ! Add scat hdr error
        if( i_hardware_status_flag.eq.2 )  i_hardware_status_flag = 3 ! Add scat hdr error

      endif
C--End scat telemetry checks


C--External temp telemetry checks
      tmp_telem_val = external_temperatures_eu(1) ! ET_OMT1_H_PROBE_TMP
      if( .not. ( tmp_telem_val.ge.ET_OMT1_H_PROBE_TMP_limits(1) .and.
     $            tmp_telem_val.le.ET_OMT1_H_PROBE_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: ET_OMT1_H_PROBE_TMP out of range: ',tmp_telem_val,ET_OMT1_H_PROBE_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 1    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif

      tmp_telem_val = external_temperatures_eu(2) ! ET_OMT1_V_PROBE_TMP
      if( .not. ( tmp_telem_val.ge.ET_OMT1_V_PROBE_TMP_limits(1) .and.
     $            tmp_telem_val.le.ET_OMT1_V_PROBE_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: ET_OMT1_V_PROBE_TMP out of range: ',tmp_telem_val,ET_OMT1_V_PROBE_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 1    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif

      tmp_telem_val = external_temperatures_eu(3) ! ET_OMT2_H_PROBE_TMP
      if( .not. ( tmp_telem_val.ge.ET_OMT2_H_PROBE_TMP_limits(1) .and.
     $            tmp_telem_val.le.ET_OMT2_H_PROBE_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: ET_OMT2_H_PROBE_TMP out of range: ',tmp_telem_val,ET_OMT2_H_PROBE_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 1    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif

      tmp_telem_val = external_temperatures_eu(4) ! ET_OMT2_V_PROBE_TMP
      if( .not. ( tmp_telem_val.ge.ET_OMT2_V_PROBE_TMP_limits(1) .and.
     $            tmp_telem_val.le.ET_OMT2_V_PROBE_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: ET_OMT2_V_PROBE_TMP out of range: ',tmp_telem_val,ET_OMT2_V_PROBE_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 1    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif

      tmp_telem_val = external_temperatures_eu(5) ! ET_OMT3_H_PROBE_TMP
      if( .not. ( tmp_telem_val.ge.ET_OMT3_H_PROBE_TMP_limits(1) .and.
     $            tmp_telem_val.le.ET_OMT3_H_PROBE_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'ET_OMT3_H_PROBE_TMP out of range: ',tmp_telem_val,ET_OMT3_H_PROBE_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 1    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif

      tmp_telem_val = external_temperatures_eu(6) ! ET_OMT3_V_PROBE_TMP
      if( .not. ( tmp_telem_val.ge.ET_OMT3_V_PROBE_TMP_limits(1) .and.
     $            tmp_telem_val.le.ET_OMT3_V_PROBE_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: ET_OMT3_V_PROBE_TMP out of range: ',tmp_telem_val,ET_OMT3_V_PROBE_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 1    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif

      tmp_telem_val = external_temperatures_eu(13) ! ET_DPLX_1H_TMP
      if( .not. ( tmp_telem_val.ge.ET_DPLX_1H_TMP_limits(1) .and.
     $            tmp_telem_val.le.ET_DPLX_1H_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: ET_DPLX_1H_TMP out of range: ',tmp_telem_val,ET_DPLX_1H_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 1    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif

      tmp_telem_val = external_temperatures_eu(14) ! ET_DPLX_1V_TMP
      if( .not. ( tmp_telem_val.ge.ET_DPLX_1V_TMP_limits(1) .and.
     $            tmp_telem_val.le.ET_DPLX_1V_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: ET_DPLX_1V_TMP out of range: ',tmp_telem_val,ET_DPLX_1V_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 1    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif
      
      tmp_telem_val = external_temperatures_eu(15) ! ET_DPLX_2H_TMP
      if( .not. ( tmp_telem_val.ge.ET_DPLX_2H_TMP_limits(1) .and.
     $            tmp_telem_val.le.ET_DPLX_2H_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: ET_DPLX_2H_TMP out of range: ',tmp_telem_val,ET_DPLX_2H_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 1    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif

      tmp_telem_val = external_temperatures_eu(16) ! ET_DPLX_2V_TMP
      if( .not. ( tmp_telem_val.ge.ET_DPLX_2V_TMP_limits(1) .and.
     $            tmp_telem_val.le.ET_DPLX_2V_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: ET_DPLX_2V_TMP out of range: ',tmp_telem_val,ET_DPLX_2V_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 1    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif

      tmp_telem_val = external_temperatures_eu(17) ! ET_DPLX_3H_TMP
      if( .not. ( tmp_telem_val.ge.ET_DPLX_3H_TMP_limits(1) .and.
     $            tmp_telem_val.le.ET_DPLX_3H_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: ET_DPLX_3H_TMP out of range: ',tmp_telem_val,ET_DPLX_3H_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 1    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif

      tmp_telem_val = external_temperatures_eu(18) ! ET_DPLX_3V_TMP
      if( .not. ( tmp_telem_val.ge.ET_DPLX_3V_TMP_limits(1) .and.
     $            tmp_telem_val.le.ET_DPLX_3V_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: ET_DPLX_3V_TMP out of range: ',tmp_telem_val,ET_DPLX_3V_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 1    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif

      tmp_telem_val = external_temperatures_eu(19) ! ET_CPLR_1H_TMP
      if( .not. ( tmp_telem_val.ge.ET_CPLR_1H_TMP_limits(1) .and.
     $            tmp_telem_val.le.ET_CPLR_1H_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: ET_CPLR_1H_TMP out of range: ',tmp_telem_val,ET_CPLR_1H_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 1    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif

      tmp_telem_val = external_temperatures_eu(20) ! ET_CPLR_1V_TMP
      if( .not. ( tmp_telem_val.ge.ET_CPLR_1V_TMP_limits(1) .and.
     $            tmp_telem_val.le.ET_CPLR_1V_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: ET_CPLR_1V_TMP out of range: ',tmp_telem_val,ET_CPLR_1V_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 1    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif

      tmp_telem_val = external_temperatures_eu(21) ! ET_CPLR_2H_TMP
      if( .not. ( tmp_telem_val.ge.ET_CPLR_2H_TMP_limits(1) .and.
     $            tmp_telem_val.le.ET_CPLR_2H_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: ET_CPLR_2H_TMP out of range: ',tmp_telem_val,ET_CPLR_2H_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 1    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif

      tmp_telem_val = external_temperatures_eu(22) ! ET_CPLR_2V_TMP
      if( .not. ( tmp_telem_val.ge.ET_CPLR_2V_TMP_limits(1) .and.
     $            tmp_telem_val.le.ET_CPLR_2V_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: ET_CPLR_2V_TMP out of range: ',tmp_telem_val,ET_CPLR_2V_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 1    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif
      
      tmp_telem_val = external_temperatures_eu(23) ! ET_CPLR_3H_TMP
      if( .not. ( tmp_telem_val.ge.ET_CPLR_3H_TMP_limits(1) .and.
     $            tmp_telem_val.le.ET_CPLR_3H_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: ET_CPLR_3H_TMP out of range: ',tmp_telem_val,ET_CPLR_3H_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 1    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif

      tmp_telem_val = external_temperatures_eu(24) ! ET_CPLR_3V_TMP
      if( .not. ( tmp_telem_val.ge.ET_CPLR_3V_TMP_limits(1) .and.
     $            tmp_telem_val.le.ET_CPLR_3V_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: ET_CPLR_3V_TMP out of range: ',tmp_telem_val,ET_CPLR_3V_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 1    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif

      tmp_telem_val = external_temperatures_eu(7) ! ET_SFE_H1_COAX_TMP
      if( .not. ( tmp_telem_val.ge.ET_SFE_H1_COAX_TMP_limits(1) .and.
     $            tmp_telem_val.le.ET_SFE_H1_COAX_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: ET_SFE_H1_COAX_TMP out of range: ',tmp_telem_val,ET_SFE_H1_COAX_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 2    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif

      tmp_telem_val = external_temperatures_eu(8) ! ET_SFE_H2_COAX_TMP
      if( .not. ( tmp_telem_val.ge.ET_SFE_H2_COAX_TMP_limits(1) .and.
     $            tmp_telem_val.le.ET_SFE_H2_COAX_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: ET_SFE_H2_COAX_TMP out of range: ',tmp_telem_val,ET_SFE_H2_COAX_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 2    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif

      tmp_telem_val = external_temperatures_eu(9) ! ET_SFE_H3_COAX_TMP
      if( .not. ( tmp_telem_val.ge.ET_SFE_H3_COAX_TMP_limits(1) .and.
     $            tmp_telem_val.le.ET_SFE_H3_COAX_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: ET_SFE_H3_COAX_TMP out of range: ',tmp_telem_val,ET_SFE_H3_COAX_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 2    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif
      
      tmp_telem_val = external_temperatures_eu(10) ! ET_SFE_V1_COAX_TMP
      if( .not. ( tmp_telem_val.ge.ET_SFE_V1_COAX_TMP_limits(1) .and.
     $            tmp_telem_val.le.ET_SFE_V1_COAX_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: ET_SFE_V1_COAX_TMP out of range: ',tmp_telem_val,ET_SFE_V1_COAX_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 2    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif

      tmp_telem_val = external_temperatures_eu(11) ! ET_SFE_V2_COAX_TMP
      if( .not. ( tmp_telem_val.ge.ET_SFE_V2_COAX_TMP_limits(1) .and.
     $            tmp_telem_val.le.ET_SFE_V2_COAX_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: ET_SFE_V2_COAX_TMP out of range: ',tmp_telem_val,ET_SFE_V2_COAX_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 2    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif

      tmp_telem_val = external_temperatures_eu(12) ! ET_SFE_V3_COAX_TMP
      if( .not. ( tmp_telem_val.ge.ET_SFE_V3_COAX_TMP_limits(1) .and.
     $            tmp_telem_val.le.ET_SFE_V3_COAX_TMP_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: ET_SFE_V3_COAX_TMP out of range: ',tmp_telem_val,ET_SFE_V3_COAX_TMP_limits

        i_anomalous_temps_flag = 1
        if( calibration_level .ge. 2    )  i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif
C--End external temp telemetry checks


C--Check Cal resistor values...
      tmp_telem_val = icds_engineering_telemetry_eu(11) ! IE_ICDS_TLM_CAL_RES1
      if( .not. ( tmp_telem_val.ge.IE_ICDS_TLM_CAL_RES1_limits(1) .and.
     $            tmp_telem_val.le.IE_ICDS_TLM_CAL_RES1_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: IE_ICDS_TLM_CAL_RES1 out of range: ',tmp_telem_val,IE_ICDS_TLM_CAL_RES1_limits

        i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif

      tmp_telem_val = icds_engineering_telemetry_eu(12) ! IE_ICDS_TLM_CAL_RES2
      if( .not. ( tmp_telem_val.ge.IE_ICDS_TLM_CAL_RES2_limits(1) .and.
     $            tmp_telem_val.le.IE_ICDS_TLM_CAL_RES2_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: IE_ICDS_TLM_CAL_RES2 out of range: ',tmp_telem_val,IE_ICDS_TLM_CAL_RES2_limits

        i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif

      tmp_telem_val = icds_engineering_telemetry_eu(13) ! IE_ICDS_TLM_CAL_RES3
      if( .not. ( tmp_telem_val.ge.IE_ICDS_TLM_CAL_RES3_limits(1) .and.
     $            tmp_telem_val.le.IE_ICDS_TLM_CAL_RES3_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: IE_ICDS_TLM_CAL_RES3 out of range: ',tmp_telem_val,IE_ICDS_TLM_CAL_RES3_limits

        i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif

      tmp_telem_val = icds_engineering_telemetry_eu(14) ! IE_ICDS_TLM_CAL_RES4
      if( .not. ( tmp_telem_val.ge.IE_ICDS_TLM_CAL_RES4_limits(1) .and.
     $            tmp_telem_val.le.IE_ICDS_TLM_CAL_RES4_limits(2) ) ) then

        if( i_dbg_level.ge.my_dbg_level ) write(6,'((a50),3f12.6)') 
     $  'WARNING: IE_ICDS_TLM_CAL_RES4 out of range: ',tmp_telem_val,IE_ICDS_TLM_CAL_RES4_limits

        i_bad_calibration_flag = 1
        if( i_hardware_status_flag.eq.0 )  i_hardware_status_flag = 2 ! add other hdr error
        if( i_hardware_status_flag.eq.1 )  i_hardware_status_flag = 3 ! add other hdr error
      endif
      
      end
      
