C--$Header: /mnt/caretr2/scatsim/CVSROOT/L1A_to_L2/L2_Faraday_cost_function.f,v 1.1 2010/03/25 17:36:12 fore Exp $
c*****************************************************************      
      subroutine L2_faraday_cost_function( obs_HH_lin,
     $                                     obs_HV_lin,
     $                                     obs_VV_lin,
     $                                     faraday_angle,
     $                                     rho_HHVV,
     $                                     f_HVHH,
     $                                     f_HVVV,
     $                                     try_HH_lin,
     $                                     try_VV_lin,
     $                                     J,
     $                                     dJ_d_try_HH,
     $                                     dJ_d_try_VV   )
c*****************************************************************      
c**   
c**   FILE NAME:   L2_faraday_cost_function.f
c**   
c**   DATE WRITTEN: 3/19/2010
c**   
c**   PROGRAMMER: Alex Fore
c**   
c**   FUNCTIONAL DESCRIPTION:  Computes the cost function and its
c**    gradient, given an trial value for sigma-0 HH and VV.
c**
c**   ROUTINES CALLED: None
c**   
c**   NOTES: 
c**
c**   UPDATE LOG: 
c**
c**   Date Changed        Reason Changed        CR # and Version #
c**   ------------       ----------------       ------------------
c**     
c*****************************************************************   
      implicit none
c     PARAMETER STATEMENTS
      integer stdout, stderr, stdin
      parameter( stdout = 6, stderr = 0, stdin = 5 )
c     INPUT VARIABLES
      real*8 obs_HH_lin     ! observed sigma-0 HH; linear units
      real*8 obs_HV_lin     ! observed sigma-0 HV; linear units
      real*8 obs_VV_lin     ! observed sigma-0 VV; linear units
      real*8 faraday_angle  ! faraday rotation angle; [deg]
      real*8 rho_HHVV       ! assumed HH-VV correlation
      real*8 f_HVHH         ! assumed co-cx pol sigma0 ratio (hh to hv)
      real*8 f_HVVV         ! assumed co-cx pol sigma0 ratio (vv to hv)
      real*8 try_HH_lin     ! trial value of sigma0 HH
      real*8 try_VV_lin     ! trial value of sigma0 HH
c     OUTPUT VARIABLES
      real*8 J              ! cost function value 
      real*8 dJ_d_try_HH    ! gradient of cost func w.r.t trial HH sigma0
      real*8 dJ_d_try_VV    ! gradient of cost func w.r.t trial VV sigma0
c     LOCAL VARIABLES
      
      real*8 M_HH_lin       ! Model HH measurement
      real*8 M_HV_lin       ! Model HV measurement
      real*8 M_VV_lin       ! Model VV measurement
      
      real*8 d_M_HH_d_try_HH
      real*8 d_M_HH_d_try_VV

      real*8 d_M_HV_d_try_HH
      real*8 d_M_HV_d_try_VV

      real*8 d_M_VV_d_try_HH
      real*8 d_M_VV_d_try_VV
      
      
      real*8 d_J_d_M_HH
      real*8 d_J_d_M_HV
      real*8 d_J_d_M_VV
      
      
      real*8 c2f, c4f
      real*8 s2f, s4f
      real*8 PI
c     DATA STATEMENTS
c     COMMON BLOCKS
c     FUNCTION STATEMENTS
c     PROCESSING STEPS
      
      PI = 4.0d0 * atan( 1.0d0 )
      
      c2f = cos( faraday_angle * PI / 180.0d0 ) ** 2
      s2f = sin( faraday_angle * PI / 180.0d0 ) ** 2
      
      c4f = cos( faraday_angle * PI / 180.0d0 ) ** 4
      s4f = sin( faraday_angle * PI / 180.0d0 ) ** 4
      
C--Construct model measurements for trial value of sigma_HH, sigma_VV
      M_HH_lin = try_HH_lin * c4f + try_VV_lin * s4f -
     $           2.0d0 * rho_HHVV * c2f * s2f * sqrt( try_VV_lin * try_HH_lin )

      M_VV_lin = try_VV_lin * c4f + try_HH_lin * s4f -
     $           2.0d0 * rho_HHVV * c2f * s2f * sqrt( try_VV_lin * try_HH_lin )
      
      M_HV_lin = 0.5d0 * ( try_HH_lin * f_HVHH + try_VV_lin * f_HVVV ) +
     $           ( try_HH_lin + try_VV_lin ) * c2f * s2f + 
     $           2.0d0 * rho_HHVV * c2f * s2f * sqrt( try_VV_lin * try_HH_lin )

C--Compute the cost function value
C      if( obs_HV_lin .gt. 0 ) then
C        J =         ( obs_HH_lin * log( obs_HH_lin / M_HH_lin ) ) ** 2 +
C     $      2.0d0 * ( obs_HV_lin * log( obs_HV_lin / M_HV_lin ) ) ** 2 +
C     $              ( obs_VV_lin * log( obs_VV_lin / M_VV_lin ) ) ** 2
C      else
        J =         ( obs_HH_lin * log( obs_HH_lin / M_HH_lin ) ) ** 2 +
     $              ( obs_VV_lin * log( obs_VV_lin / M_VV_lin ) ) ** 2
C      endif

C--Compute gradients of model measurements w.r.t. trial values of sigma_HH, sigma_VV
      d_M_HH_d_try_HH = c4f - rho_HHVV * c2f * s2f * sqrt( try_VV_lin / try_HH_lin )
      d_M_HH_d_try_VV = s4f - rho_HHVV * c2f * s2f * sqrt( try_HH_lin / try_VV_lin )
      
      d_M_VV_d_try_HH = s4f - rho_HHVV * c2f * s2f * sqrt( try_VV_lin / try_HH_lin )
      d_M_VV_d_try_VV = c4f - rho_HHVV * c2f * s2f * sqrt( try_HH_lin / try_VV_lin )
      
      d_M_HV_d_try_HH = 0.5d0 * f_HVHH + c2f * s2f +
     $                  rho_HHVV * c2f * s2f * sqrt( try_VV_lin / try_HH_lin )
      d_M_HV_d_try_VV = 0.5d0 * f_HVVV + c2f * s2f +
     $                  rho_HHVV * c2f * s2f * sqrt( try_HH_lin / try_VV_lin )

C--Compute gradients of cost function w.r.t. M_HH, M_HV, M_VV
      d_J_d_M_HH = - 2.0d0 * obs_HH_lin ** 2 
     $            * log( obs_HH_lin / M_HH_lin ) / M_HH_lin
      d_J_d_M_HV = - 4.0d0 * obs_HV_lin ** 2 
     $            * log( obs_HV_lin / M_HV_lin ) / M_HV_lin
      d_J_d_M_VV = - 2.0d0 * obs_VV_lin ** 2 
     $            * log( obs_VV_lin / M_VV_lin ) / M_VV_lin

C--Compute gradients of cost function w.r.t. try_HH and try_VV
C      if( obs_HV_lin .gt. 0 ) then
C        dJ_d_try_HH = d_J_d_M_HH * d_M_HH_d_try_HH + 
C     $                d_J_d_M_HV * d_M_HV_d_try_HH + 
C     $                d_J_d_M_VV * d_M_VV_d_try_HH
C
C        dJ_d_try_VV = d_J_d_M_HH * d_M_HH_d_try_VV + 
C     $                d_J_d_M_HV * d_M_HV_d_try_VV + 
C     $                d_J_d_M_VV * d_M_VV_d_try_VV
C      else
        dJ_d_try_HH = d_J_d_M_HH * d_M_HH_d_try_HH + 
     $                d_J_d_M_VV * d_M_VV_d_try_HH

        dJ_d_try_VV = d_J_d_M_HH * d_M_HH_d_try_VV + 
     $                d_J_d_M_VV * d_M_VV_d_try_VV
C      endif
      end
      