C--$Header: /mnt/caretr2/scatsim/CVSROOT/L1A_to_L2/L2_Faraday_rotation_correction.f,v 1.1 2010/03/25 17:36:12 fore Exp $
c*****************************************************************
      subroutine L2_Faraday_rotation_correction( L2_sigma_TOI,
     $                                           L2_var_sigma_TOI,
     $                                           L2_faraday_rotation_angle,
     $                                           rho_HHVV,
     $                                           f_HVHH,
     $                                           f_HVVV,
     $                                           flag_L2_no_far_corr_entered,
     $                                           L2_sigma_TOA,
     $                                           L2_var_sigma_TOA,
     $                                           L2_good_far_corr )
c*****************************************************************      
c**   
c**   FILE NAME: L2_Faraday_rotation_correction.f
c**   
c**   DATE WRITTEN: 3/18/2010
c**   
c**   PROGRAMMER: Alex Fore
c**   
c**   FUNCTIONAL DESCRIPTION:  Driver for BFGS minimization of
c**    Faraday rotation correction cost function problem.
c**    Computes faraday corrected sig-0 HH and VV;
c**    consevation of sig-0 gives us HV for free.
c**    Then it uses Gaussian error propagation to compute
c**    a variance for faraday corrected (hh,hv,vv).
c**
c**   ROUTINES CALLED: 
c**   
c**   NOTES: 
c**
c**   UPDATE LOG:
c**
c**   Date Changed        Reason Changed        CR # and Version #
c**   ------------       ----------------       ------------------
c**   01/31/11           Remove call exit for bad iflag   JMG
c*****************************************************************   
      implicit none
c     PARAMETER STATEMENTS
      integer stdout, stderr, stdin
      parameter( stdout = 6, stderr = 0, stdin = 5 )

      integer*4 m_corr                 ! BFGS stuff
      parameter( m_corr = 3 )          ! BFGS stuff
c     INPUT VARIABLES
      real*8 L2_sigma_TOI(4)           ! top-of-ionosphere sigma0; [linear units]
      real*8 L2_var_sigma_TOI(4)       ! top-of-ionosphere variance of sigma0
      real*8 L2_faraday_rotation_angle ! faraday rotation angle; [deg]
      real*8 rho_HHVV                  ! assumed HH-VV correlation; linear units
      real*8 f_HVHH                    ! assumed co-cx pol sigma0 ratio (hv to hh); linear units
      real*8 f_HVVV                    ! assumed co-cx pol sigma0 ratio (hv to vv); linear units
      logical flag_L2_no_far_corr_entered ! turns off faraday rotation correction
c     OUTPUT VARIABLES
      real*8 L2_sigma_TOA(4)           ! faraday corrected top-of-atmosphere sigma0; linear units
      real*8 L2_var_sigma_TOA(4)       ! variance of corrected TOA sigma0;
      logical L2_good_far_corr         ! tells main software is faraday removal is successful
c     LOCAL VARIABLES
      integer i_chan
      
      real*8 toi_HH_lin                ! top-of-ionosphere sigma-0 HH; linear units
      real*8 toi_HV_lin                ! top-of-ionosphere sigma-0 HV; linear units
      real*8 toi_VV_lin                ! top-of-ionosphere sigma-0 VV; linear units

      real*8 toa_HH_lin                ! faraday corrected sigma-0 HH; linear units
      real*8 toa_HV_lin                ! faraday corrected sigma-0 HV; linear units
      real*8 toa_VV_lin                ! faraday corrected sigma-0 VV; linear units
      
      integer*4 MP,   LP                                ! BFGS stuff
      real*8    GTOL, STPMIN,STPMAX                     ! BFGS stuff
      logical   diagco                                  ! BFGS stuff
      real*8    diag(2)                                 ! BFGS stuff
      integer*4 iprint(2)                               ! BFGS stuff
      real*8    eps                                     ! BFGS stuff
      real*8    xtol                                    ! BFGS stuff
      real*8    workspace(2*(2*m_corr+1)+2*m_corr)      ! BFGS stuff
      integer*4 iflag                                   ! BFGS stuff
      
      logical   l_done
      integer*4 i_iter
      real*8    try_sig0(2)
      real*8    J
      real*8    grad_J(2)

      real*8 c2f, c4f
      real*8 s2f, s4f
      
      real*8 var_sig_toa_hh
      real*8 var_sig_toa_hv
      real*8 var_sig_toa_vv
      real*8 var_sig_toi_hh
      real*8 var_sig_toi_hv
      real*8 var_sig_toi_vv
      
      real*8 A, B, C, D
      real*8 d_toi_HH_d_toa_HH
      real*8 d_toi_HH_d_toa_VV
      real*8 d_toi_HV_d_toa_HH
      real*8 d_toi_HV_d_toa_VV
      real*8 d_toi_VV_d_toa_HH
      real*8 d_toi_VV_d_toa_VV
      
      real*8 PI      
      
c     DATA STATEMENTS
c     COMMON BLOCKS
      integer i_dbg_level
      common /debug/ i_dbg_level

      COMMON /LB3/MP,LP,GTOL,STPMIN,STPMAX ! BFGS stuff
c     FUNCTION STATEMENTS
c     PROCESSING STEPS
      EXTERNAL LB2                         ! BFGS stuff

      PI = 4.0d0 * atan( 1.0d0 )
      
      if( .not.flag_L2_no_far_corr_entered ) then
      
        toi_HH_lin  = L2_sigma_TOI(4)
        toi_HV_lin  = 0.5d0 * ( L2_sigma_TOI(1) + L2_sigma_TOI(3) )
        toi_VV_lin  = L2_sigma_TOI(2)
      
        L2_good_far_corr = .false.
        
        diagco      = .false.
        iprint(1)   = -1
        iprint(2)   = 0
        eps         = 10.0d0 ** ( -12 ) ! stopping criteria
        xtol        = 10.0d0 ** ( -16 ) ! line-search tolerance; should be ~ machine precision
        try_sig0(1) = toi_HH_lin        ! initial guess for sigma-0 HH; linear units
        try_sig0(2) = toi_VV_lin        ! initial guess for sigma-0 VV; linear units
        i_iter      = 0      
        l_done      = .false.
        iflag       = 0 

        do while( .not.l_done )
          i_iter = i_iter + 1
        
          call L2_faraday_cost_function( toi_HH_lin,
     $                                   toi_HV_lin,
     $                                   toi_VV_lin,
     $                                   L2_faraday_rotation_angle,
     $                                   rho_HHVV,
     $                                   f_HVHH,
     $                                   f_HVVV,
     $                                   try_sig0(1),
     $                                   try_sig0(2),
     $                                   J,
     $                                   grad_J(1),
     $                                   grad_J(2)   )
     
          call lbfgs( 2,        ! input; # of variables
     $                m_corr,   ! input; # of corrections used in BFGS update
     $                try_sig0, ! input; trial state vector 
     $                J,        ! output; cost function value @ try_sig0
     $                grad_J,   ! output; gradent of cost function value @ try_sig0
     $                diagco,   ! BFGS control parameter
     $                diag,     ! BFGS control parameter
     $                iprint,   ! controls BFGS debuging output
     $                eps,      ! minimization terminates when ||G|| < EPS max(1,||X||)
     $                xtol,     ! line-search tolerance?
     $                workspace,! working arrays used by BFGS
     $                iflag )   ! iflag <0 => error
                                ! iflag==0 => terminated without errors
                                ! iflag==1 => user must recompute J and grad_J and iterate
                              
          if( iflag .lt. 0 ) then
            write(stdout,'(a,I6)') 'WARNING: L2_faraday_rotation_correction: lbfgs returned iflag: ', iflag
            write(stdout,'(a50,f16.6)') 'L2_faraday_rotation_angle [deg]: ', L2_faraday_rotation_angle
            write(stdout,'(a50,4f16.6)') 'L2_sigma_TOI [db]: ', (10*log10(L2_sigma_TOI(i_chan)),i_chan=1,4)
            write(stdout,'(a50,4d16.6)') 'L2_sigma_TOI [lin]: ', L2_sigma_TOI
            write(stdout,'(a50,2d16.6,2f16.6)') 'try_sig0 (HH,VV) [lin,dB]: ', try_sig0, (10*log10(try_sig0(i_chan)),i_chan=1,4)
            write(stdout,'(a50,3f16.6)') 'J, grad_J (HH,VV): ', J, grad_J
          
            L2_good_far_corr = .false.
            l_done           = .true.
          elseif( iflag .eq. 0 ) then
            L2_good_far_corr = .true.
            l_done           = .true.  ! minimization has terminated; break out of do while loop.
          elseif( iflag .eq. 1 ) then
            l_done = .false. ! need to keep iterating...
          else
            write(stdout,*) 'WARNING: L2_faraday_rotation_correction: Unexpected value of iflag: ', iflag
          endif ! if( iflag .lt. 0 ) then
        
          if( i_iter .gt. 100 ) then
            write(stdout,*) 'ERROR: L2_faraday_rotation_correction: Exceeded 100 iterations in BFGS minimization'
            stop
          endif ! if( i_iter .gt. 100 )
        enddo ! do while( .not.l_done )
      
        if( L2_good_far_corr ) then
      
C--Set output variables      
          toa_HH_lin = try_sig0(1) ! top-of-atmosphere sigma-0 HH
          toa_VV_lin = try_sig0(2) ! top-of-atmosphere sigma-0 VV
      
C--Conservation of total sigma-0 will give us the cross-pol channel.
          toa_HV_lin = 0.5d0 * ( toi_HH_lin + 2.0d0 * toi_HV_lin + toi_VV_lin 
     $                         - toa_HH_lin - toa_VV_lin )
      
C--Compute variance propagation through the Faraday rotation model
          var_sig_toi_hh = L2_var_sigma_TOI(4) ! variance of top-of-ionosphere sig-0; hh
          var_sig_toi_hv = 0.25d0 * ( L2_var_sigma_TOI(1) + L2_var_sigma_TOI(3) ) ! hv;
          var_sig_toi_vv = L2_var_sigma_TOI(2) ! variance of top-of-ionosphere sig-0; vv

C--From model for Faraday rotation: (sig_M_*) is model meas in presence of farday rotation;
C-- try_HH_lin is the TOA sigma0, which we have just found via the solution of a 
C-- 2-d unconstraned optimization problem.
C
C--  var(sig_M_hh) = [ c4f - rho_HHVV * c2f * s2f * sqrt( toa_VV_lin / toa_HH_lin ) ] ** 2 * var(toa_HH_lin) +
C--                  [ s4f - rho_HHVV * c2f * s2f * sqrt( toa_HH_lin / toa_VV_lin ) ] ** 2 * var(toa_VV_lin)
C
C--  var(sig_M_vv) = [ s4f - rho_HHVV * c2f * s2f * sqrt( toa_VV_lin / toa_HH_lin ) ] ** 2 * var(toa_HH_lin) +
C--                  [ c4f - rho_HHVV * c2f * s2f * sqrt( toa_HH_lin / toa_VV_lin ) ] ** 2 * var(toa_VV_lin)
C
C    We already know the stuff on the LHS (variance of TOI sig-0).
C    thus we solve for the RHS (var_sig_t_hh, var_sig_t_vv) via a 2x2 matrix inversion...

          c2f = cos( L2_faraday_rotation_angle * PI / 180.0d0 ) ** 2
          s2f = sin( L2_faraday_rotation_angle * PI / 180.0d0 ) ** 2
      
          c4f = cos( L2_faraday_rotation_angle * PI / 180.0d0 ) ** 4
          s4f = sin( L2_faraday_rotation_angle * PI / 180.0d0 ) ** 4      

C--Compute gradients of model measurements w.r.t. trial values of sigma_HH, sigma_VV
          d_toi_HH_d_toa_HH = c4f - rho_HHVV * c2f * s2f * sqrt( toa_VV_lin / toa_HH_lin )
          d_toi_HH_d_toa_VV = s4f - rho_HHVV * c2f * s2f * sqrt( toa_HH_lin / toa_VV_lin )
      
          d_toi_VV_d_toa_HH = s4f - rho_HHVV * c2f * s2f * sqrt( toa_VV_lin / toa_HH_lin )
          d_toi_VV_d_toa_VV = c4f - rho_HHVV * c2f * s2f * sqrt( toa_HH_lin / toa_VV_lin )
      
          d_toi_HV_d_toa_HH = 0.5d0 * f_HVHH + c2f * s2f +
     $                      rho_HHVV * c2f * s2f * sqrt( toa_VV_lin / toa_HH_lin  )
          d_toi_HV_d_toa_VV = 0.5d0 * f_HVVV + c2f * s2f +
     $                      rho_HHVV * c2f * s2f * sqrt( toa_HH_lin / toa_VV_lin )


          A = d_toi_HH_d_toa_HH ** 2
          B = d_toi_HH_d_toa_VV ** 2
          C = d_toi_VV_d_toa_HH ** 2
          D = d_toi_VV_d_toa_VV ** 2

          var_sig_toa_hh = (   D * var_sig_toi_hh - B * var_sig_toi_vv ) / ( A * D - B * C )
          var_sig_toa_vv = ( - C * var_sig_toi_hh + A * var_sig_toi_vv ) / ( A * D - B * C )
      
C--This is derived from conservation of total sigma-0 through faraday rotation.
C--I considered toa_HH and toa_VV to be the independant variables for this 
C--computation. ( toi_{HH,HV,VV} are functions of toa_{HH,VV} for doing
C--the Gaussian error propagation calculation. )
          var_sig_toi_hv = 0.25d0 * ( 
     $                     d_toi_HH_d_toa_HH ** 2 * var_sig_toa_hh
     $                   + d_toi_HH_d_toa_VV ** 2 * var_sig_toa_vv
     $                   + d_toi_VV_d_toa_HH ** 2 * var_sig_toa_hh
     $                   + d_toi_VV_d_toa_VV ** 2 * var_sig_toa_vv
     $                   + d_toi_HV_d_toa_HH ** 2 * var_sig_toa_hh * 4.0d0
     $                   + d_toi_HV_d_toa_VV ** 2 * var_sig_toa_vv * 4.0d0
     $                   + var_sig_toa_hh 
     $                   + var_sig_toa_vv )
        
          if( var_sig_toi_hv .le. 0.0d0 ) then
            write(6,'(a,d16.6)')    'ERROR: L2_Faraday_rotation_correction: negative hv variance!', var_sig_toi_hv
            write(6,'(a50,f16.6)')  'L2_faraday_rotation_angle [deg]: ', L2_faraday_rotation_angle
            write(6,'(a50,4f16.6)') 'A,B,C,D: ', A, B, C, D
            write(6,'(a50,3f16.6)') '1/(AD-BC), (D-B)/(AD-BC), (A-C)/(AD-BC),: ', 
     $                               1/(A*D-B*C), (D-B)/(A*D-B*C), (A-C)/(A*D-B*C)
          endif ! if( var_sig_toi_hv .le. 0.0d0 ) then
        
C--Copy to output arrays      
          L2_sigma_TOA(1)     = toa_HV_lin
          L2_sigma_TOA(2)     = toa_VV_lin
          L2_sigma_TOA(3)     = toa_HV_lin
          L2_sigma_TOA(4)     = toa_HH_lin
      
          L2_var_sigma_TOA(1) = var_sig_toi_hv
          L2_var_sigma_TOA(2) = var_sig_toi_vv
          L2_var_sigma_TOA(3) = var_sig_toi_hv
          L2_var_sigma_TOA(4) = var_sig_toi_hh
          
        else ! if( L2_good_far_corr ) then

C--Faraday correction was not sucessful; copy TOI to TOA, pass logical back to main s/w
C-- so it can set the corresponding L2 flag.
          L2_sigma_TOA(1)     = L2_sigma_TOI(1)
          L2_sigma_TOA(2)     = L2_sigma_TOI(2)
          L2_sigma_TOA(3)     = L2_sigma_TOI(3)
          L2_sigma_TOA(4)     = L2_sigma_TOI(4)
      
          L2_var_sigma_TOA(1) = L2_var_sigma_TOI(1)
          L2_var_sigma_TOA(2) = L2_var_sigma_TOI(2)
          L2_var_sigma_TOA(3) = L2_var_sigma_TOI(3)
          L2_var_sigma_TOA(4) = L2_var_sigma_TOI(4)
        endif

      else ! if( .not.flag_L2_no_far_corr_entered ) then

        L2_sigma_TOA(1)     = L2_sigma_TOI(1)
        L2_sigma_TOA(2)     = L2_sigma_TOI(2)
        L2_sigma_TOA(3)     = L2_sigma_TOI(3)
        L2_sigma_TOA(4)     = L2_sigma_TOI(4)
      
        L2_var_sigma_TOA(1) = L2_var_sigma_TOI(1)
        L2_var_sigma_TOA(2) = L2_var_sigma_TOI(2)
        L2_var_sigma_TOA(3) = L2_var_sigma_TOI(3)
        L2_var_sigma_TOA(4) = L2_var_sigma_TOI(4)
      endif
      end
