C--$Header: /u/caretr2/scatsim/CVSROOT/L1A_to_L2/L2_retrive_wind_dualpol.f,v 1.8 2012/04/27 16:33:20 fore Exp $
c*****************************************************************      
      subroutine L2_retrive_wind_dualpol( incang, 
     $                                    relazi_in,
     $                                    i_sigwave,
     $                                    obs_s0_vv,
     $                                    obs_s0_hh,
     $                                    kpc_vv,
     $                                    kpc_hh,
     $                                    i_beam,
     $                                    wind_speed,
     $                                    model_speed )
c*****************************************************************      
c**   
c**   FILE NAME: L2_retrive_wind_dualpol.f
c**   
c**   DATE WRITTEN: 10/10/2011
c**   
c**   PROGRAMMER: Alex Fore, GMF development by W. Tang and S. Yueh
c**   
c**   FUNCTIONAL DESCRIPTION:  Searches speed space for
c**    the speed which gives max objective function value
c**
c**   ROUTINES CALLED: 
c**   
c**   NOTES: Based on QuikSCAT processing code Line maximize subroutines.
c**
c**   UPDATE LOG: Updated to work with SY subroutines 1/24/2012
c**
c**   Date Changed        Reason Changed        CR # and Version #
c**   ------------       ----------------       ------------------
c**     
c*****************************************************************   
      implicit none
c     PARAMETER STATEMENTS
      integer stdout, stderr, stdin
      parameter( stdout = 6, stderr = 0, stdin = 5 )
      
      real*8   COARSE_DSPD     ! coarse speed search spacing
      real*8   DELTA_SPD       ! refined search spacing
      real*8   UPPER_SPD_BOUND ! maximum speed in GMF 
      real*8   LOWER_SPD_BOUND ! mimimum speed in GMF (should be 0!)
      integer  MAX_NUM_AMBIGS  ! max number of ambigs (arrary size)
      parameter( COARSE_DSPD=0.2, DELTA_SPD=0.1, LOWER_SPD_BOUND=0, UPPER_SPD_BOUND=50, MAX_NUM_AMBIGS=10 )
      
c     INPUT VARIABLES      
      real*8   incang      ! incidence angle in degrees
      real*8   relazi_in   ! relative azi angle (cell azimuth - wind direction) in degrees
      real*8   obs_s0_vv   ! L2 sigma0 vv [linear units]
      real*8   obs_s0_hh   ! L2 sigma0 hh [linear units]
      real*8   kpc_vv
      real*8   kpc_hh      
      real*8   model_speed ! NCEP wind speed [ambig removal]
      integer  i_sigwave   ! significant wave height flag
      integer  i_beam      ! which beam is this? (1,2,3)
c     OUTPUT VARIABLES
      real*8   wind_speed  ! retrieved wind speed in m/s
c     LOCAL VARIABLES
      real*8   kpc(3)      ! vv hh hv
      real*8   obs_s0(3)   ! vv hh hv

      real*8   coarse_obj(0:250)  ! must be >= than nint( 50 / COARSE_DSPD )
      real*8   amb_obj(MAX_NUM_AMBIGS)
      real*8   amb_spd(MAX_NUM_AMBIGS)
      real*8   minus_s0(3)
      real*8   center_s0(3)
      real*8   plus_s0(3)
      real*8   minus_spd
      real*8   center_spd
      real*8   plus_spd
      real*8   minus_obj
      real*8   center_obj
      real*8   plus_obj
      real*8   offset
      real*8   max_obj
      real*8   diff1, diff2
      real*8   spd_diff, obj_best
      real*8   relazi
      
      integer  num_ambigs
      integer  ispd
      integer  iamb
      integer  best_amb_spd, best_amb_obj
      
      integer  speed_iterations
            
      logical  do_center
      logical  do_plus
      logical  do_minus
      logical  do_interp      
      logical  found
      logical  is_local_max
  
c     DATA STATEMENTS

c     COMMON BLOCKS
      integer i_dbg_level
      common /debug/ i_dbg_level

c     FUNCTION STATEMENTS
c     PROCESSING STEPS
      
      ! switch wind direction conventions to use S. Yueh's subroutines
      relazi = relazi_in + 180.0
      if( relazi .ge. 180 ) relazi = relazi - 360
      if( relazi .lt.-180 ) relazi = relazi + 360
      
      obs_s0(1) = obs_s0_vv
      obs_s0(2) = obs_s0_hh
      
      kpc(1)    = kpc_vv
      kpc(2)    = kpc_hh
      
      num_ambigs = 0 
      do iamb=1,MAX_NUM_AMBIGS
        amb_obj(iamb) = -999.0
        amb_spd(iamb) = -999.0
      enddo
      
      do ispd=0,nint(UPPER_SPD_BOUND/COARSE_DSPD) ! coarse speed search; sample obj at every 1 m/s
        center_spd = real(ispd) * COARSE_DSPD
        call aqradargmf_op(incang,center_spd,relazi,center_s0)
        call obj_func(center_s0,model_speed,center_spd,obs_s0,kpc,coarse_obj(ispd))
      enddo
      
      ! refine ambigs found during coarse speed search
      do ispd=0,nint(UPPER_SPD_BOUND/COARSE_DSPD) 

        is_local_max = .false.
        if( ispd.eq.0 .and. coarse_obj(ispd).gt.coarse_obj(ispd+1) )then
          is_local_max = .true.

        elseif( ispd.eq.int(UPPER_SPD_BOUND) .and. 
     $          coarse_obj(ispd).gt.coarse_obj(ispd-1) ) then
          is_local_max = .true.       

        elseif( coarse_obj(ispd).gt.coarse_obj(ispd+1) .and. 
     $          coarse_obj(ispd).gt.coarse_obj(ispd-1) ) then
          is_local_max = .true.   
        endif
        
        if( .not.is_local_max ) cycle

        num_ambigs          = num_ambigs + 1
        amb_obj(num_ambigs) = coarse_obj(ispd)
        amb_spd(num_ambigs) = real(ispd) * COARSE_DSPD
        
        if( num_ambigs.gt.MAX_NUM_AMBIGS ) then
          write(6,*) 'Increase MAX_NUM_AMBIGS in L2_retrieve_wind_dualpol!'
        endif
        
C        write(6,*) 'iamb, obj, spd, modelspd: ', 
C     $   num_ambigs, amb_obj(num_ambigs), amb_spd(num_ambigs), model_speed
        
        !-Refined search for a local maxima about the coarse maxima------------
        !-Kick it off the boundaries (it will go back)-------------------------
        center_spd = amb_spd(num_ambigs)
        if( center_spd.le.LOWER_SPD_BOUND ) center_spd = center_spd + DELTA_SPD
        if( center_spd.ge.UPPER_SPD_BOUND ) center_spd = center_spd - DELTA_SPD

        do_interp        = .true.
        do_minus         = .true.
        do_center        = .true.
        do_plus          = .true.
        found            = .false.
        speed_iterations = 0
        
        do while( .not.found .and. speed_iterations.lt.500 ) 
        
          if( do_minus ) then
            minus_spd = center_spd - DELTA_SPD
            if( minus_spd .lt. LOWER_SPD_BOUND ) minus_spd = LOWER_SPD_BOUND
            call aqradargmf_op(incang,minus_spd,relazi,minus_s0)
            call obj_func(minus_s0,model_speed,minus_spd,obs_s0,kpc,minus_obj)
          endif
        
          if( do_center ) then
            center_spd = center_spd
            call aqradargmf_op(incang,center_spd,relazi,center_s0)
            call obj_func(center_s0,model_speed,center_spd,obs_s0,kpc,center_obj)
          endif
        
          if( do_plus ) then
            plus_spd = center_spd + DELTA_SPD
            if( plus_spd .gt. UPPER_SPD_BOUND ) plus_spd = UPPER_SPD_BOUND
            call aqradargmf_op(incang,plus_spd,relazi,plus_s0)
            call obj_func(plus_s0,model_speed,plus_spd,obs_s0,kpc,plus_obj)
          endif
        
          do_interp = .true.        
          found     = .true.
        
          if( plus_obj .gt. center_obj .and. plus_obj .gt. minus_obj ) then 
            found      = .false.
            minus_obj  = center_obj
            minus_spd  = center_spd
            center_obj = plus_obj
            center_spd = plus_spd
          
            if( plus_spd .ge. UPPER_SPD_BOUND ) then
              center_spd = UPPER_SPD_BOUND
              found      = .true.
              do_interp  = .false.
            endif
          
            do_minus  = .false.
            do_center = .false.
            do_plus   = .true.
          endif

          if( minus_obj .gt. center_obj .and. minus_obj .gt. plus_obj ) then
            found      = .false.
            plus_obj   = center_obj
            plus_spd   = center_spd
            center_obj = minus_obj
            center_spd = minus_spd
          
            if( minus_spd .le. LOWER_SPD_BOUND ) then
              center_spd = LOWER_SPD_BOUND
              found      = .true.
              do_interp  = .false.            
            endif
          
            do_minus  = .true.
            do_center = .false.
            do_plus   = .false.
          endif
        
          speed_iterations = speed_iterations + 1
        enddo
      
        diff1 = plus_obj - minus_obj
        diff2 = plus_obj + minus_obj - 2 * center_obj
      
        if( do_interp .and. diff2 .ne. 0.0 ) then
          offset              = - 0.5 * diff1 / diff2
          amb_obj(num_ambigs) = center_obj - ( diff1 * diff1 / diff2 ) / 8.0;
          amb_spd(num_ambigs) = center_spd + offset * DELTA_SPD
        
          ! Keep in bounds of GMF table
          if( amb_spd(num_ambigs).gt.UPPER_SPD_BOUND ) amb_spd(num_ambigs) = UPPER_SPD_BOUND
          if( amb_spd(num_ambigs).lt.LOWER_SPD_BOUND ) amb_spd(num_ambigs) = LOWER_SPD_BOUND
        else
          amb_spd(num_ambigs) = center_spd
          amb_obj(num_ambigs) = center_obj
        endif
        
C        write(6,*) 'iamb, obj, spd, modelspd: ', 
C     $   num_ambigs, amb_obj(num_ambigs), amb_spd(num_ambigs), model_speed
        
      enddo
      
      ! Pick closest ambig to NCEP wind speed
      best_amb_spd = 1
      best_amb_obj = 1
      spd_diff     = abs( amb_spd(1) - model_speed )
      obj_best     = amb_obj(1)
      
      if( num_ambigs.gt.1 ) then
        do iamb=2,num_ambigs
          if( abs( amb_spd(iamb) - model_speed ) .lt. spd_diff ) then
            best_amb_spd = iamb
            spd_diff     = abs( amb_spd(iamb) - model_speed )
          endif
          if( amb_obj(iamb) .gt. obj_best ) then
            best_amb_obj = iamb
            obj_best     = amb_obj(iamb)
          endif
        enddo
      endif
      
      wind_speed = amb_spd(best_amb_spd)
      !wind_speed = amb_spd(best_amb_obj)
      end
 
c*****************************************************************      
      subroutine obj_func( model_sigma0, ! gmf sigma0
     $                     model_speed,  ! NCEP wind
     $                     wind_speed,   ! scat speed
     $                     obs_s0,       ! obs sigma0 [linear]
     $                     kpc,          ! nomalized std of obs sigma0
     $                     obj )         ! obj function value
c*****************************************************************
c**   
c**   FILE NAME: L2_retrive_wind_dualpol.f
c**   
c**   DATE WRITTEN: 1/24/2012
c**   
c**   PROGRAMMER: Alex Fore
c**   
c**   FUNCTIONAL DESCRIPTION:  Formulates the objective function.
c**
c**   ROUTINES CALLED: 
c**   
c**   NOTES: 
c**
c**   UPDATE LOG:
c**
c**   Date Changed        Reason Changed        CR # and Version #
c**   ------------       ----------------       ------------------
c**     
c*****************************************************************   
      implicit none
c     PARAMETER STATEMENTS
      integer stdout, stderr, stdin
      parameter( stdout = 6, stderr = 0, stdin = 5 )
c     INPUT VARIABLES
      real*8        model_sigma0(3) ! VV HH VH [linear units]
      real*8        model_speed
      real*8        wind_speed
      real*8        obs_s0(3)
      real*8        kpc(3)
c     OUTPUT VARIABLES
      real*8        obj
c     LOCAL VARIABLES
c     DATA STATEMENTS
c     COMMON BLOCKS
c     FUNCTION STATEMENTS
c     PROCESSING STEPS
      
      obj = -(obs_s0(1)-model_sigma0(1))**2 / ( obs_s0(1) * kpc(1) )**2
     $      -(obs_s0(2)-model_sigma0(2))**2 / ( obs_s0(2) * kpc(2) )**2
      
      end

