! Module and Reader for Aquarius L2 Uncertainty Files
! Thomas Meissner
! Remote Sensing Systems


module SSS_uncertainties_module
implicit none
save
public


! files for retrievals with uncertainties
! to be specified by user
character(len=200)  :: L2_dir_pert = '/disk02/joel/ocssw/test/l2gen_aquarius/PERT_L2_'


 
! Parameters

! max number of block in l1a file = (5872 sec/orbit)*(1.2 overlap)/(1.44 sec/block) + overlapping tails
integer(4), parameter                                           :: max_cyc= 5000 

! number of radiometers (i.e., horns)  (1=inner, 2=middle, 3=outer)
integer(4), parameter                                           :: n_rad=3       

! place holder for missing value
real(4),    parameter                                           :: missing_val4=-9999.0
real(8),    parameter                                           :: missing_val8=-9999.d0

! L2 Variables

! orbit (rev) number
integer(4)                                                      ::  iorbit

! Time 
! observation time sonce 01 January 2000 00Z [sec]
!real(8), dimension(max_cyc)                                     :: time_utc_2000

integer(4), parameter                                           :: nerrors = 20

contains


subroutine read_SSS_uncertainties(jorbit, L2_file_pert, sec_of_day, time_utc_2000, &
     cellat, cellon, &
     dSSS, dlambda, dSSS_dlambda, dSSS_tot, ierror)
implicit none

integer(4), intent(in)      ::      jorbit              ! rev #
character(200),intent(in)   :: L2_file_pert

! sec of day
real(4), intent(out)        ::      sec_of_day(max_cyc)
real(8), intent(out)        ::      time_utc_2000(max_cyc)

! geolocation
! Earth point boresight latitude, longitude
real(4), intent(out), dimension(n_rad, max_cyc)   :: cellat, cellon

! arrays of perturbed retrievals and uncertainties

real(4), intent(out), dimension(nerrors, n_rad, max_cyc)  :: dSSS, dlambda, dSSS_dlambda
!                                                     TYPE
! 1 = NEDT v-pol                                      random
! 2 = NEDT h-pol                                      random
! 3 = NEDT S3                                         random  
! 4 = EIA (deviation from plumb)                      random
! 5 = wind speed (random component)                   random
! 6 = wind direction                                  random
! 7 = wind speed (systematic component)               systematic
! 8 = SST                                             systematic
! 9 = IU coupling                                     systematic 
!10 = TA galaxy  (I/2)                                systematic
!11 = TA moon    (I/2)                                systematic  
!12 = TB surf v-pol land    contamination             systematic
!13 = TB surf v-pol sea-ice contamination             systematic
!14 = RFI level on SSS                                systematic
!15 - 19: sparse

real(4), intent(out), dimension(3, n_rad, max_cyc)         :: dSSS_tot
! 1= random     component
! 2= systematic component
! 3= total RMS

integer(4), intent(out)     ::      ierror              ! =0. normal return 
                                                        ! =1. L2 file does not exist

logical(4)                  ::      lexist
integer(4)                  ::       iunit               ! input unit


    iunit = 5

    inquire(file=L2_file_pert,exist=lexist)
        
    if (.not.lexist) then ! L2 uncertainty file does not exist
        ierror=1
        return
    endif
    
    ierror=0
    open(unit=iunit, file=L2_file_pert, status='old', form='unformatted', access='stream', action='read')

    read(iunit)  iorbit
    if (iorbit /= jorbit) then
        write(*,*) ' mismatch in orbit number when reading L2 pert file. pgm stopped.'
        write(*,*) jorbit, iorbit
        call exit(1)
    endif
    

    ! time
    read(iunit)     time_utc_2000
    read(iunit)     sec_of_day  

    ! geolocation
    read(iunit)     cellat
    read(iunit)     cellon

    ! array of error in parameters
    read(iunit)     dlambda

    ! array of sensitivies of SSS to the parameters
    read(iunit)     dSSS_dlambda

    ! array of uncertainties
    read(iunit)     dSSS
    read(iunit)     dSSS_tot

    close(iunit)

return
end subroutine read_SSS_uncertainties

end module SSS_uncertainties_module

