module wind_speed_retrieval_module

use static_data_module
use salinity_module

implicit none
private
public :: fd_wspd
public :: fd_est_error
save

! global variable accessed by chi2 functions 
real(4)                                     :: ysigma0_VV, ysigma0_HH, ydew_V, ydew_H, ywin_0, yphir  
real(4)                                     :: ywt_HH, ywt_H, ywt_W
integer(4)                                  :: yirad

real(4)                                     :: ysst, ysss
real(4), dimension(2)                       :: ydelta_1, ydelta_adj   
! added in June 2013, as dEW is now SST/SSS dependent

real(4), parameter, dimension(n_rad)        :: tht_ref = (/29.36, 38.44, 46.29/)
real(4), parameter                          :: sst_ref=20.0, sss_ref=35.0, freq_aq=1.413, wcasp=11.0
real(4), parameter                          :: kappascale=1.4 ! empirical scale factor for delta form factor

type(static_data_struct)                    :: stdata

contains

subroutine fd_wspd(irad, phir, wspd_0, dew, xsigma0_VV, xsigma0_HH, &
     wspd_bg, sst, sss, staticdata, &
     wspd_scat, chisq_scat, iflag_wspd_scat)
! scat HH
! December 27, 2012: substitute the NUMREC BRENT with my Golden Section minimum search
  use static_data_module

  implicit none

  integer(4), intent(in)                  :: irad
  real(4), intent(in)                     :: phir     ! relative wind direction
  real(4), intent(in)                     :: wspd_0   ! first guess wind speed (use NCEP)
  real(4), intent(in)                     :: wspd_bg
  real(4), intent(in)                     :: dew(2)
  real(4), intent(in)                     :: xsigma0_VV, xsigma0_HH
  real(4), intent(in)                     :: sst, sss

  type(static_data_struct), intent(in)    :: staticdata

  real(4), intent(out)                    :: wspd_scat, chisq_scat
  integer(4), intent(out)                 :: iflag_wspd_scat

  integer(4), parameter           :: mmax=4
  real(4), dimension(mmax)        :: wa, wb, wc
  real(4), parameter              :: bw=0.5  
  integer(4)                      :: iw, nmax, min_count
  real(4)                         :: w0, w1, w2, y0, y1, y2

  real(4)                         :: dsigma0, ddew, est_err(3), dw

  real(4), dimension(mmax)        :: x_min, y_min
  integer(4), dimension(mmax)     :: m_search

  real(4), parameter              :: very_large_number = 1.0E15
  real(4)                         :: dmin, diff
  integer(4)                      :: imin, jmin

  real(4), parameter              :: eps=1.0E-2
  real(4), parameter              :: eps_brent=5.0E-5
  real(4), parameter              :: wmin=0.0, wmax=30.0
  integer(4),save  :: istart=1

  real(8)                         :: machep, x_min_dbl, chi2_min, snd_derv_bnd, glomin
  real(4), dimension(2)           :: ze0, ze0ref, zdelta_adj
  real(4)                         :: xtht
  
  if(istart.eq.1) then
     istart=0
     stdata = staticdata
  endif

  ! save global variables to be accessed through chi2_HH
  ysigma0_VV = xsigma0_VV
  ysigma0_HH = xsigma0_HH
        
  yphir      = phir
  yirad      = irad 
        
  ywin_0     = wspd_0

  ysst       = sst
  ysss       = sss

  if (loc(wspd_bg).ne.0) then
   ! added in June 2013, as dEW is now dependent on SST/SSS
     ysst=sst
     ysss=sss_ref ! I decided yo use 35 as standard value. 

     ! SST dependent form factor
     xtht=tht_ref(irad)
     call fdem0_meissner_wentz_salinity(freq_aq,xtht,sst,    sss_ref, ze0) 
     call fdem0_meissner_wentz_salinity(freq_aq,xtht,sst_ref,sss_ref, ze0ref) 
     ydelta_1 = ze0/ze0ref
    
     ! adjustment for SST dependent form factor
     call find_delta_fac_adjust (irad, sst, zdelta_adj)
     ydelta_adj = zdelta_adj
  endif
  
  call fd_est_error (irad, wspd_0,  est_err)

  dsigma0 = est_err(1)
  ddew    = est_err(2)
  dw      = est_err(3)
  ywt_HH  = 1.0/(dsigma0**2) ! channel weighted by expected variance
  ywt_W   = 1.0/(dw**2)      ! weight for background field

  if (loc(dew).ne.0) then
     ydew_V     = dew(1)
     ydew_H     = dew(2)
     ywt_H      = 1.0/(ddew**2) ! rad channel H weighted by expected variance 
  endif


!  if (irad.eq.1) write(*,*) 'find minima'
  min_count=0
  nmax = nint(wmax/bw)
  do iw=1,nmax-1
     w0 = wmin + float(iw-1)*bw
     w1 = wmin + float(iw)*bw  
     w2 = wmin + float(iw+1)*bw   
     if (w0<0.0) w0=0.0
     if (loc(dew).ne.0) then
        y0 = chi2_HHH(w0)
        y1 = chi2_HHH(w1)
        y2 = chi2_HHH(w2)
     else
        y0 = chi2_HH(w0)
        y1 = chi2_HH(w1)
        y2 = chi2_HH(w2)
     endif
     if (y1<=y0 .and. y1<y2) then ! found minimum with brackets
        min_count=min_count+1
        wa(min_count)= w0
        wb(min_count)= w1
        wc(min_count)= w2
     endif
  enddo
  
!  if (irad.eq.1) write(*,*) 'min_count: ', min_count 
  if (min_count==0) then ! function is either monotonously rising or falling
     if (loc(dew).ne.0) then
        y0 = chi2_HHH(wmin)
        y1 = chi2_HHH(wmax)
     else
        y0 = chi2_HH(wmin)
        y1 = chi2_HH(wmax)
     endif
     if (y0<y1) then
        x_min(1)=wmin
        y_min(1)=y0
     else
        x_min(1)=wmax
        y_min(1)=y1
     endif
     iflag_wspd_scat=0 ! o.k.
     wspd_scat=x_min(1)
     chisq_scat=y_min(1) 
     return   
  endif
  
  if (min_count>mmax) then
     write(*,*)' chi2 has more than mmax=4 minima'
     write(*,*) irad
     write(*,*) wspd_0, phir
     write(*,*) min_count
     write(*,*) wb(1:min_count)
     iflag_wspd_scat=1 ! no scat wspd
     wspd_scat=missing_val
     chisq_scat=missing_val
     return
  endif

  !  if (irad.eq.1) write(*,*) 'minimum search'
  ! improve on minima using Golden Section             
  ! if more than one minimum choose the one that is closest to NCEP 
  jmin=0
  dmin=very_large_number
  machep = epsilon(1d0)
  snd_derv_bnd = 100.0
  do imin =1,min_count
        
     if (loc(dew).ne.0) then
        call minimum_search(wa(imin),wb(imin),wc(imin),chi2_HHH,eps,x_min(imin),y_min(imin),m_search(imin))

        ! brent
        x_min_dbl = dble(x_min(imin))
        chi2_min = glomin( dble(wa(imin)), dble(wc(imin)), x_min_dbl, snd_derv_bnd, machep, &
             dble(eps_brent), dble(eps_brent), chi2_HHH_dbl, x_min_dbl)
!        write(*,*) 'HHH_winds: ', x_min_dbl, chi2_min

        x_min(imin) = x_min_dbl
        chisq_scat = chi2_min
     else
        call minimum_search(wa(imin),wb(imin),wc(imin),chi2_HH, eps,x_min(imin),y_min(imin),m_search(imin))
     endif

     if (m_search(imin)/=0) cycle            
       
     diff  = abs(x_min(imin)-ywin_0)
     if (diff < dmin) then
        dmin=diff
        jmin=imin
     endif
  enddo
        
  if (jmin==0) then
     iflag_wspd_scat=1 ! no scat wspd
     write(*,*) ' Golden Section method did not find minimum'
     wspd_scat=missing_val
     chisq_scat=missing_val
  else 
     iflag_wspd_scat=0 ! o.k.
     wspd_scat=x_min(jmin)
     chisq_scat=y_min(jmin)
  endif
        
  return
end subroutine fd_wspd


subroutine fd_est_error (irad, wspd,  est_err)
  implicit none

  integer(4), intent(in)              ::  irad
  real(4), intent(in)                 ::  wspd

  real(4), dimension(3), intent(out)  ::  est_err
  ! 1= sigma0 2= emiss 3=NCEP wind speed

  real(4), parameter    :: wbin=1.0, wlow = 0.5
  integer(4)            :: iw1, iw2 
  real(4)               :: ww, brief, w1, w2
  real(4), dimension(7) :: d1, d2, d
  integer(4), parameter :: iwmax=22

  est_err=0.0
    
  ww = wspd 
  if (ww < wlow)  ww=wlow
   
  iw1 = floor((ww-wlow)/wbin) + 1
  if (iw1<1)           iw1 = 1
  if (iw1> iwmax-1)    iw1 = iwmax-1 ! linear interpolation above iwmax
  iw2 = iw1 + 1
        
  w1 = wlow + wbin*(iw1-1)
  w2 = w1   + wbin
  brief = (ww-w1)/(w2-w1)
    
  d1(:) = stdata%estimated_error_array(:,iw1)
  d2(:) = stdata%estimated_error_array(:,iw2)
    
  d(:) = d1(:)*(1.0-brief) + d2(:)*brief ! linear interpolation
    
  if (irad==1) then
     est_err(1) = d(1)
     est_err(2) = d(4)
     est_err(3) = d(7)
  else if (irad==2) then
     est_err(1) = d(2)
     est_err(2) = d(5)
     est_err(3) = d(7)       
  else if (irad==3) then
     est_err(1) = d(3)
     est_err(2) = d(6)
     est_err(3) = d(7)
  endif

  return
end subroutine fd_est_error


real(4) function chi2_HH(w)
  use static_data_module

  !     sos evaluation: HH scatterometer + NCEP background wind speed
  !     fixed phir
  !     variable w
  
  implicit none
  real(4)                            :: w

  integer(4)                         :: irad
  real(4), dimension(0:2,npol_sc)    :: bharm  !1=VV, 2=HH, 3=VH, 4=HV
  real(4), dimension(0:2,npol_sc)    :: dbharm  

  real(4)                            :: f0_HH

  real(4) cosd

  irad=yirad
!  if (irad.eq.1) write(*,*) 'In chi2_HH'

  call fd_TM_scat_harmonics (irad, w, stdata, bharm, dbharm)

  f0_HH = bharm(0,2) + bharm(1,2)*cosd(yphir) + bharm(2,2)*cosd(2.0*yphir)
  chi2_HH = ywt_HH*(ysigma0_HH - f0_HH)**2 + ywt_W*((w - ywin_0)**2)

  return
end function chi2_HH


real(4) function chi2_HHH(w)
  use static_data_module
  !     combined radiometer (H-pol) and scatterometer (HH) + NCEP background
  !     sos evaluation
  !     fixed phir
  !     variable w

  implicit none
  real(4)                            :: w
       
  integer(4)                         :: irad
  real(4), dimension(0:2,2)          :: aharm, aharm_casp  !1=VV, 2=HH, 3=VH, 4=HV
  real(4), dimension(0:2,npol_sc)    :: bharm  !1=VV, 2=HH, 3=VH, 4=HV
  real(4), dimension(0:2,2)          :: daharm, daharm_casp  
  real(4), dimension(0:2,npol_sc)    :: dbharm  

  real(4)                            :: f0_HH, fdew_H, f1, f2

  real(4), dimension(2)              :: dew_1, dew_r, demiss_sst_wspd

  real(4) cosd
  integer, pointer :: nptr => null()
  real(4), parameter :: teff=290. ! used for wind induced emissivity harmonics

  irad=yirad
  call fd_TM_emiss_harmonics (irad, w,     nptr, nptr, stdata, aharm,      daharm)
  call fd_TM_emiss_harmonics (irad, wcasp, nptr, nptr, stdata, aharm_casp, daharm_casp)

  call fd_TM_scat_harmonics (irad, w, stdata, bharm, dbharm)
	
  f0_HH = bharm(0,2) + bharm(1,2)*cosd(yphir) + bharm(2,2)*cosd(2.0*yphir)

  f1 = (aharm(0,2) + aharm(1,2)*cosd(yphir) + aharm(2,2)*cosd(2.0*yphir))
  f2 = (aharm_casp(0,2) + aharm_casp(1,2)*cosd(yphir) + aharm_casp(2,2)*cosd(2.0*yphir))  
    
  ! I am cutting the delta adjustment off at wcasp = 11 m/s 
  if (w<=wcasp) then
     fdew_H = f1*(ydelta_1(2) + ydelta_adj(2))
  else
     fdew_H = f1*ydelta_1(2) + f2*ydelta_adj(2)
  endif

  !fdew_H= aharm(0,2) + aharm(1,2)*cosd(yphir) + aharm(2,2)*cosd(2.0*yphir) + demiss_sst_wspd(2)

  chi2_HHH = ywt_HH*(ysigma0_HH - f0_HH)**2 + ywt_H*(ydew_H - fdew_H)**2  + ywt_W*((w - ywin_0)**2)

return
end function chi2_HHH


real(8) function chi2_HHH_dbl(wdbl)
  implicit none
  real(4)                            :: w
  real(8)                            :: wdbl

  w = wdbl
  chi2_HHH_dbl = dble(chi2_HHH(w))
  
  return
end function chi2_HHH_dbl



subroutine find_delta_fac_adjust (irad, sst, delta_fac_adj)
! adjustment to SST form factor in wind induced emissivity
implicit none

integer(4), intent(in)                              ::  irad
real(4), intent(in)                                 ::  sst  ! Celsius
 
real(4), intent(out), dimension(2)                  ::  delta_fac_adj

real(4)                                             ::  xsst
real(4), dimension(2)                               ::  delta

integer(4)                                          ::  ipol

integer(4)                                          ::  ksst1, ksst2
real(4)                                             ::  x1, x2, brief, y1, y2

integer(4), parameter                               ::  msst=35, n_rad=3
integer(4), save                                    ::  istart=1

real(4), parameter                                  ::  sst_step=1.0, sst0=0.0, sstmax=30.0


xsst=sst
if (xsst<sst0+sst_step/2)   xsst=sst0+sst_step/2
if (xsst>sstmax) xsst=sstmax

ksst1 = floor((xsst - (sst0+sst_step/2))/sst_step) + 1
if (ksst1< 1)           ksst1 = 1
if (ksst1> msst-1)      ksst1 = msst-1
ksst2 = ksst1 + 1

do ipol=1,2
    
    x1 = sst0 + (ksst1-1)*sst_step + sst_step/2
    x2 = x1   + sst_step
    brief = (xsst-x1)/sst_step 
    y1 = stdata%delta_EW(ipol,ksst1,irad)
    y2 = stdata%delta_EW(ipol,ksst2,irad)
    delta(ipol) = y1*(1.0-brief) + y2*brief
    
    delta_fac_adj(ipol) = kappascale*delta(ipol)

enddo ! ipol

return
end subroutine find_delta_fac_adjust

end module wind_speed_retrieval_module
