c*****************************************************************      
      subroutine faraday_rotation_IONEX( sc_position, spacecraft_lon, 
     $ spacecraft_lat, spacecraft_altitude, earth_position_center, 
     $ date, f_far_rot )
c*****************************************************************      
c**   
c**   FILE NAME: faraday_rotation.f
c**   
c**   DATE WRITTEN: 5/28/2009
c**   
c**   PROGRAMMER: Alex Fore
c**   
c**   FUNCTIONAL DESCRIPTION:  Uses IONEX TEC data products and
c**    igrf_sub to compute B field of Earth.
c**
c**   ROUTINES CALLED: 
c**   
c**   NOTES: 
c**
c**   UPDATE LOG:
c**
c**   Date Changed        Reason Changed        CR # and Version #
c**   ------------       ----------------       ------------------
c**     
c*****************************************************************   
      implicit none
c     PARAMETER STATEMENTS
      real*8 pi,r_dtor,r_rtod,r_log2,r_twopi
      parameter(pi=3.141592653589793d0)     !if you have to ask, give it up
      parameter(r_rtod=180.d0/pi,r_dtor=pi/180.d0)  !radian to degree conversions
      parameter(r_log2=.69314718056) !natural logarithm of 2
      parameter(r_twopi=2.0*pi)
      
      real*8 d_e_charge,d_e_mass, f0,r_f0
      parameter(d_e_charge=1.6021d-19)
      parameter(d_e_mass=9.1091d-31)
      parameter(r_f0=1.26)       
      
      real*8 r_sol, r_gm, r_spindot, r_g, r_ftom, r_mtof, r_mtokm, r_kmtom
      parameter(r_sol=299792458.0d0)            !in vacuum speed of light (m/sec)
      parameter(r_gm=3.98600448073d14)          !G*(Earth mass) (m^3/s^2)
      parameter(r_spindot=7.29211573052d-5)     !earth spin rate (rad/sec)
      parameter(r_g=9.8d0)                                      !acceleration due to gravity (m/sec^2)
      parameter(r_ftom = 0.3048d0)          !multiply to convert feet to meters
      parameter(r_mtof = 1.0d0/r_ftom)      !multiply to convert meters to feet
      parameter(r_mtokm = 0.001d0)          !multiply to convert meters to kilometers
      parameter(r_kmtom = 1000.0d0)         !multiply to convert kilometers to meters
      
c     INPUT VARIABLES
      real*8 sc_position(3)
      real*8 spacecraft_lon
      real*8 spacecraft_lat
      real*8 spacecraft_altitude
      real*8 earth_position_center(3)   
      real*8 date(6)                                    ! order: year,month,day,hour,minute,second
c     OUTPUT VARIABLES
      real*8 f_far_rot
c     LOCAL VARIABLES
      character*320 media_a_iripath  ! path to the IRI data file directory
      
      real*8 r_look_unit(3)
      real*8 r_sc_pos_unit(3)
      real*8 r_range
      real*8 r_antposdist
      
      real*8 slant_range
      real*8 iono_position(3)
      real*8 iono_nadir_vector(3)
      real*8 iono_lon
      real*8 iono_lat
      real*8 iono_altitude
      
      real*4 media_f_alati
      real*4 media_f_along
      real*4 media_f_aheight
      real*4 media_f_Bdec
      real*4 media_f_Binc
            
      real*8 MJD, MJD0, julian_time
      integer*4 ierr
      
      real*4 f_TEC            ! TEC from 50km to pltfm ht in TEC units (10^16 m^-2)
      real*8 d_TEC                        ! output from get_ionex_tec is a real*8...
      real*4 f_dyear          ! Decimal year for IGRF computation
      real*4 f_babs           ! IGRF field strength in gauss
      real*4 f_dec,f_dip      ! Field declination (angle from north) and dip in degrees
      real*4 f_l              ! Field L value
      
      real*8 d_dip, d_dec, d_babs, sc_height_km
      
      integer*4 i_year
      integer*4 i_month
      integer*4 i_day
      integer*4 i_day_of_year
      
      real*4 f_B_xyz(3)
      real*8 d_B_enu(3)
      
      integer*4 i_coord
      integer*4 ii
      real*8 d_coslook
      real*8 d_coslkfld
      real*8 d_fr_factor
      integer lnblnk

      logical good_solution
      
c     DATA STATEMENTS
c     COMMON BLOCKS
      character*200 AQ_SCAT_SIM_DIR_OUT, AQ_SCAT_SIM_DIR_DAT, AQ_SCAT_SIM_DIR_SCRATCH
      common /AQ_SCAT_SIM_DIRS/ AQ_SCAT_SIM_DIR_OUT, AQ_SCAT_SIM_DIR_DAT, AQ_SCAT_SIM_DIR_SCRATCH
      
      integer i_dbg_level
      common /debug_level/ i_dbg_level               
c     FUNCTION STATEMENTS
c     PROCESSING STEPS

      if(spacecraft_lon > 180.0) spacecraft_lon=spacecraft_lon-360.0d0
      
      media_a_iripath = AQ_SCAT_SIM_DIR_DAT(1:lnblnk(AQ_SCAT_SIM_DIR_DAT))
     $ // '/faraday/data.iri_2007/'
      
      i_year  = int(date(1))
      i_month = int(date(2))
      i_day   = int(date(3))
      
C--Get the day of year from (month,day,year) information
      call CAL2DAY( i_year, i_month, i_day, i_day_of_year )
      
C--Get the Julian time
      call iau_CAL2JD( i_year, i_month, i_day, MJD0, MJD, ierr)
      julian_time = MJD0 + MJD + date(4)/24.0d0 + date(5)/(60.0d0*24.0d0) 
     $ + date(6)/(24.0d0*60.0d0*60.0d0)
      
      if( i_dbg_level.ge.2)then
        write(6,*) 'in faraday_rotation_IONEX: MJD0, MJD,julian_time: ',MJD0, MJD,julian_time
      endif
           
C--Compute unit vector from s/c to ground footprint.
      r_look_unit(1) = earth_position_center(1) - sc_position(1)
      r_look_unit(2) = earth_position_center(2) - sc_position(2)
      r_look_unit(3) = earth_position_center(3) - sc_position(3)
      
      call norm(r_look_unit,r_range)
      r_look_unit(1) = r_look_unit(1) / r_range
      r_look_unit(2) = r_look_unit(2) / r_range
      r_look_unit(3) = r_look_unit(3) / r_range

C--Compute unit vector that points from origin to s/c location.
      call norm(sc_position,r_antposdist)
      r_sc_pos_unit(1) = sc_position(1) / r_antposdist
      r_sc_pos_unit(2) = sc_position(2) / r_antposdist
      r_sc_pos_unit(3) = sc_position(3) / r_antposdist

C--Locate the cell on the ionosphere      
      call locate_cell_on_ionosphere( sc_position, r_look_unit, 
     $ iono_position, slant_range, good_solution )
      
      if( .not.good_solution )then
        write(6,*) 'ERROR: faraday_rotation_IONEX: Imaginary solution found!'
        call exit(1)
      endif
      
      call compute_sc_nadir_components( iono_position, iono_lon, 
     $ iono_lat, iono_altitude, iono_nadir_vector)      
     
      media_f_alati   =  iono_lat       ! media_f_alati is real*4, iono_lat is r*8
      media_f_along   =  iono_lon   ! media_f_along is real*4, iono_lon is r*8
      media_f_aheight = spacecraft_altitude / 1000.0

C--Get the TEC from IONEX file
      call get_ionex_tec( iono_lat, iono_lon, julian_time, d_TEC )
      if ( d_TEC .eq. -999.0) then
         f_far_rot = -999.0
         return
      endif
      d_TEC = d_TEC * 1.0d16    ! convert from TEC units to electrons / m^2

C--Scale the TEC for Aquarius Height...
      d_TEC = d_TEC * 1.0d0
      f_TEC = d_TEC                             ! f_TEC is real*4, d_TEC is real*8...

C--Compute decimal year := year + day-of-year / num_days_in_year      
      if(  mod(i_year,400).eq.0 .or. 
     $  ( mod(i_year,4).eq.0 .and. mod(i_year,100).ne.0 ) )then
        f_dyear = real(i_year) + real(i_day_of_year)/real(366)
      else
        f_dyear = real(i_year) + real(i_day_of_year)/real(365)
      endif

C--Compute the B field of the Earth      
      call igrf_sub_dip_dec( media_f_alati, media_f_along, f_dyear, 
     $ media_f_aheight, media_a_iripath, f_l, ierr, f_dip, 
     $ f_dec, f_babs)
      
      sc_height_km = spacecraft_altitude / 1000.0d0
      
      call igrf11_wrapper( iono_lat, ! [I] lat in deg 
     $                     iono_lon, ! [I] lon in deg
     $                     i_year,   ! [I] integer year
     $                     i_day_of_year,    ! [I] integer day-of-year
     $                     sc_height_km, ! [I] altitude in km above surface
     $                     d_dip,    ! [O] dip angle in deg
     $                     d_dec,    ! [O] declination angle in deg
     $                     d_babs)   ! [O] magitude of B field (Gauss)
      
      if( i_dbg_level.ge.2 )then
        write(6,'((a50),3f12.6)') 'IGRF10 dip, dec, abs(B): ', f_dip, f_dec, f_babs
        write(6,'((a50),3f12.6)') 'IGRF11 dip, dec, abs(B): ', d_dip, d_dec, d_babs
      endif
      
      f_dip  = d_dip  ! igrf11syn wants doubles and returns doubles
      f_dec  = d_dec  ! igrf11syn wants doubles and returns doubles
      f_babs = d_babs ! igrf11syn wants doubles and returns doubles
      
      media_f_Binc = f_dip*r_dtor
      media_f_Bdec = f_dec*r_dtor
      
      call b_xyz(media_f_Binc, media_f_Bdec, media_f_alati, media_f_along, 
     $ f_B_xyz, d_B_enu)
     
C--Dot product of look unit vector and B field unit vector
      d_coslkfld = 0.
      do i_coord = 1,3
         d_coslkfld = d_coslkfld + f_B_xyz(i_coord)*r_look_unit(i_coord)
      enddo

C--Dot product of look unit vector and unit vector that points from
C-- s/c to origin. (geocentric nadir vector) 
C--(i.e. cosine of look angle).
      d_coslook =0                      
      do i_coord = 1,3
         d_coslook = d_coslook + (-1.d0)*r_sc_pos_unit(i_coord)
     &          *r_look_unit(i_coord)
      enddo     
     
      d_fr_factor = 81.d-4*f_babs*d_e_charge
     &      /(2*r_f0*r_f0*d_e_mass*r_sol)*1e-18

      f_far_rot = d_fr_factor*f_TEC*d_coslkfld/d_coslook


      if( i_dbg_level.ge.2 .or. f_far_rot.ne.f_far_rot )then
        write(6,*) ' '
        write(6,*) 'In subroutine faraday_rotation_IONEX:'
        write(6,'(a50,3f12.6)') 'media (lat, lon, height): ', 
     $   media_f_alati, media_f_along, media_f_aheight
        write(6,'(a50,6f12.6)') 'date: ',date
        write(6,'(a50,1e12.6,1f12.6)') 'tec, f_far_rot: ',f_TEC, f_far_rot
        write(6,'(a50,3f12.6)') 'abs(B), f_dip, f_dec: ', f_babs, f_dip, f_dec
        write(6,'(a50,3f12.6)') 'hat(B): ',f_B_xyz
        write(6,'(a50,3f12.6)') 'cos(look), look dot hat(B): ', d_coslook, d_coslkfld
      endif

C--For making a map of the ionosphere...     
C      write(27,'(f20.8,8f16.8)') sc_position(3), media_f_alati, media_f_along, f_TEC/1.0d16, f_babs, f_B_xyz, f_far_rot
     
      end


c*****************************************************************      
      subroutine faraday_rotation_IRI_2007( sc_position, spacecraft_lon, 
     $ spacecraft_lat, spacecraft_altitude, earth_position_center, 
     $ date, f_far_rot )
c*****************************************************************      
c**   
c**   FILE NAME: faraday_rotation.f
c**   
c**   DATE WRITTEN: 5/28/2009
c**   
c**   PROGRAMMER: Alex Fore
c**   
c**   FUNCTIONAL DESCRIPTION:  
c**
c**   ROUTINES CALLED: 
c**   
c**   NOTES: 
c**
c**   UPDATE LOG:
c**
c**   Date Changed        Reason Changed        CR # and Version #
c**   ------------       ----------------       ------------------
c**     
c*****************************************************************   
      implicit none
c     PARAMETER STATEMENTS
      real*8 pi,r_dtor,r_rtod,r_log2,r_twopi
      parameter(pi=3.141592653589793d0)     !if you have to ask, give it up
      parameter(r_rtod=180.d0/pi,r_dtor=pi/180.d0)  !radian to degree conversions
      parameter(r_log2=.69314718056) !natural logarithm of 2
      parameter(r_twopi=2.0*pi)
      
      real*8 d_e_charge,d_e_mass, f0,r_f0
      parameter(d_e_charge=1.6021d-19)
      parameter(d_e_mass=9.1091d-31)
      parameter(r_f0=1.26)       
      
      real*8 r_sol, r_gm, r_spindot, r_g, r_ftom, r_mtof, r_mtokm, r_kmtom
      parameter(r_sol=299792458.0d0)            !in vacuum speed of light (m/sec)
      parameter(r_gm=3.98600448073d14)          !G*(Earth mass) (m^3/s^2)
      parameter(r_spindot=7.29211573052d-5)     !earth spin rate (rad/sec)
      parameter(r_g=9.8d0)                                      !acceleration due to gravity (m/sec^2)
      parameter(r_ftom = 0.3048d0)          !multiply to convert feet to meters
      parameter(r_mtof = 1.0d0/r_ftom)      !multiply to convert meters to feet
      parameter(r_mtokm = 0.001d0)          !multiply to convert meters to kilometers
      parameter(r_kmtom = 1000.0d0)         !multiply to convert kilometers to meters
      
c     INPUT VARIABLES
      real*8 sc_position(3)
      real*8 spacecraft_lon
      real*8 spacecraft_lat
      real*8 spacecraft_altitude
      real*8 earth_position_center(3)   
      real*8 date(6)                                    ! order: year,month,day,hour,minute,second
c     OUTPUT VARIABLES
      real*8 f_far_rot
c     LOCAL VARIABLES
      character*320 media_a_iripath  ! path to the IRI data file directory
      
      real*8 r_look_unit(3)
      real*8 r_sc_pos_unit(3)
      real*8 r_range
      real*8 r_antposdist
      
      real*8 slant_range
      real*8 iono_position(3)
      real*8 iono_nadir_vector(3)
      real*8 iono_lon
      real*8 iono_lat
      real*8 iono_altitude
                 
C     Description of What IRI Wants for the Position in Space
C     and Time of Conditions to Simulate
C         ALATI,ALONG   LATITUDE NORTH AND LONGITUDE EAST IN DEGREES
C         IYYYY         Year as YYYY, e.g. 1985
C         MMDD (-DDD)   DATE (OR DAY OF YEAR AS A NEGATIVE NUMBER)
C         DHOUR         LOCAL TIME (OR UNIVERSAL TIME + 25) IN DECIMAL 
C                          HOURS
      real*4    media_f_alati     ! platform center position (latitude)
      real*4    media_f_along     ! platform center position (longitude)
      real*4    media_f_aheight   ! platform height (km)
      real*4    media_f_dhour     ! platform center time
      integer*4 media_i_mmdd      ! platform center day of year
      integer*4 media_i_yyyy      ! platform center year

C     Returned from background ionosphere simulation (IRI model)
      real*4 media_f_TEC          ! Total Electron Content (in m^-2) up to platform
      real*4 media_f_f2ht         ! F2 layer height (km)
      real*4 media_f_Babs         ! Magnitude of Geomagnetic field (gauss)
      real*4 media_f_Bdec         ! Magnetic field declination (rad)
      real*4 media_f_Binc         ! Magnetic field inclination (rad)
            
            
      real*8 MJD, MJD0, julian_time
      integer*4 ierr
      
      real*4 f_TEC            ! TEC from 50km to pltfm ht in TEC units (10^16 m^-2)
      real*8 d_TEC                        ! output from get_ionex_tec is a real*8...
      real*4 f_dyear          ! Decimal year for IGRF computation
      real*4 f_babs           ! IGRF field strength in gauss
      real*4 f_dec,f_dip      ! Field declination (angle from north) and dip in degrees
      real*4 f_l              ! Field L value
      
      integer*4 i_year
      integer*4 i_month
      integer*4 i_day
      integer*4 i_day_of_year
      
      real*4 f_B_xyz(3)
      real*8 d_B_enu(3)
      
      integer*4 i_coord
      integer*4 ii
      real*8 d_coslook
      real*8 d_coslkfld
      real*8 d_fr_factor

      integer lnblnk

      logical good_solution
      logical media_l_ionoheight      ! over-riding height default?
      real*4  media_f_ionolayerheight ! over-ride ionosphere height      
      
c     DATA STATEMENTS
c     COMMON BLOCKS
      character*200 AQ_SCAT_SIM_DIR_OUT, AQ_SCAT_SIM_DIR_DAT, AQ_SCAT_SIM_DIR_SCRATCH
      common /AQ_SCAT_SIM_DIRS/ AQ_SCAT_SIM_DIR_OUT, AQ_SCAT_SIM_DIR_DAT, AQ_SCAT_SIM_DIR_SCRATCH
      
      integer i_dbg_level
      common /debug_level/ i_dbg_level               
c     FUNCTION STATEMENTS
c     PROCESSING STEPS

      if(spacecraft_lon > 180.0) spacecraft_lon=spacecraft_lon-360.0d0
      
      media_a_iripath = AQ_SCAT_SIM_DIR_DAT(1:lnblnk(AQ_SCAT_SIM_DIR_DAT))
     $ // '/faraday/data.iri_2007/'
      
      i_year   = int(date(1))
      i_month  = int(date(2))
      i_day    = int(date(3))
      
C--Get the day of year from (month,day,year) information
      call CAL2DAY( i_year, i_month, i_day, i_day_of_year )
      
C--Get the Julian time
      call iau_CAL2JD( i_year, i_month, i_day, MJD0, MJD, ierr)
      julian_time = MJD0 + MJD + date(4)/24.0d0 + date(5)/(60.0d0*24.0d0) 
     $ + date(6)/(24.0d0*60.0d0*60.0d0)
      
      if( i_dbg_level.ge.2)then
        write(6,*) 'in faraday_rotation_IONEX: MJD0, MJD,julian_time: ',MJD0, MJD,julian_time
      endif
           
C--Compute look unit vector from s/c to ground footprint.
      r_look_unit(1) = earth_position_center(1) - sc_position(1)
      r_look_unit(2) = earth_position_center(2) - sc_position(2)
      r_look_unit(3) = earth_position_center(3) - sc_position(3)
      
      call norm(r_look_unit,r_range)
      r_look_unit(1) = r_look_unit(1) / r_range
      r_look_unit(2) = r_look_unit(2) / r_range
      r_look_unit(3) = r_look_unit(3) / r_range

C--Compute unit vector that points from origin to s/c location.
      call norm(sc_position,r_antposdist)
      r_sc_pos_unit(1) = sc_position(1) / r_antposdist
      r_sc_pos_unit(2) = sc_position(2) / r_antposdist
      r_sc_pos_unit(3) = sc_position(3) / r_antposdist

C--Locate the cell on the ionosphere      
      call locate_cell_on_ionosphere( sc_position, r_look_unit, 
     $ iono_position, slant_range, good_solution )

      if( .not.good_solution )then
        write(6,*) 'ERROR: faraday_rotation_IONEX: Imaginary solution found!'
        call exit(1)
      endif

      call compute_sc_nadir_components( iono_position, iono_lon, 
     $ iono_lat, iono_altitude, iono_nadir_vector)      
     
C--Set up call to compute_iri     
      media_f_alati   = iono_lat        ! media_f_alati is real*4, iono_lat is r*8
      media_f_along   = iono_lon    ! media_f_along is real*4, iono_lon is r*8
      media_f_aheight = spacecraft_altitude / 1000.0
      media_f_dhour   = date(4) + date(5) / 60.0d0
      media_i_mmdd    = i_day + 100 * i_month
      media_i_yyyy    = i_year
      
      call compute_iri( media_a_iripath, media_l_ionoheight,
     &          media_f_ionolayerheight, media_f_alati,
     &          media_f_along, media_f_aheight,
     &          media_f_dhour, media_i_mmdd, media_i_yyyy,
     &          media_f_TEC, media_f_f2ht, media_f_Babs,
     &          media_f_Bdec, media_f_Binc )

      f_TEC  = media_f_TEC
      f_babs = media_f_Babs
      f_dip  = media_f_Binc*r_rtod
      f_dec  = media_f_Bdec*r_rtod
      
      call b_xyz(media_f_Binc, media_f_Bdec, media_f_alati, media_f_along, 
     $ f_B_xyz, d_B_enu)

C--Dot product of look unit vector and B field unit vector
      d_coslkfld = 0.
      do i_coord = 1,3
         d_coslkfld = d_coslkfld + f_B_xyz(i_coord)*r_look_unit(i_coord)
      enddo

C--Dot product of look unit vector and unit vector that points from
C-- s/c to origin. (geocentric nadir vector) 
C--(i.e. cosine of look angle).
      d_coslook =0                      
      do i_coord = 1,3
         d_coslook = d_coslook + (-1.d0)*r_sc_pos_unit(i_coord)
     &          *r_look_unit(i_coord)
      enddo     
     
      d_fr_factor = 81.d-4*f_babs*d_e_charge
     &      /(2*r_f0*r_f0*d_e_mass*r_sol)*1e-18

      f_far_rot = d_fr_factor*f_TEC*d_coslkfld/d_coslook

      if( i_dbg_level.ge.2)then
        write(6,*) ' '
        write(6,*) 'In subroutine faraday_rotation_IRI_2007:'
        write(6,'(a50,3f12.6)') 'media (lat, lon, height): ', 
     $   media_f_alati, media_f_along, media_f_aheight
        write(6,'(a50,6f12.6)') 'date: ',date
        write(6,'(a50,1e12.6,1f12.6)') 'tec: ',f_TEC, f_far_rot
        write(6,'(a50,3f12.6)') 'abs(B), f_dip, f_dec: ', f_babs, f_dip, f_dec
        write(6,'(a50,3f12.6)') 'hat(B): ',f_B_xyz
        write(6,'(a50,3f12.6)') 'cos(look), look dot hat(B): ', d_coslook, d_coslkfld
      endif

C--For making a map of the ionosphere...     
C      write(27,'(f20.8,8f16.8)') sc_position(3), media_f_alati, media_f_along, f_TEC/1.0d16, f_babs, f_B_xyz, f_far_rot
     
      end

c****************************************************************

      subroutine compute_iri(media_a_iripath,media_l_ionoheight,
     &      media_f_ionolayerheight,media_f_alati,
     &      media_f_along,media_f_aheight, 
     &      media_f_dhour,media_i_mmdd , media_i_yyyy,
     &      media_f_TEC,media_f_f2ht,media_f_Babs,
     &      media_f_Bdec,media_f_Binc)

c****************************************************************
c**     
c**   FILE NAME: compute_iri(media)
c**     
c**   DATE WRITTEN: 13/November/2003
c**     
c**   PROGRAMMER: Jan Martin
c**     
c**   FUNCTIONAL DESCRIPTION: 
c**   This subroutine will set up arguments for and call
c**   the 'iri_sub_ne' subroutine which will compute the
c**   International Reference Ionosphere TEC and height profile
c**   for electron density at the point defined in the
c**   media structure.  
c**     
c**   ROUTINES CALLED:
c**     
c**   NOTES: 
c**     
c**   UPDATE LOG:
c**
c**   Date Changed        Reason Changed                  CR # and Version #
c**   ------------       ----------------                 -----------------
c**     
c*****************************************************************

      implicit none

c     MATHEMATICAL CONSTANTS

      real*8 pi,r_dtor,r_rtod,r_log2,r_twopi
      parameter(pi=3.141592653589793d0)     !if you have to ask, give it up
      parameter(r_rtod=180.d0/pi,r_dtor=pi/180.d0)  !radian to degree conversions
      parameter(r_log2=.69314718056) !natural logarithm of 2
      parameter(r_twopi=2.0*pi)


C     INCLUDE FILES:

c     PARAMETER STATEMENTS

      real*4 f_minht            ! minimum height for ionosphere effect
      parameter (f_minht=100.0)

c     INPUT VARIABLES:

        character*320 media_a_iripath  ! path to the IRI data file directory
        logical media_l_ionoheight     ! over-riding height default?
        real*4 media_f_ionolayerheight ! over-ride ionosphere height


C     Description of What IRI Wants for the Position in Space
C     and Time of Conditions to Simulate
C         ALATI,ALONG   LATITUDE NORTH AND LONGITUDE EAST IN DEGREES
C         IYYYY         Year as YYYY, e.g. 1985
C         MMDD (-DDD)   DATE (OR DAY OF YEAR AS A NEGATIVE NUMBER)
C         DHOUR         LOCAL TIME (OR UNIVERSAL TIME + 25) IN DECIMAL 
C                          HOURS

        real*4 media_f_alati    ! platform center position (latitude)
        real*4 media_f_along    ! platform center position (longitude)
        real*4 media_f_aheight  ! platform height (km)
        real*4 media_f_dhour    ! platform center time
        integer*4 media_i_mmdd  ! platform center day of year
        integer*4 media_i_yyyy  ! platform center year
        
C     Returned from background ionosphere simulation (IRI model)
        real*4 media_f_TEC      ! Total Electron Content (in m^-2) up to platform
        real*4 media_f_f2ht     ! F2 layer height (km)
        real*4 media_f_Babs     ! Magnitude of Geomagnetic field (gauss)
        real*4 media_f_Bdec     ! Magnetic field declination (rad)
        real*4 media_f_Binc     ! Magnetic field inclination (rad)
                
c     OUTPUT VARIABLES:
        
c     LOCAL VARIABLES:
        
        logical l_jf(30)        ! switches for options - default .true.
        integer*4 i_jmag        ! geographic (0) or geomagnetic(1) coordinates
        integer*4 i_code        ! validity of L value
        integer*4 i_month       ! Month
        integer*4 i_day         ! Day of month
        integer*4 ii,i_idx
        
        real*4 f_hour           ! UT decimal hour of observation
        real*4 f_htbeg,f_htend  ! starting and ending heights for outf
        real*4 f_htstp          ! height step for outf
        real*4 f_outf(1:20,1:500) ! output iono param array
        real*4 f_oarr(1:50)     ! additional output iono parameters
        real*4 f_dyear          ! Decimal year for IGRF computation
        
        real*4 f_TEC            ! TEC from 50km to pltfm ht in TEC units (10^16 m^-2)
        real*4 f_tectop         ! Amount of tec above platform ht (%)
        real*4 f_tecbot         ! Amount of TEC below 50km (%)
        real*4 f_babs           ! IGRF field strength in gauss
        real*4 f_dec,f_dip      ! Field declination (angle from north) and dip in degrees
        real*4 f_l              ! Field L value
        real*4 f_scrnht         ! Mean ionosphere height up to platform height
        real*4 f_sumNe

c     COMMON BLOCKS:
      integer i_dbg_level
      common /debug_level/ i_dbg_level   
c     EQUIVALENCE STATEMENTS:

c     DATA STATEMENTS:

C     FUNCTION STATEMENTS:

c     PROCESSING STEPS:

c     First check to see if there really is ionosphere up to the 
c     platform height - if not, just return with 0
      if (media_f_aheight .le. f_minht) then
         write (6,'(a)') 'Platform Height 
     &   is below the Ionosphere So Setting TEC to 0'
         write (6,'(a,f20.3)') 'Platform Height (km) = '
     &       , media_f_aheight
         write (6,'(a,f20.3)') 'Minimum Ionosphere 
     &        Height (km) = ',f_minht
         f_TEC = 0.
         return
      endif
      
c     First fill the l_jf array with default values
      do ii = 1,30
         l_jf(ii) = .true.
      enddo
      
C--AGF modify 5/19/2010
C--Set to defaults (taken from irisub.for from iri2007 release).
      l_jf(2)  = .false. ! Te, Ti not computed 
      l_jf(3)  = .false. ! Ni not computed
      l_jf(5)  = .false. ! foF2 - URSI
      l_jf(6)  = .false. ! Ni - DS-95 & TTS-03 
      l_jf(12) = .false. ! suppress IRI output to stdout
      l_jf(21) = .false. ! ion drift not computed
      l_jf(23) = .false. ! Te_topside (Intercosmos)
      l_jf(28) = .false. ! spread-F probability not computed
      l_jf(29) = .false. ! new options as def. by JF(30)
      l_jf(30) = .false. ! NeQuick topside model
      ! l_jf(29) ==.false. && l_jf(3) == .false => (f,f) NeQuick
C--End of AGF modify iri input flags.

c     Coordinates are geographic
      i_jmag = 0

c     Time is UTC (or so close that it doesn't matter)
      f_hour = media_f_dhour + 25

c     Set heights where the ionosphere values such as N_e, T_e, etc.
c     are calculated.  Used for finding the mean height for screen.
      f_htbeg = 80.0
      if (media_f_aheight .gt. 1200.0) then
         f_htend = 1200.0
      else
         f_htend = media_f_aheight
      endif
      f_htstp = (f_htend - f_htbeg)/99
      
      if( i_dbg_level.ge.2) then
        write(6,*) 'In subroutine compute_iri: '
        write(6,*) 'height start, step, end [km]: ', f_htbeg, f_htstp, f_htend
      endif
      
c     alati,along,iyyyy,mmdd, and dhour are already set by determine_iri_point

c     Call iri_sub_ne to generate IRI values for this position and time
c     It fills all the quantities in the common blocks which iri_tec will
c     use to integrate up TEC
C
C--AGF changed call to iri_sub since iri2007 doesn't have iri_sub_ne 5/19/2010
      call iri_sub( l_jf, i_jmag, media_f_alati, media_f_along,
     &     media_i_yyyy, media_i_mmdd, f_hour,f_htbeg, f_htend,
     &     f_htstp, media_a_iripath, f_outf, f_oarr)

c     Find mean Ne height for screen, unless already chosen by user
      if (.not.media_l_ionoheight) then
         f_scrnht = 0.
         f_sumNe = 0.
         do i_idx = 1,100
            f_scrnht = f_scrnht + (f_htbeg + (i_idx - 1)*f_htstp)*
     &              f_outf(1,i_idx)
            f_sumNe = f_sumNe + f_outf(1,i_idx)
         enddo
         media_f_ionolayerheight = f_scrnht/f_sumNe
      endif

      media_f_f2ht = f_oarr(2)

c     Now call iri_tec to integrate N_e from 50km up to platform height
c     Use most accurate integration algorithm (istep=2)
      call iri_tec(f_minht,media_f_aheight,2,f_TEC,f_tectop,f_tecbot)
      
c     Load TEC into media structure
      media_f_TEC = f_TEC

c     Calculate geomagnetic field
      i_month = media_i_mmdd / 100
      i_day   = media_i_mmdd - i_month*100
      f_dyear = media_i_yyyy + (i_month  - 1 + i_day/30.)/12.
      
      call igrf_sub_dip_dec( media_f_alati, media_f_along, f_dyear, 
     $ media_f_aheight, media_a_iripath, f_l, i_code, f_dip, 
     $ f_dec, f_babs )
      
      
c     Load geomagnetic field parameters into media structure
      media_f_Babs = f_babs
      media_f_Bdec = f_dec*r_dtor
      media_f_Binc = f_dip*r_dtor

      if ( i_dbg_level .ge. 2) then
         write(6,'(a)') ' '
         write(6,'(a)') 'TEC calculated for IRI Background '//
     +        'Ionosphere Model'
         write(6,732) 'TEC      (hex) = ',f_TEC/1.e16
         write(6,732) 'topTEC      (%) = ',f_tectop
         write(6,732) 'botTEC      (%) = ',f_tecbot
         write(6,'(a)') ' '
         write(6,'(a)') 'Background IGRF'
         write(6,732) 'Field strength (gauss) = ',f_babs
         write(6,732) 'Dip angle   (deg) = ',f_dip
         write(6,732) 'Decl angle  (deg) = ',f_dec
 732     format(a24,f20.6)
      endif

      return
      end


c****************************************************************

        subroutine norm(r_v,r_n)

c****************************************************************
c**
c**     FILE NAME: norm.for
c**
c**     DATE WRITTEN: 8/3/90
c**
c**     PROGRAMMER:Scott Hensley
c**
c**     FUNCTIONAL DESCRIPTION: The subroutine takes vector and returns 
c**     its norm.
c**
c**     ROUTINES CALLED:none
c**  
c**     NOTES: none
c**
c**     UPDATE LOG:
c**
c*****************************************************************

        implicit none

c       INPUT VARIABLES:
        real*8 r_v(3)                              !3x1 vector
   
c       OUTPUT VARIABLES:see input

c       LOCAL VARIABLES:
        real*8 r_n

c       PROCESSING STEPS:

c       compute vector norm

        r_n = sqrt(r_v(1)**2 + r_v(2)**2 + r_v(3)**2)
      
        end  


c****************************************************************

        subroutine lincomb(r_k1,r_u,r_k2,r_v,r_w)

c****************************************************************
c**
c**     FILE NAME: lincomb.for
c**
c**     DATE WRITTEN: 8/3/90
c**
c**     PROGRAMMER:Scott Hensley
c**
c**     FUNCTIONAL DESCRIPTION: The subroutine forms the linear combination
c**     of two vectors.
c**
c**     ROUTINES CALLED:none
c**  
c**     NOTES: none
c**
c**     UPDATE LOG:
c**
c*****************************************************************

        implicit none

c       INPUT VARIABLES:
        real*8 r_u(3)                              !3x1 vector
        real*8 r_v(3)                              !3x1 vector
        real*8 r_k1                              !scalar
        real*8 r_k2                              !scalar
   
c       OUTPUT VARIABLES:
        real*8 r_w(3)                              !3x1 vector

c       LOCAL VARIABLES:none

c       PROCESSING STEPS:

c       compute linear combination

        r_w(1) = r_k1*r_u(1) + r_k2*r_v(1)
        r_w(2) = r_k1*r_u(2) + r_k2*r_v(2)
        r_w(3) = r_k1*r_u(3) + r_k2*r_v(3)
      
        end  


c****************************************************************

        subroutine matvec(r_t,r_v,r_w)

c****************************************************************
c**
c**     FILE NAME: matvec.for
c**
c**     DATE WRITTEN: 7/20/90
c**
c**     PROGRAMMER:Scott Hensley
c**
c**     FUNCTIONAL DESCRIPTION: The subroutine takes a 3x3 matrix 
c**     and a 3x1 vector a multiplies them to return another 3x1
c**     vector.
c**
c**     ROUTINES CALLED:none
c**  
c**     NOTES: none
c**
c**     UPDATE LOG:
c**
c****************************************************************

        implicit none

c       INPUT VARIABLES:
        real*8 r_t(3,3)                            !3x3 matrix
        real*8 r_v(3)                              !3x1 vector
   
c       OUTPUT VARIABLES:
        real*8 r_w(3)                              !3x1 vector

c       LOCAL VARIABLES:none

c       PROCESSING STEPS:

c       compute matrix product

        r_w(1) = r_t(1,1)*r_v(1) + r_t(1,2)*r_v(2) + r_t(1,3)*r_v(3)
        r_w(2) = r_t(2,1)*r_v(1) + r_t(2,2)*r_v(2) + r_t(2,3)*r_v(3)
        r_w(3) = r_t(3,1)*r_v(1) + r_t(3,2)*r_v(2) + r_t(3,3)*r_v(3)
          
        end  

c****************************************************************
        subroutine latlon( r_a ,r_e2,r_v,r_llh,i_type) 

c****************************************************************
c**   
c**   FILE NAME: latlon.f
c**   
c**   DATE WRITTEN:7/22/93 
c**   
c**   PROGRAMMER:Scott Hensley
c**   
c**   FUNCTIONAL DESCRIPTION:This program converts a vector to 
c**   lat,lon and height above the reference ellipsoid or given a
c**   lat,lon and height produces a geocentric vector. 
c**   
c**   ROUTINES CALLED:none
c**   
c**   NOTES: none
c**   
c**   UPDATE LOG:
c**   
c****************************************************************
        
        implicit none
        
c       INPUT VARIABLES:

        integer i_type          !1=lat,lon to vector,2= vector to lat,lon

        real*8 r_v(3)                    !geocentric vector (meters)
        real*8 r_llh(3)                  !latitude (deg -90 to 90),longitude (deg -180 to 180),height
   
c       OUTPUT VARIABLES: see input

c       LOCAL VARIABLES:

        integer i_ft
        real*8 pi,r_dtor,r_re,r_q2,r_q3,r_b,r_q
        real*8 r_p,r_tant,r_theta,r_a,r_e2

c       DATA STATEMENTS:

        data pi /3.141592653589793238d0/
        data r_dtor /1.74532925199d-2/

C       FUNCTION STATEMENTS:

c       PROCESSING STEPS:


        if(i_type .eq. 1)then  !convert lat,lon to vector
           
           r_re = r_a/sqrt(1.d0 - r_e2*sin(r_llh(1))**2)
           
           r_v(1) = (r_re + r_llh(3))*cos(r_llh(1))*cos(r_llh(2))
           r_v(2) = (r_re + r_llh(3))*cos(r_llh(1))*sin(r_llh(2))
           r_v(3) = (r_re*(1.d0-r_e2) + r_llh(3))*sin(r_llh(1))               
           
        elseif(i_type .eq. 2)then  !convert vector to lat,lon 
           
           r_q2 = 1.d0/(1.d0 - r_e2)
           r_q = sqrt(r_q2)
           r_q3 = r_q2 - 1.d0
           r_b = r_a*sqrt(1.d0 - r_e2)
           
           r_llh(2) = atan2(r_v(2),r_v(1))
           
           r_p = sqrt(r_v(1)**2 + r_v(2)**2)
           r_tant = (r_v(3)/r_p)*r_q
           r_theta = atan(r_tant)
           r_tant = (r_v(3) + r_q3*r_b*sin(r_theta)**3)/
     +          (r_p - r_e2*r_a*cos(r_theta)**3)
           r_llh(1) =  atan(r_tant)
           r_re = r_a/sqrt(1.d0 - r_e2*sin(r_llh(1))**2)
           r_llh(3) = r_p/cos(r_llh(1)) - r_re          
  
        endif
      
        end  



c****************************************************************
      subroutine b_xyz(media_f_Binc, media_f_Bdec,
     &   media_f_alati, media_f_along,f_xyz,d_benu)

c****************************************************************
c**     
c**   FILE NAME: b_xyz(media,f_xyz)
c**     
c**   DATE WRITTEN: 10/January/2004
c**     
c**   PROGRAMMER: Jan Martin
c**     
c**   FUNCTIONAL DESCRIPTION: 
c**   This subroutine takes the IGRF B-field representation
c**   contained in the media struct and converts it to
c**   XYZ coordinates
c**     
c**   ROUTINES CALLED:
c**     
c**   NOTES: 
c**     
c**   UPDATE LOG:
c**
c**   Date Changed        Reason Changed                  CR # and Version #
c**   ------------       ----------------                 -----------------
c**     
c*****************************************************************

      implicit none

C     INCLUDE FILES:


c     PARAMETER STATEMENTS

      real*8 pi,r_dtor,r_rtod,r_log2,r_twopi
      parameter(pi=3.141592653589793d0)     !if you have to ask, give it up
      parameter(r_rtod=180.d0/pi,r_dtor=pi/180.d0)  !radian to degree conversions
      parameter(r_log2=.69314718056) !natural logarithm of 2
      parameter(r_twopi=2.0*pi)

c     INPUT VARIABLES:

        
c     OUTPUT VARIABLES:

      real*4 f_xyz(3)

c     LOCAL VARIABLES:


      real*8 d_enuxyzmat(3,3)
      real*8 d_benu(3),d_xyz(3)
      real*4 media_f_Binc, media_f_Bdec,media_f_alati,media_f_along

c     COMMON BLOCKS:

c     EQUIVALENCE STATEMENTS:

c     DATA STATEMENTS:

C     FUNCTION STATEMENTS:

c     PROCESSING STEPS:

c     Calculate the rotation matrix between ENU and XYZ coordinate systems


c      write(*,*) media_f_alati*r_dtor,media_f_along*r_dtor
      call enubasis(media_f_alati*r_dtor,media_f_along*r_dtor,
     &               d_enuxyzmat)


c     Calculate the unit vector in the B-field direction, in ENU coordinates
      d_benu(1) = cos(media_f_Binc)*sin(media_f_Bdec)
      d_benu(2) = cos(media_f_Binc)*cos(media_f_Bdec)
      d_benu(3) = -sin(media_f_Binc)


c     Rotate to XYZ basis
      call matvec(d_enuxyzmat,d_benu,d_xyz)

      f_xyz(1) = d_xyz(1)
      f_xyz(2) = d_xyz(2)
      f_xyz(3) = d_xyz(3)



      end



c****************************************************************

        subroutine enubasis(r_lat,r_lon,r_enuxyzmat)

c****************************************************************
c**
c**     FILE NAME: enubasis.f
c**
c**     DATE WRITTEN: 7/22/93
c**
c**     PROGRAMMER:Scott Hensley
c**
c**     FUNCTIONAL DESCRIPTION: Takes a lat and lon and returns a 
c**     change of basis matrix from ENU to XYZ geocentric coordinates.  
c**
c**     ROUTINES CALLED:none
c**  
c**     NOTES: none
c**
c**     UPDATE LOG:
c****************************************************************

        implicit none

c       INPUT VARIABLES:

        real*8 r_lat                   !latitude (rad)
        real*8 r_lon                   !longitude (rad)
   
c       OUTPUT VARIABLES:

        real*8 r_enuxyzmat(3,3)

c       LOCAL VARIABLES:

        real*8 r_slt,r_clt,r_clo,r_slo

c       DATA STATEMENTS:

C       FUNCTION STATEMENTS:

c       PROCESSING STEPS:

        r_clt = cos(r_lat)
        r_slt = sin(r_lat)
        r_clo = cos(r_lon)
        r_slo = sin(r_lon)

c     North  vector

        r_enuxyzmat(1,2) = -r_slt*r_clo        
        r_enuxyzmat(2,2) = -r_slt*r_slo
        r_enuxyzmat(3,2) = r_clt
      
c     East vector

        r_enuxyzmat(1,1) = -r_slo        
        r_enuxyzmat(2,1) = r_clo
        r_enuxyzmat(3,1) = 0.d0

c     Up vector 
      
        r_enuxyzmat(1,3) = r_clt*r_clo        
        r_enuxyzmat(2,3) = r_clt*r_slo
        r_enuxyzmat(3,3) = r_slt

        end  
