C--$Header: /mnt/caretr2/scatsim/CVSROOT/AQ_SCATSIM_OBJS/get_mask2.f,v 1.2 2011/01/07 17:01:34 fore Exp $
c*****************************************************************      
      subroutine get_mask2( lat, lon, land_mask )
c*****************************************************************      
c**   
c**   FILE NAME: get_mask2.f
c**   
c**   DATE WRITTEN: 12/27/2010
c**   
c**   PROGRAMMER: Alex Fore, based on F. Patt's get_mask 
c**     subroutine. 
c**   
c**   FUNCTIONAL DESCRIPTION:  Returns 0 for .not.land, 1 for land
c**
c**   ROUTINES CALLED: ffloor
c**   
c**   NOTES: Uses much more memory but much faster than get_mask.
c**     Written for the ODPS maskfile.  Verified that it gives 
c**     exactly the same results for 3 * 10^7 random lat, lon 
c**     trials.
c**
c**   UPDATE LOG: 
c**
c**   Date Changed        Reason Changed        CR # and Version #
c**   ------------       ----------------       ------------------
c**     
c*****************************************************************   
      implicit none
c     PARAMETER STATEMENTS
      integer stdout, stderr, stdin
      parameter( stdout = 6, stderr = 0, stdin = 5 )
      integer NRECS, REC_LENGTH
      parameter( NRECS = 9644, REC_LENGTH = 2048 )
c     INPUT VARIABLES
      real*4        lat
      real*4        lon
c     OUTPUT VARIABLES
      integer*1     land_mask
c     LOCAL VARIABLES
      character*320 maskfile

      logical       maskfile_read    /.false./
      logical       land_flag
      integer       i_rec
      integer       ii
      
      integer*2     datarec(REC_LENGTH/2)       ! temporary
      integer*2     allrecs(REC_LENGTH/2,NRECS) ! holds all of maskfile in RAM
      integer*2     rec_ptr                     ! pointer to record for 1x1 deg "cell"

      integer       rec_ptr_ind     ! index of the record pointer in map header
      integer       byte_offset     ! byte offset into record for i_lat,i_lon
      integer       bit_test        ! bit of byte to test for land/.not.land
      integer       i_lat           ! lon index in 1x1 degree "cell"
      integer       i_lon           ! lon index in 1x1 degree "cell"
      
      integer       mask_resolution ! bits per degree lat,lon
      integer       mask_num_recs   ! number of records in maskfile
      integer       mask_recl       ! record length in bytes of maskfile
      
      integer*2     short_var
      byte          byte_var(2)
      byte          byte_test
      integer lnblnk

c     DATA STATEMENTS
c     COMMON BLOCKS
      integer i_dbg_level
      common /debug/ i_dbg_level

      character*200 AQ_SCAT_SIM_DIR_OUT, AQ_SCAT_SIM_DIR_DAT, AQ_SCAT_SIM_DIR_SCRATCH
      common /AQ_SCAT_SIM_DIRS/ AQ_SCAT_SIM_DIR_OUT, AQ_SCAT_SIM_DIR_DAT, AQ_SCAT_SIM_DIR_SCRATCH
      
c     FUNCTION STATEMENTS
      integer ffloor
c     EQUIVALENCE STATEMENTS
      equivalence ( short_var, byte_var(1) )
c     PROCESSING STEPS
      save allrecs, mask_resolution, mask_num_recs, mask_recl, maskfile_read, maskfile
      
      
!      maskfile = AQ_SCAT_SIM_DIR_DAT(1:lnblnk(AQ_SCAT_SIM_DIR_DAT)) //
!     $           '/winds/landmask_little_endian' 
      
      land_mask = 0      
      if( .not.maskfile_read ) then
         maskfile = AQ_SCAT_SIM_DIR_DAT(1:lnblnk(AQ_SCAT_SIM_DIR_DAT)) //
     $        '/landmask_big_endian' 
        ! Only read maskfile the 1st time.
        open(17,file=maskfile,access='direct',recl=REC_LENGTH)
        do i_rec = 1, NRECS
          read(17,rec=i_rec) datarec  
          do ii = 1, REC_LENGTH/2
            allrecs(ii,i_rec) = datarec(ii)
          enddo
        enddo
        close(17)
        
        mask_resolution = allrecs(1,1)
        mask_num_recs   = allrecs(2,1)
        mask_recl       = allrecs(3,1)
        
        ! Sanity check, maybe this could be done better?
        if( mask_recl       .ne. REC_LENGTH .or.
     $      mask_num_recs   .ne. NRECS      .or.
     $      mask_resolution .ne. 128 )then
          write(stdout,*) 'get_mask2: ERROR, unexpected sizes in mask header!'
          stop
        endif
        
        if( i_dbg_level .gt. 1 ) then
          write(stdout,'((a20),I6)') 'Mask resolution: ', allrecs(1,1)
          write(stdout,'((a20),I6)') 'Mask # records:  ', allrecs(2,1)
          write(stdout,'((a20),I6)') 'Mask recl:       ', allrecs(3,1)
          write(stdout,'((a20),I6)') 'Mask west bndy:  ', allrecs(4,1)
          write(stdout,'((a20),I6)') 'Mask east bndy:  ', allrecs(5,1)
          write(stdout,'((a20),I6)') 'Mask south bndy: ', allrecs(6,1)
          write(stdout,'((a20),I6)') 'Mask north bndy: ', allrecs(7,1)
        endif
        
        maskfile_read = .true.
      endif
      
      ! Require lon in [-180,180) lat in [-90,90)
      if( lon .lt. -180.0 .or. lon .ge. 180.0 .or.
     $    lat .lt.  -90.0 .or. lat .ge.  90.0 ) then
        write(stdout,*) 'ERROR: get_mask2: Inputs out of range!'
        land_mask = 0
        return
      endif
      
      ! Get pointer to record
      
      ! rec_ptr_ind = 360 * ffloor(lat+90.0) + ffloor(lon+180.0)
      ! 
      ! Replicate floating point logic used in F. Patt's version (no floating point
      ! round-off differences then). 1/7/2011
      i_lat       = lat + 91.0
      i_lon       = lon + 181.0
      rec_ptr_ind = 360 * ( i_lat - 1 ) + i_lon - 1
      rec_ptr     = allrecs( rec_ptr_ind + 1 , 2 )

      if( i_dbg_level .gt. 2 ) then
        write(stdout,'((a50),2f12.6)') 'lat, lon: ', lat, lon
        write(stdout,'((a50),2I6)') 'rec_ptr_ind, rec_ptr: ', rec_ptr_ind, rec_ptr
        write(stdout,*) ' '
      endif
      
      if( rec_ptr .eq. 0 )then       
        land_mask = 0 ! no land in this cell
      else if( rec_ptr .eq. 1 ) then 
        land_mask = 1 ! all land in this cell
      else
        ! Mix of land and not land in cell; need to test bits...
        ! Each record is 128x128 bit array of flags, for each 1x1 deg 
        ! lat, lon "cell".
        i_lat = int( 128 * mod( lat +  90.0, 1.0 ) + 1.0 )
        i_lon = int( 128 * mod( lon + 180.0, 1.0 ) + 1.0 )
        
        if( i_lat .lt. 1 .or. i_lat .gt. 128 .or.
     $      i_lon .lt. 1 .or. i_lon .gt. 128 ) then
          write(stdout,*) 'get_mask: ERROR, i_lat or i_lon out of range: ', 
     $      i_lat, i_lon
          stop
        endif
        
        ! 0-based byte offset into record
        byte_offset = ( mask_resolution * (i_lat-1) + (i_lon-1) ) / 8
        
        ! bit of char to test
        bit_test = 7 - mod( i_lon - 1 + 8, 8 )
        
        ! copy short record into short_var
        short_var = allrecs( byte_offset / 2 + 1, rec_ptr + 1 )
        
        ! set byte_test to byte to do bit testing on
        ! short_var and byte_var(1) are "equivilent" above.
        ! Note: for little endian switch byte_var indicies 1 <--> 2
        if( mod( byte_offset + 2, 2 ) .eq. 0 )then
          byte_test = byte_var(1)
        else
          byte_test = byte_var(2)
        endif
        
        ! test bit for land
        if( btest( byte_test, bit_test ) ) then
          land_mask = 1
        else
          land_mask = 0
        endif
        
      endif
      
      return
      end
      
      
      function ffloor( b )
      implicit none
      
      integer ffloor
      real*4  b
      
      if( b .ge. 0 ) then
        ffloor = int( b )
      else
        ffloor = int( b ) - 1
      endif
      
      return
      end
            
