c*****************************************************************      
      subroutine iri2007_VTEC_Scale_Factor( sc_position,
     $                                      earth_position_center,
     $                                      date,
     $                                      tec_scale_factor )
c*****************************************************************      
c**   
c**   FILE NAME: iri2007_VTEC_Scale_Factor.f
c**   
c**   DATE WRITTEN: 8/18/2010
c**   
c**   PROGRAMMER: Alex Fore
c**   
c**   FUNCTIONAL DESCRIPTION:  Uses IRI-2007 to compute a scale
c**    factor to give the fraction of the VTEC that is below
c**    the satellite/platform.
c**
c**   ROUTINES CALLED: 
c**   
c**   NOTES: 
c**
c**   UPDATE LOG:
c**
c**   Date Changed        Reason Changed        CR # and Version #
c**   ------------       ----------------       ------------------
c**     
c*****************************************************************   
      implicit none
c     PARAMETER STATEMENTS
      integer stdout, stderr, stdin
      parameter( stdout = 6, stderr = 0, stdin = 5 )
c     INPUT VARIABLES
      real*8 sc_position(3)
      real*8 earth_position_center(3)
      real*8 date(6)
c     OUTPUT VARIABLES
      real*8 tec_scale_factor
c     LOCAL VARIABLES
      character*320 media_a_iripath  ! path to the IRI data file directory
      
      real*8 r_look_unit(3)
      real*8 r_sc_pos_unit(3)
      real*8 iono_position(3)
      real*8 slant_range
      real*8 iono_lon
      real*8 iono_lat
      real*8 iono_alt
      real*8 iono_nadir_vector(3)
      
      real*8 sc_lon
      real*8 sc_lat
      real*8 sc_alt
      real*8 sc_nadir_vector(3)
      
      real*8 r_range
      real*8 r_antposdist
      
      logical good_solution

      real*4    media_f_alati     ! platform center position (latitude)
      real*4    media_f_along     ! platform center position (longitude)
      real*4    media_f_aheight   ! platform height (km)
      real*4    media_f_dhour     ! platform center time
      integer*4 media_i_mmdd      ! platform center day of year
      integer*4 media_i_yyyy      ! platform center year

C     Returned from background ionosphere simulation (IRI model)
      real*4 media_f_TEC_sat      ! Total Electron Content (in m^-2) up to platform
      real*4 media_f_TEC_20k      ! Total Electron Content (in m^-2) up to 20.2 km
      real*4 media_f_f2ht         ! F2 layer height (km)
      real*4 media_f_Babs         ! Magnitude of Geomagnetic field (gauss)
      real*4 media_f_Bdec         ! Magnetic field declination (rad)
      real*4 media_f_Binc         ! Magnetic field inclination (rad)

      logical media_l_ionoheight      ! over-riding height default?
      real*4  media_f_ionolayerheight ! over-ride ionosphere height      
      
      integer lnblnk

c     DATA STATEMENTS
c     COMMON BLOCKS
      integer i_dbg_level
      common /debug/ i_dbg_level
      
      character*200 AQ_SCAT_SIM_DIR_OUT, AQ_SCAT_SIM_DIR_DAT, AQ_SCAT_SIM_DIR_SCRATCH
      common /AQ_SCAT_SIM_DIRS/ AQ_SCAT_SIM_DIR_OUT, AQ_SCAT_SIM_DIR_DAT, AQ_SCAT_SIM_DIR_SCRATCH      
c     FUNCTION STATEMENTS
c     PROCESSING STEPS
      
      media_a_iripath = AQ_SCAT_SIM_DIR_DAT(1:lnblnk(AQ_SCAT_SIM_DIR_DAT))
     $ // '/faraday/data.iri_2007/'

C--Compute look unit vector from s/c to ground footprint.
      r_look_unit(1) = earth_position_center(1) - sc_position(1)
      r_look_unit(2) = earth_position_center(2) - sc_position(2)
      r_look_unit(3) = earth_position_center(3) - sc_position(3)
      
      call norm(r_look_unit,r_range)
      r_look_unit(1) = r_look_unit(1) / r_range
      r_look_unit(2) = r_look_unit(2) / r_range
      r_look_unit(3) = r_look_unit(3) / r_range

C--Compute unit vector that points from origin to s/c location.
      call norm(sc_position,r_antposdist)
      r_sc_pos_unit(1) = sc_position(1) / r_antposdist
      r_sc_pos_unit(2) = sc_position(2) / r_antposdist
      r_sc_pos_unit(3) = sc_position(3) / r_antposdist

C--Locate the cell on the ionosphere      
      call locate_cell_on_ionosphere( sc_position, r_look_unit, 
     $ iono_position, slant_range, good_solution )

      if( .not.good_solution )then
        write(6,*) 'ERROR: iri2007_VTEC_Scale_Factor: Imaginary solution found!'
!        pause
      endif

C--Compute spacecraft altitude above ellipsoid
      call compute_sc_nadir_components( sc_position, 
     $                                  sc_lon, 
     $                                  sc_lat,
     $                                  sc_alt,
     $                                  sc_nadir_vector)      

C--Compute lat, lon of pierce-point on ionosphere
      call compute_sc_nadir_components( iono_position, 
     $                                  iono_lon, 
     $                                  iono_lat,
     $                                  iono_alt,
     $                                  iono_nadir_vector)      

C--Set up call to compute_iri     
      media_f_alati   = iono_lat        ! media_f_alati is real*4, iono_lat is r*8
      media_f_along   = iono_lon    ! media_f_along is real*4, iono_lon is r*8
      media_f_aheight = sc_alt / 1000.0
      media_f_dhour   = date(4) + date(5) / 60.0d0
      media_i_mmdd    = int(date(3)) + 100 * int(date(2))
      media_i_yyyy    = int(date(1))
      
C--Compute VTEC @ platform height
      call compute_iri( media_a_iripath, media_l_ionoheight,
     &          media_f_ionolayerheight, media_f_alati,
     &          media_f_along, media_f_aheight,
     &          media_f_dhour, media_i_mmdd, media_i_yyyy,
     &          media_f_TEC_sat, media_f_f2ht, media_f_Babs,
     &          media_f_Bdec, media_f_Binc )

C--Compute VTEC @ 20,200 km altitude (as suggested by D. LeVine)
      media_f_aheight = 20200.0      
      call compute_iri( media_a_iripath, media_l_ionoheight,
     &          media_f_ionolayerheight, media_f_alati,
     &          media_f_along, media_f_aheight,
     &          media_f_dhour, media_i_mmdd, media_i_yyyy,
     &          media_f_TEC_20k, media_f_f2ht, media_f_Babs,
     &          media_f_Bdec, media_f_Binc )
      
      tec_scale_factor = media_f_TEC_sat / media_f_TEC_20k
      
      if( i_dbg_level.ge.2 ) then
        write(6,*) 'iri2007_VTEC_Scale_Factor: '
        write(6,'(a50,3f12.6)') 'r_look_unit: ',r_look_unit
        write(6,'(a50,3f12.6)') 'r_sc_pos_unit: ',r_sc_pos_unit
        write(6,'(a50,3f12.6)') 'tec_sat, tec_20k, factor: ', media_f_TEC_sat/1d16, media_f_TEC_20k/1d16, tec_scale_factor
      endif

      return
      end
