c*******************************************************************

      subroutine model_function(r_incidence_angle,r_wind_speed,r_direction_angle,i_sig_wave,
     &     r_sigma0)

c*******************************************************************
c**   
c**    FILE NAME:  model_function.f
c**
c**    DATE WRITTEN:   12-FEB-2007
c**
c**    PROGRAMMER:  Adam P. Freedman
c**
c**    FUNCTIONAL DESCRIPTION: This routine computes the values for sigma-0 based
c**     on Simon Yuehs model function. It takes incidence angle, wind speed, 
c**     wind direction, and a significant wave height parameter and returns
c**     sigma-0 for the ocean for each of three polarizations: HH, VV, and VH.
c**
c**    ROUTINES CALLED: 
c**
c**    INCLUDE FILES:  
c**
c**    NOTES: 
c**
c**    UPDATE LOG: 
c**
c**    Date Changed        Reason Changed                  CR # and Version #
c**    ------------       ----------------                 -----------------  
c**
c*******************************************************************

      implicit none

c     PARAMETER AND STRUCTURE DEFINITION FILES: 

      ! include 'constants.inc'
      integer I_N_WINDS
      parameter (I_N_WINDS = 50)
      integer I_N_INCANGS
      parameter (I_N_INCANGS = 51)
      integer I_N_DIRECTS
      parameter (I_N_DIRECTS = 72)
      integer I_N_POLS
      parameter (I_N_POLS = 9)
      integer I_N_SWAVS
      parameter (I_N_SWAVS = 4)

c     INPUT VARIABLES:

      real*8  r_incidence_angle     ! incidence angle in degrees of incident beam
      real*8  r_wind_speed          ! wind speed in meters/second
      real*8  r_direction_angle     ! angle between wind direction and azimuth angle of incident beam
      integer i_sig_wave            ! significant wave height indication flag (legitimate values of 1, 2, 3, or 4)

c     OUTPUT VARIABLES:

      real*8  r_sigma0(9)           ! array of sigma-nought values in dB, for polarizations (HH,HV,VV)

c     LOCAL VARIABLES:

      real*8  r_wnd_speeds(I_N_WINDS)
      real*8  r_inc_angles(I_N_INCANGS)
      real*8  r_directs(I_N_DIRECTS)
      character*2  a_polarizs(I_N_POLS)
      integer i_sigwaves(I_N_SWAVS)

      logical l_firsttime

      real*8 r_sigma0_arr(I_N_WINDS,I_N_INCANGS,I_N_POLS,I_N_SWAVS,I_N_DIRECTS)
                                ! zeros(Nu,Nthe,Npol,Nbl,Nphi);   sigma0 matrix (wind speed, incident angle, polarization, bl, wind direction)
      
      integer i_Nrows
      integer i_Ncols
      integer i_osn

      character*320  a_filepath
      character*320  a_filename(5)  ! (3)
      character*320  a_infile

      integer i_sw, i_inc, i_wnd, i_wdr
      integer i_ios, i, j, k
      real*4  f_temp(14)

      real*8  r_wndfrac,pi,linearHH,linearVH,linearVV
      real*8  r_incfrac
      real*8  r_dirfrac

      logical speed_lt_min_idx 

      character*200 AQ_SCAT_SIM_DIR_OUT, AQ_SCAT_SIM_DIR_DAT, AQ_SCAT_SIM_DIR_SCRATCH

c     COMMON BLOCKS:
      common /AQ_SCAT_SIM_DIRS/ AQ_SCAT_SIM_DIR_OUT, AQ_SCAT_SIM_DIR_DAT, AQ_SCAT_SIM_DIR_SCRATCH

c     EQUIVALENCE STATEMENTS: none

c     DATA STATEMENTS:
c10,11,12,13,14,15,16,17,18,& 19,20,21,22,23,24,,55,56,57,58,59,60
      data r_wnd_speeds/1,2,3,4,5,6,7,8,9,10,
     & 11,12,13,14,15,16,17,18,19,20,21,22,23,
     & 24,25,26,27,28,29,30,31,32,33,34,35,36,
     & 37,38,39,40,41,42,43,44,45,46,47,48,49,50/                      ! wind speeds in table in m/s
      data r_inc_angles/10,11,12,13,14,15,16,17,18,
     & 19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,
     & 34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,
     & 49,50,51,52,53,54,55,56,57,58,59,60/                                              ! incidence angles in table in degrees
      data r_directs/0,5,10,15,20,25,30,35,40,45,
     & 50,55,60,65,70,75,80,85,90,95,100,105,110,
     & 115,120,125,130,135,140,145,150,155,160,165,
     & 170,175,180,185,190,195,200,205,210,215,220,
     & 225,230,235,240,245,250,255,260,265,270,275,
     & 280,285,290,295,300,305,310,315,320,325,330,
     & 335,340,345,350,355/                                            ! direction angles in table in degrees
      data a_polarizs/'HH','HV','VV','M_RHO_HHVV','P_RHO_HHVV',
     & 'M_RHO_HHHV','P_RHO_HHHV','M_RHO_HVVV','P_RHO_HVVV'/                  ! polarization states in table

      data i_sigwaves/2,3,4,5/                         ! significant wave height parameter

      data l_firsttime /.TRUE./                        ! if first time in subroutine, read and store tables
      data i_Nrows / 72/, i_Ncols /14/

c     SAVE STATEMENTS:

      save r_sigma0_arr

c     FORMAT STATEMENTS:

 1001 format (/1X,A/)    ! skip line before and after
 1002 format (/1X,A)     ! skip line before
 1003 format (1X,A/)     ! skip line after

c     FUNCTION STATEMENTS:

 !!!     integer strlen
      integer lnblnk

c     PROCESSING STEPS:

c     Contents of input files:
c     Sample file name:  f1u1t30a0kd6ts15s35.dat
c     Contents of each line:  Ts,Salinity,U,theta,phi,sigma_hh,sigma_hv,sigma_vv,
c                             M_RHO_HHVV,P_RHO_HHVV,M_RHO_HHHV,P_RHO_HHHV,M_RHO_HVVV,P_RHO_HVVV
c     Models appear to have been run with Ts = 15.0 C, Salinity = 35.0 psu
c     We ignore complex cross-pol cross-correlations for Aquarius.
c     In each file, only the wind direction angle (phi) changes.
c     Thus, each file contains 5 lines, and 14 elements.

c     If first time in program, read all input files and save data

      if (l_firsttime) then
c         write(*,*) "First time: ",l_firsttime
ccc LOOP THROUGH READING FILES

c         a_filepath = '/Users/fore/Documents/AQUARIUS/scat_science/Aquarius_data/model_funct'
         
         
         a_filepath = AQ_SCAT_SIM_DIR_DAT(1:lnblnk(AQ_SCAT_SIM_DIR_DAT)) // '/model_funct'
         
         a_filename(1) = '/Bl00X'
         a_filename(2) = '/ltXX'
         a_filename(3) = '/f1uX'
         a_filename(4) = 'tXX'
         a_filename(5) = 'a0kd6ts15S35.dat'

         pi = 3.141592653589793d0

         do i_sw = 1, I_N_SWAVS
            do i_inc = 1, I_N_INCANGS
               do i_wnd = 1, I_N_WINDS
                  write(a_filename(1),'(''/Bl00'',I1)') i_sigwaves(i_sw)
                  write(a_filename(2),'(''/lt'',I2)') int(r_inc_angles(i_inc))
                  if (r_wnd_speeds(i_wnd).lt.10) then
                     write(a_filename(3),'(''/f1u'',I1,''t'',I2,
     $                    ''a0kd6ts15S35.dat'')')  int(r_wnd_speeds(i_wnd)), int(r_inc_angles(i_inc))
                  else
                     write(a_filename(3),'(''/f1u'',I2,''t'',I2,
     $                    ''a0kd6ts15S35.dat'')') 
     $                    int(r_wnd_speeds(i_wnd)), int(r_inc_angles(i_inc))
                  endif

                  a_infile = a_filepath(1:lnblnk(a_filepath))//a_filename(1)(1:lnblnk(a_filename(1)))
     $                 //a_filename(2)(1:lnblnk(a_filename(2)))
                  a_infile = a_infile(1:lnblnk(a_infile))//a_filename(3)(1:lnblnk(a_filename(3)))
                 !write(*,*) 'a_infile',a_infile !****************
                 ! open(unit=12,file=a_infile,status='old',iostat=i_ios,readonly,form='formatted')
                  open(unit=12,file=a_infile,iostat=i_ios,
     $                  form='formatted')
                  if (i_ios.ne.0) then
                     write(*,*) 'read error, ios = ', i_ios
                     return
                  end if
                  
                  do i_wdr = 1, i_Nrows
                     read(12,*,iostat=i_osn) (f_temp(j),j=1,i_Ncols)
                     r_sigma0_arr(i_wnd,i_inc,1,i_sw,i_wdr) = f_temp(6)
c                     f_temp(7)=-60.0d0
c                     f_temp(8)=f_temp(6)
                     r_sigma0_arr(i_wnd,i_inc,2,i_sw,i_wdr) = f_temp(7)
                     r_sigma0_arr(i_wnd,i_inc,3,i_sw,i_wdr) = f_temp(8)
c                     f_temp(9)=.8d0
c                     f_temp(10)=0.0d0
c                     f_temp(11)=0.0d0
c                     f_temp(13)=0.0d0
                     linearHH=10.0d0**(f_temp(6)/10.0d0)
                     linearVH=10.0d0**(f_temp(7)/10.0d0)
                     linearVV=10.0d0**(f_temp(8)/10.0d0)
                     r_sigma0_arr(i_wnd,i_inc,4,i_sw,i_wdr) = sqrt(linearHH*linearVV)*f_temp(9)*dcos(f_temp(10)*pi/180.d0)
                     r_sigma0_arr(i_wnd,i_inc,5,i_sw,i_wdr) = sqrt(linearHH*linearVV)*f_temp(9)*dsin(f_temp(10)*pi/180.d0)
                     r_sigma0_arr(i_wnd,i_inc,6,i_sw,i_wdr) = sqrt(linearHH*linearVH)*f_temp(11)*dcos(f_temp(12)*pi/180.d0)
                     r_sigma0_arr(i_wnd,i_inc,7,i_sw,i_wdr) = sqrt(linearHH*linearVH)*f_temp(11)*dsin(f_temp(12)*pi/180.d0)
                     r_sigma0_arr(i_wnd,i_inc,8,i_sw,i_wdr) = sqrt(linearVV*linearVH)*f_temp(13)*dcos(f_temp(14)*pi/180.d0)
                     r_sigma0_arr(i_wnd,i_inc,9,i_sw,i_wdr) = sqrt(linearVV*linearVH)*f_temp(13)*dsin(f_temp(14)*pi/180.d0)
                  enddo

                  close(12)

               enddo
            end do
         end do

         l_firsttime = .FALSE.
      end if
         
!open (unit=i_io, file=a_infile, status='old', iostat=i_ios, readonly, form='unformatted', access='direct', recl=34478)

c     Interpolate to desired wind speed, incidence angle, wind direction
 
c     Use multilinear interpolation
!**      subroutine model_function(r_incidence_angle,r_wind_speed,r_direction_angle,i_sig_wave,   r_sigma0)

c     Wind speed

      i_wnd = -99
      r_wndfrac = 0.0d0

      ! AGF modify 3/29/2011 to interpolate down to 0 m/s (with sigma0 == -Inf in dB at 0 m/s)
      if (r_wind_speed.lt.r_wnd_speeds(1))then
        i_wnd = 1
        speed_lt_min_idx = .true.
      else
        speed_lt_min_idx = .false.
      endif
      ! End AGF mods 3/29/2011 
      
      if (r_wind_speed.ge.r_wnd_speeds(I_N_WINDS)) i_wnd = I_N_WINDS
      if (i_wnd.lt.0) then
         do i = 1, I_N_WINDS-1
            if ((r_wind_speed.ge.r_wnd_speeds(i)).and.(r_wind_speed.lt.r_wnd_speeds(i+1))) then
               i_wnd = i
               r_wndfrac = (r_wind_speed - r_wnd_speeds(i))/(r_wnd_speeds(i+1) - r_wnd_speeds(i))
            end if
         end do
      end if

c     Incidence angle

      i_inc = -99
      r_incfrac = 0.0d0
      !if (r_incidence_angle.lt.r_inc_angles(1)) i_inc = 1
      !if (r_incidence_angle.ge.r_inc_angles(I_N_INCANGS)) i_inc = I_N_INCANGS
      if (r_incidence_angle.lt.r_inc_angles(1)) then
         i_inc = 1
         r_incfrac = (r_incidence_angle - r_inc_angles(1))/(r_inc_angles(2) - r_inc_angles(1))
      end if
      if (r_incidence_angle.ge.r_inc_angles(I_N_INCANGS)) then
         i_inc = I_N_INCANGS-1
         r_incfrac = (r_incidence_angle - r_inc_angles(I_N_INCANGS-1))/(r_inc_angles(I_N_INCANGS) - r_inc_angles(I_N_INCANGS-1))
      end if
      if (i_inc.lt.0) then
         do i = 1, I_N_INCANGS-1
            if ((r_incidence_angle.ge.r_inc_angles(i)).and.(r_incidence_angle.lt.r_inc_angles(i+1))) then
               i_inc = i
               r_incfrac = (r_incidence_angle - r_inc_angles(i))/(r_inc_angles(i+1) - r_inc_angles(i))
            end if
         end do
      end if

c     Direction angle

      i_wdr = -99
      r_dirfrac = 0.0
      if (r_direction_angle.eq.r_directs(I_N_DIRECTS)) i_wdr = I_N_DIRECTS
      do i = 1, I_N_DIRECTS-1
         if ((r_direction_angle.ge.r_directs(i)).and.(r_direction_angle.lt.r_directs(i+1))) then
            i_wdr = i
            r_dirfrac = (r_direction_angle - r_directs(i))/(r_directs(i+1) - r_directs(i))
         end if
      end do

c     Significant wave height

      i_sw = -99
      do i = 1, I_N_SWAVS
         if (i_sig_wave.eq.i_sigwaves(i)) then
            i_sw = i
         end if
      end do

      ! write(*,*) 'i_wnd, i_inc, i_wdr, i_sw ', i_wnd, i_inc, i_wdr, i_sw
      ! write(*,*) 'r_wndfrac, r_incfrac, r_dirfrac ', r_wndfrac, r_incfrac, r_dirfrac

c     Multi-dimensional linear interpolation

      if (r_incidence_angle.lt.10.d0) then
         r_sigma0(1) = +10.0d0
         r_sigma0(3) = +10.0d0
         r_sigma0(2) = -60.0d0
         r_sigma0(4) = 0.0d0
         r_sigma0(5) = 0.0d0
         r_sigma0(6) = 0.0d0
         r_sigma0(7) = 0.0d0
         r_sigma0(8) = 0.0d0
         r_sigma0(9) = 0.0d0
      else if (i_wnd+i_inc+i_wdr+i_sw .gt. 0) then
         do j = 1,9
            r_sigma0(j) = (1.0d0-r_wndfrac) * (1.0d0-r_incfrac) * (1.0d0-r_dirfrac) * r_sigma0_arr(i_wnd,i_inc,j,i_sw,i_wdr)
            
            r_sigma0(j) = r_sigma0(j) +     (r_wndfrac) * (1.0d0-r_incfrac) * (1.0d0-r_dirfrac) * 
     $           r_sigma0_arr(i_wnd+1,i_inc,j,i_sw,i_wdr)
            r_sigma0(j) = r_sigma0(j) + (1.0d0-r_wndfrac) *     (r_incfrac) * (1.0d0-r_dirfrac) * 
     $           r_sigma0_arr(i_wnd,i_inc+1,j,i_sw,i_wdr)
            r_sigma0(j) = r_sigma0(j) + (1.0d0-r_wndfrac) * (1.0d0-r_incfrac) *     (r_dirfrac) * 
     $           r_sigma0_arr(i_wnd,i_inc,j,i_sw,i_wdr+1)
            
            r_sigma0(j) = r_sigma0(j) + (1.0d0-r_wndfrac) *     (r_incfrac) *     (r_dirfrac) * 
     $           r_sigma0_arr(i_wnd,i_inc+1,j,i_sw,i_wdr+1)
            r_sigma0(j) = r_sigma0(j) +     (r_wndfrac) * (1.0d0-r_incfrac) *     (r_dirfrac) * 
     $           r_sigma0_arr(i_wnd+1,i_inc,j,i_sw,i_wdr+1)
            r_sigma0(j) = r_sigma0(j) +     (r_wndfrac) *     (r_incfrac) * (1.0d0-r_dirfrac) * 
     $           r_sigma0_arr(i_wnd+1,i_inc+1,j,i_sw,i_wdr)
            
            r_sigma0(j) = r_sigma0(j) +     (r_wndfrac) *     (r_incfrac) *     (r_dirfrac) * 
     $           r_sigma0_arr(i_wnd+1,i_inc+1,j,i_sw,i_wdr+1)
         end do

! AGF modify 3/29/2011 to interpolate down to 0 m/s (with sigma0 == -Inf in dB at 0 m/s)
! check for wind speed less than 1st posting point in table
         if( speed_lt_min_idx ) then
! only modify HH, HV, and VV elements
            r_sigma0(1) = r_sigma0(1) + 10.d0*log10( r_wind_speed / r_wnd_speeds(1) )
            r_sigma0(2) = r_sigma0(2) + 10.d0*log10( r_wind_speed / r_wnd_speeds(1) )
            r_sigma0(3) = r_sigma0(3) + 10.d0*log10( r_wind_speed / r_wnd_speeds(1) )
         endif
! End AGF mods 3/29/2011          

      else
         r_sigma0(1) = -999.d0
         r_sigma0(2) = -999.d0
         r_sigma0(3) = -999.d0
         r_sigma0(4) = -999.d0
         r_sigma0(5) = -999.d0
         r_sigma0(6) = -999.d0
         r_sigma0(7) = -999.d0
         r_sigma0(8) = -999.d0
         r_sigma0(9) = -999.d0
      endif
      
      ! write(*,*) 'r_sigma0(1) ', r_sigma0(1)  !********
      ! write(*,*) (((r_sigma0_arr(i_wnd+i,i_inc+j,1,i_sw,i_wdr+k),i=0,1),j=0,1),k=0,1)  !*******


c     END OF SUBROUTINE

      return

      end

c *******************************************************************


c****************************************************************

      function strlen(a_string)

c****************************************************************
c**   Modified from rdflen of rdf_reader.f of Scott Shaffer
c*****************************************************************

c     INPUT VARIABLES:

      character*(*) a_string
      integer strlen

c     LOCAL VARIABLES:

      integer i_len

c     PROCESSING STEPS:

      i_len=len(a_string)

      do while(i_len .gt. 0 .and. (a_string(i_len:i_len) .eq. ' ' .or. 
     &     ichar(a_string(i_len:i_len)) .eq. 0))
         i_len=i_len-1
      enddo

      strlen=i_len

      return

      end
