c*****************************************************************      
      subroutine points_edge( sc_position, sc_velocity, yaw_deg,
     $ pitch_deg, roll_deg, i_beam, lat_points, lon_points,
     $ all_good_solutions )
c*****************************************************************      
c**   
c**   FILE NAME:  points_edge.f
c**   
c**   DATE WRITTEN:  2/4/2009
c**   
c**   PROGRAMMER:  Alex Fore
c**   
c**   FUNCTIONAL DESCRIPTION:  Uses the beam edge points in the 
c**    common block /beam_edge_points/ to compute the lat,lon 
c**    coordinates of the 3 dB points on the ground.  
c**
c**    You must populate the commmon block /beam_edge_points/
c**    using the subroutine find_edge_points before using this
c**    subroutine.
c**
c**   ROUTINES CALLED: 
c**   
c**   NOTES: 
c**
c**   UPDATE LOG: Updated 8/18/2009 to set a logical 
c**    all_good_solutions to .false.
c**    if any of the edge points are off-earth.
c**
c**   Date Changed        Reason Changed        CR # and Version #
c**   ------------       ----------------       ------------------
c**     
c*****************************************************************   
      implicit none
c     PARAMETER STATEMENTS
c     INPUT VARIABLES
      real*8 sc_position(3) ! units?
      real*8 sc_velocity(3) ! units?
      real*8 yaw_deg            ! must be in degrees...
      real*8 pitch_deg          ! must be in degrees...
      real*8 roll_deg           ! must be in degrees...
      integer i_beam            ! which beam   (1-3)

c     OUTPUT VARIABLES
      real*8 lat_points(4), lon_points(4)
      logical all_good_solutions
      
c     LOCAL VARIABLES
      integer i_pol                     ! polarization ( 1 - cross-pol, 2 - VV co-pol, 3 - HH co-pol)
      real*8 vector_beam(3)
      real*8 slant_range
      real*8 earth_position(3)
      real*8 lat_geocentric_deg, lat_deg, lon_deg
      real*8 theta_B, phi_B
      real*8 PI
      integer ii, jj
      logical good_solution

c     DATA STATEMENTS
c     COMMON BLOCKS
      integer i_dbg_level
      common /debug_level/ i_dbg_level
      
      real*8 theta_B_edge(3,3,4)
      real*8 phi_B_edge(3,3,4)
      common /beam_edge_points/ theta_B_edge, phi_B_edge
c     FUNCTION STATEMENTS
c     PROCESSING STEPS
      PI = 4.0d0 * atan( 1.0d0 )
      
      all_good_solutions = .true.
      
      do ii = 1,4
c        theta_B = theta_B_edge(i_beam,i_pol,ii)
c        phi_B   = phi_B_edge(i_beam,i_pol,ii)
C--Use average beam edge points since the difference between the 
C--edge points of the different polarizations is ~ a few hundreths
C--of a degree (in lat,lon space) when projected on the ground.
C
C--You can use the ones for each polarization by adding i_pol to 
C--argument list, and replacing the averaging code with the 
C--commented out assignments above this comment.
        theta_B = 0.0d0
        phi_B = 0.0d0
        do i_pol=2,3
          theta_B = theta_B + theta_B_edge(i_beam,i_pol,ii) / 2.0d0 
          phi_B   = phi_B + phi_B_edge(i_beam,i_pol,ii) / 2.0d0
        enddo
        
        vector_beam(1) = sin( theta_B * PI / 180.0d0 ) * cos( phi_B * PI / 180.0d0 )
        vector_beam(2) = sin( theta_B * PI / 180.0d0 ) * sin( phi_B * PI / 180.0d0 )
        vector_beam(3) = cos( theta_B * PI / 180.0d0 )
        
        call geolocate( vector_beam, sc_position, sc_velocity,
     $ yaw_deg, pitch_deg, roll_deg, i_beam, earth_position,
     $ slant_range, good_solution )
        
        if( good_solution )then
          call compute_lat_long( earth_position, lat_geocentric_deg, lat_deg, lon_deg)
          lat_points(ii) = lat_deg
          lon_points(ii) = lon_deg
        else
          !write(6,*) 'Imaginary solution found in points_edge!'
          lat_points(ii) = 0.0d0 ! what should be the value if error?
          lon_points(ii) = 0.0d0
          all_good_solutions = .false.
        endif
        
      enddo
      end
      
      
c*****************************************************************      
      subroutine find_edge_points( antenna_pattern_relative_1h,
     $ antenna_pattern_relative_1v, antenna_pattern_relative_2h,
     $ antenna_pattern_relative_2v, antenna_pattern_relative_3h,
     $ antenna_pattern_relative_3v )
c*****************************************************************      
c**   
c**   FILE NAME:   points_edge.f
c**   
c**   DATE WRITTEN:  2/3/2009
c**   
c**   PROGRAMMER:  Alex Fore
c**   
c**   FUNCTIONAL DESCRIPTION:  Fills the common block with the 
c**    edge points in the beam coordinate system.  They are
c**    always the same points in the beam system, so we call this 
c**    subroutine once at the start to compute them.  
c**
c**   ROUTINES CALLED: 
c**   
c**   NOTES: 
c**
c**   UPDATE LOG:
c**
c**   Date Changed        Reason Changed        CR # and Version #
c**   ------------       ----------------       ------------------
c**     
c*****************************************************************   
      implicit none
c     PARAMETER STATEMENTS
c     INPUT VARIABLES
      real*8 antenna_pattern_relative_1h(360,721,6)
      real*8 antenna_pattern_relative_1v(360,721,6)
      real*8 antenna_pattern_relative_2h(360,721,6)
      real*8 antenna_pattern_relative_2v(360,721,6)
      real*8 antenna_pattern_relative_3h(360,721,6)
      real*8 antenna_pattern_relative_3v(360,721,6)    
c     OUTPUT VARIABLES
c     LOCAL VARIABLES
      real*8 phi_points(4), theta_points(4)
      integer i_point
c     DATA STATEMENTS
c     COMMON BLOCKS
      integer i_dbg_level
      common /debug_level/ i_dbg_level
      
      real*8 theta_B_edge(3,3,4)
      real*8 phi_B_edge(3,3,4)
      common /beam_edge_points/ theta_B_edge, phi_B_edge
c     FUNCTION STATEMENTS
c     PROCESSING STEPS
      
C--Beam 1 TX H RX V--      
      call find_3db_2_way_points( antenna_pattern_relative_1h, 
     $ antenna_pattern_relative_1v, phi_points, theta_points )      
     
      do i_point = 1,4
        theta_B_edge(1,1,i_point) = theta_points(i_point)
        phi_B_edge(1,1,i_point)   = phi_points(i_point)
      enddo
      
C--Beam 1 TX V RX V--      
      call find_3db_2_way_points( antenna_pattern_relative_1v, 
     $ antenna_pattern_relative_1v, phi_points, theta_points )

      do i_point = 1,4
        theta_B_edge(1,2,i_point) = theta_points(i_point)
        phi_B_edge(1,2,i_point)   = phi_points(i_point)
      enddo

C--Beam 1 TX H RX H-- 
      call find_3db_2_way_points( antenna_pattern_relative_1h, 
     $ antenna_pattern_relative_1h, phi_points, theta_points )
      do i_point = 1,4
        theta_B_edge(1,3,i_point) = theta_points(i_point)
        phi_B_edge(1,3,i_point)   = phi_points(i_point)
      enddo


C--Beam 2 TX H RX V--      
      call find_3db_2_way_points( antenna_pattern_relative_2h, 
     $ antenna_pattern_relative_2v, phi_points, theta_points )      
      do i_point = 1,4
        theta_B_edge(2,1,i_point) = theta_points(i_point)
        phi_B_edge(2,1,i_point)   = phi_points(i_point)
      enddo
      
C--Beam 2 TX V RX V--      
      call find_3db_2_way_points( antenna_pattern_relative_2v, 
     $ antenna_pattern_relative_2v, phi_points, theta_points )
      do i_point = 1,4
        theta_B_edge(2,2,i_point) = theta_points(i_point)
        phi_B_edge(2,2,i_point)   = phi_points(i_point)
      enddo

C--Beam 2 TX H RX H-- 
      call find_3db_2_way_points( antenna_pattern_relative_2h, 
     $ antenna_pattern_relative_2h, phi_points, theta_points )
      do i_point = 1,4
        theta_B_edge(2,3,i_point) = theta_points(i_point)
        phi_B_edge(2,3,i_point)   = phi_points(i_point)
      enddo


C--Beam 3 TX H RX V--      
      call find_3db_2_way_points( antenna_pattern_relative_3h, 
     $ antenna_pattern_relative_3v, phi_points, theta_points )      
      do i_point = 1,4
        theta_B_edge(3,1,i_point) = theta_points(i_point)
        phi_B_edge(3,1,i_point)   = phi_points(i_point)
      enddo
      
C--Beam 3 TX V RX V--      
      call find_3db_2_way_points( antenna_pattern_relative_3v, 
     $ antenna_pattern_relative_3v, phi_points, theta_points )
      do i_point = 1,4
        theta_B_edge(3,2,i_point) = theta_points(i_point)
        phi_B_edge(3,2,i_point)   = phi_points(i_point)
      enddo

C--Beam 3 TX H RX H-- 
      call find_3db_2_way_points( antenna_pattern_relative_3h, 
     $ antenna_pattern_relative_3h, phi_points, theta_points )
      do i_point = 1,4
        theta_B_edge(3,3,i_point) = theta_points(i_point)
        phi_B_edge(3,3,i_point)   = phi_points(i_point)
      enddo
      end

c*****************************************************************      
      subroutine find_3db_2_way_points( tx_patt, rx_patt, 
     $ phi_points, theta_points )
c*****************************************************************      
c**   
c**   FILE NAME:  points_edge.f
c**   
c**   DATE WRITTEN:  2/3/2009
c**   
c**   PROGRAMMER:  Alex Fore
c**   
c**   FUNCTIONAL DESCRIPTION:  Given a tx one-way pattern, and a 
c**    rx one-way pattern we find the two-way 3 dB points in the
c**    beam coordinate system.  We only consider two cuts across
c**    boresight, one with phi = 0 and another with phi = 90.  We
c**    then compute the theta values where it crosses the -3dB
c**    line. (recall that these antenna patterns have theta in 
c**    -180,180) 
c**
c**    One dimensional linear interpolation is used to improve 
c**    accuracy of the 3dB points.
c**
c**   ROUTINES CALLED: 
c**   
c**   NOTES: 
c**
c**   UPDATE LOG:
c**
c**   Date Changed        Reason Changed        CR # and Version #
c**   ------------       ----------------       ------------------
c**     
c*****************************************************************   
      implicit none
c     PARAMETER STATEMENTS
c     INPUT VARIABLES
      real*8 tx_patt(360,721,6)
      real*8 rx_patt(360,721,6)
c     OUTPUT VARIABLES
      real*8 phi_points(4), theta_points(4)
c     LOCAL VARIABLES
      real*8 two_way_gain_1, two_way_gain_2
      real*8 phi_B, theta_B_1, theta_B_2
      integer i_phi_patt, i_theta_1_patt, i_theta_2_patt
      integer i_theta, i_phi
      real*8 theta_B_low, theta_B_high
c     DATA STATEMENTS
c     COMMON BLOCKS
      integer i_dbg_level
      common /debug_level/ i_dbg_level
c     FUNCTION STATEMENTS
c     PROCESSING STEPS
             
      do i_phi = 0,1
        phi_B = real(i_phi) * 90.0d0

        phi_points(i_phi+1)   = phi_B
        phi_points(i_phi+1+2) = phi_B
        
        i_phi_patt   = 1 + nint( 359.0d0 * phi_B / 180.0d0 )
        
        theta_B_low  = 0
        theta_B_high = 0
        
        do i_theta = -30,-1,1
          theta_B_1 = real( i_theta ) * 0.5d0
          theta_B_2 = real( i_theta + 1 ) * 0.5d0
          
          i_theta_1_patt = 1 + nint( 720.0d0 * (theta_B_1+180.0d0) / 360.0d0 )     
          i_theta_2_patt = 1 + nint( 720.0d0 * (theta_B_2+180.0d0) / 360.0d0 )
          
          two_way_gain_1 = tx_patt(i_phi_patt,i_theta_1_patt,1) +
     $     rx_patt(i_phi_patt,i_theta_1_patt,1)
          two_way_gain_2 = tx_patt(i_phi_patt,i_theta_2_patt,1) +
     $     rx_patt(i_phi_patt,i_theta_2_patt,1)

C--Test if the 2-way gain crosses -3 dB in the interval [theta_B_1,theta_B_2]   
C----(There is an implicit assumption that the 2-way gain is monotonic
C-----as it approaches the max gain at theta_B = 0).
          if( two_way_gain_1 .lt. -3.0d0 .and. two_way_gain_2 .gt. -3.0d0)then
C--One dimensional interpolation...          
            theta_B_low = theta_B_1 + (theta_B_2-theta_B_1) * 
     $       ( -3.0d0 - two_way_gain_1 ) / ( two_way_gain_2 - two_way_gain_1 )
          endif
        enddo
        
C--Same logic for the other side (positive theta_B) of the cut across boresight.        
        do i_theta = 30,1,-1
          theta_B_1 = real( i_theta ) * 0.5d0
          theta_B_2 = real( i_theta - 1 ) * 0.5d0
          
          i_theta_1_patt = 1 + nint( 720.0d0 * (theta_B_1+180.0d0) / 360.0d0 )     
          i_theta_2_patt = 1 + nint( 720.0d0 * (theta_B_2+180.0d0) / 360.0d0 )
          
          two_way_gain_1 = tx_patt(i_phi_patt,i_theta_1_patt,1) + 
     $     rx_patt(i_phi_patt,i_theta_1_patt,1)
          two_way_gain_2 = tx_patt(i_phi_patt,i_theta_2_patt,1) + 
     $     rx_patt(i_phi_patt,i_theta_2_patt,1)
          
          if( two_way_gain_1 .lt. -3.0d0 .and. two_way_gain_2 .gt. -3.0d0)then
            theta_B_high = theta_B_1 + (theta_B_2-theta_B_1) *
     $       ( -3.0d0 - two_way_gain_1 ) / ( two_way_gain_2 - two_way_gain_1 )
          endif
        enddo        
        
        theta_points(i_phi+1)   = theta_B_high
        theta_points(i_phi+1+2) = theta_B_low
        
      enddo
      if( i_dbg_level.ge.2)then
        write(6,*) 'theta_points, phi_points:'
        do i_phi=1,4
          write(6,*) theta_points(i_phi), phi_points(i_phi)
        enddo
      endif
      end