c****************************************************************
c**
c**   Time functions routines
c**
c****************************************************************
c**
c**   How to use these Fortran routines:
c**
c**   To go from GPS second to anything else, use
c**       call gps2time(r_gpssec,r_juldate,r_juldatelps,
c**     &   i_yyyy,i_mm,i_dd,i_hh,i_mn,r_ss,i_doy,r_sod)
c**   Given a time in GPS seconds, it returns the Julian date (r_juldate),
c**   and then dates and times corrected for leap seconds, including
c**   a leap-second corrected Julian date (r_juldatelps), the
c**   year, month, day, hour, minute, and second, the doy-of-year
c**   and the second of day.
c**   If the GPS time occurs during an actual leap second, the time
c**   returned in 23:59:60.xxx of the final day of the month, and the
c**   second-of-day is 86400.xxx.
c**
c**   (Note that in the calling sequence, r_xxx indicates a real*8
c**   variable, while i_xxx indicates an integer variable.)
c**
c**   To go from UTC to GPS time, use one of the following:
c**        YMDHMS2gps(i_yyyy,i_mm,i_dd,i_hh,i_mn,r_ss,r_gpssec,i_leap)
c**        YdoyHMS2gps(i_yyyy,i_doy,i_hh,i_mn,r_ss,r_gpssec,i_leap)
c**        YMDsod2gps(i_yyyy,i_mm,i_dd,r_sod,r_gpssec,i_leap)
c**        Ydoysod2gps(i_yyyy,i_doy,r_sod,r_gpssec,i_leap)
c**   In these routines i_leap is an input parameter normally set equal
c**   to zero. If the input time is an actual leap second, the UTC time
c**   can have the following commonly used values:
c**           a)  23:59:59.xxx   or sod 86399.xxx
c**           b)  23:59:60.xxx   or sod 86400.xxx
c**           c)  00:00:00.xxx   or sod 00000.xxx
c**   If format (b) is used, the routines know that this is a leap second
c**   and make the correction automatically.
c**   If format (a) or (c) is used, the i_leap flag must be set equal to
c**   one so that the routines know this is the leap second with that time
c**   tag, and not the normal UTC second with that time tag.
c**   
c**   Once you have GPS time, the Julian date can be obtained directly by
c**      secgps2jd(r_sec)
c**   and the inverse   jd2secgps(r_juldate)
c**
c**   Note that, since both Julian date and the GPS second have such large magnitudes,
c**   even using real*8 the time tag is not accurate at the microsecond level due to
c**   roundoff errors. For more accuracy, a smaller integer offset should be used, or the
c**   integer and fractional portions of the second or day should be dealt with separately.
c**
c**
c****************************************************************
 


c****************************************************************

      subroutine gps2time(r_gpssec,r_juldate,r_juldatelps,
     &   i_yyyy,i_mm,i_dd,i_hh,i_mn,r_ss,i_doy,r_sod)

c****************************************************************
c**
c**   FILE NAME: 
c**     
c**   DATE WRITTEN: 7/24/2009
c**     
c**   PROGRAMMER: Adam Freedman
c**     
c**   FUNCTIONAL DESCRIPTION: 
c**      This subroutine takes a GPS second value as input and outputs a
c**             number of time quantities, including julian date (no leap second
c**             correction), julian date after leap seconds have been applied 
c**             (possible duplicate seconds correspond to leap second), UTC time
c**             as year, month, day, hour, minute, second, day of year, and 
c**             seconds of day.
c**  
c**   ROUTINES CALLED: 
c**     
c**   NOTES: 
c**     
c**   UPDATE LOG:
c**
c**   Date Changed        Reason Changed                  CR # and Version #
c**   ------------       ----------------                 -----------------
c**     
c*****************************************************************

      implicit none

c     PARAMETER STATEMENTS:

c     INPUT VARIABLES:

      real*8 r_gpssec   ! seconds of time since GPS epoch of 1/6/1980
      
c     OUTPUT VARIABLES:
      
      real*8 r_juldate
      real*8 r_juldatelps
      integer i_yyyy, i_mm, i_dd
      integer i_hh, i_mn 
      real*8 r_ss
      integer i_doy
      real*8 r_sod
      
c     LOCAL VARIABLES:
     
      integer i_lps     ! leap seconds at this epoch
      integer i_lpf     ! output leap second flag
      integer i_ss
      real*8  r_frac
      real*8  r_jultemp
      
c     COMMON BLOCKS:
      
c     EQUIVALENCE STATEMENTS:
      
c     DATA STATEMENTS:
      
c     FUNCTION STATEMENTS:

      real*8  secgps2jd

c     PROCESSING STEPS:
      
c     Convert input time into julian date  

      r_juldate = secgps2jd(r_gpssec) ! / SECDAY + GPSREF
      
c         Adjust julian date for leap seconds
      
      call leap_sec_corr(r_juldate,1,0,r_juldatelps,i_lps,i_lpf)
       
c     Compute calendar date and time
      
      call JD2CAL_v2(r_juldatelps,i_yyyy,i_mm,i_dd,i_hh,i_mn,i_ss,r_frac,i_doy,r_sod)
        
      if (i_lpf.eq.2) then !i_ss = i_ss + 1
         r_jultemp = r_juldatelps - 1.d0/86400.d0
         call JD2CAL_v2(r_jultemp,i_yyyy,i_mm,i_dd,i_hh,i_mn,i_ss,r_frac,i_doy,r_sod)
         i_ss = i_ss + 1
      endif    
      
      r_ss = i_ss + r_frac
       
      return
      
      end
      
c****************************************************************

      subroutine JD2CAL_v2(JD,YEAR,MONTH,DAY,HOUR,MINUTE,SECOND,FRAC,DOY,SOD)

c****************************************************************
c**     
c**   FILE NAME: 
c**     
c**   DATE WRITTEN: 1/1/91
c**     
c**   PROGRAMMER: Frank Webb
c**     
c**   FUNCTIONAL DESCRIPTION:
c**
c**   This sub routine (Julian Date 2 CALendar date) takes an input Julian date  
c**   and returns the various components of the corresponding calendar date.     
c**   The components of the calendar date are all returned as numbers to allow   
c**   for use in computation. For instance, the month is returned as the         
c**   integer month number rather than as a character string.                    
c**     
c**   ROUTINES CALLED: J2DATE
c**     
c**   NOTES: 
c**
c**   JD   is the Julian date.           
c**                                      
c**   YEAR     is the year.              
c**   MONTH    is the month number.      
c**   DAY      is the day.               
c**   HOUR     is the hour.              
c**   MINUTE   is the minute.            
c**   SECOND   is the second.            
c**   FRAC     is the fractional seconds.
c**   DOY      is the day of year.
c**   SOD      is the seconds of day.
c**  
c**   UPDATE LOG:
c**
c**   Date Changed        Reason Changed                  CR # and Version #
c**   ------------       ----------------                 -----------------
c**     7/24/2009               Added DOY and SOD for Aquarius
c**
c*****************************************************************

      implicit none

c     PARAMETER STATEMENTS:

      real*8             SECDAY
      PARAMETER        ( SECDAY = 86400.0D0 )

c     INPUT VARIABLES:

      real*8             JD
        
c     OUTPUT VARIABLES:

      INTEGER            YEAR
      INTEGER            MONTH
      INTEGER            DAY
      INTEGER            HOUR
      INTEGER            MINUTE
      INTEGER            SECOND
      real*8             FRAC
      integer            DOY
      real*8             SOD

c     LOCAL VARIABLES:

      real*8             JDPLUS
      INTEGER            JDINT
      real*8             DSEC
      INTEGER            ISEC
      integer            JDYEAR

c     COMMON BLOCKS:

c     EQUIVALENCE STATEMENTS:

c     DATA STATEMENTS:

C     FUNCTION STATEMENTS:

      integer            DATE2J

c     PROCESSING STEPS:

c    Compute JDINT, the integer Julian date at noon of the current day
c    compute DSEC, the number of seconds elapsed since the start of the
c    current day.

      JDPLUS = JD + 0.5D0
      JDINT  = INT( JDPLUS )
      DSEC   = SECDAY * MOD( JDPLUS, 1.0D0 )
      IF( DSEC .GE. SECDAY )THEN
         JDINT = JDINT + 1
         DSEC  = DSEC  - SECDAY
      END IF
      ISEC = INT( DSEC )
      SOD = ISEC

c     Call J2DATE with JDINT to compute the year, month, and day of the
c     calendar date.

      CALL J2DATE( JDINT, YEAR, MONTH, DAY )

c     Compute HOUR.

      HOUR   = ISEC/3600
      ISEC   = ISEC - 3600*HOUR

c     Compute MINUTE.

      MINUTE = ISEC/60
      ISEC   = ISEC - 60*MINUTE

c     Compute SECOND.
      SECOND = ISEC

c     Compute FRAC.
      FRAC = MOD( DSEC, 1.0D0 )
      
c     Compute DOY

      JDYEAR =  DATE2J( YEAR, 1, 1 )
        
      ! DOY = INT(JD - JDYEAR - 0.5) + 1
      DOY = JDINT - JDYEAR + 1
      
      SOD = SOD + FRAC

      end  

Cc****************************************************************
C
C      subroutine J2DATE(JD,YEAR,MONTH,DAY)
C
Cc****************************************************************
Cc**     
Cc**   FILE NAME: 
Cc**     
Cc**   DATE WRITTEN: 1/1/91
Cc**     
Cc**   PROGRAMMER: Frank Webb
Cc**     
Cc**   FUNCTIONAL DESCRIPTION: 
Cc**
Cc**   This sub routine (Julian date 2 calendar DATE) converts the input integer    
Cc**   Julian date to the corresponding Gregorian calendar date. Since the Julian  
Cc**   date is an integer, this correspondence is exact for noon of the caledar    
Cc**   date.                                                                       
Cc**                                                                               
Cc**   The algorithm for this conversion is taken from the following article       
Cc**   Tantzen,R.T., "Communications of the ACM", Volume 6, Number 8, August       
Cc**   Algorithm 199, page 444.                                                    
Cc**  
Cc**   ROUTINES CALLED: None
Cc**     
Cc**   NOTES: 
Cc**
Cc**   JD       is the integer Julian date.  
Cc**                                         
Cc**   YEAR     is the year number.          
Cc**   MONTH    is the month number.         
Cc**   DAY      is the day number.           
Cc**  
Cc**   UPDATE LOG:
Cc**
Cc**   Date Changed        Reason Changed                  CR # and Version #
Cc**   ------------       ----------------                 -----------------
Cc**     
Cc*****************************************************************
C
C      implicit none
C
Cc     INPUT VARIABLES:
C
C      INTEGER            JD
C       
Cc     OUTPUT VARIABLES:
C
C      INTEGER            YEAR
C      INTEGER            MONTH
C      INTEGER            DAY
C
Cc     LOCAL VARIABLES:
C
C      INTEGER            J
C      INTEGER            Y
C      INTEGER            M
C      INTEGER            D
C
Cc     COMMON BLOCKS:
C
Cc     EQUIVALENCE STATEMENTS:
C
Cc     DATA STATEMENTS:
C
CC     FUNCTION STATEMENTS:
C
Cc     PROCESSING STEPS:
C
C      J = JD
C
C      J = J - 1721119
C      Y = (4*J-1)/146097
C      J = 4*J - 1 - 146097*Y
C      D = J/4
C      J = (4*D+3)/1461
C      D = 4*D + 3 -1461*J
C      D = (D+4)/4
C      M = (5*D-3)/153
C      D = 5*D - 3 - 153*M
C      D = (D+5)/5
C      Y = 100*Y + J
C      IF( M .LT. 10 )THEN
C         M = M + 3
C      ELSE
C         M = M - 9
C         Y = Y + 1
C      END IF
C
C      YEAR   = Y
C      MONTH  = M
C      DAY    = D
C       
C      end  
C 
 

      

c****************************************************************
            
      subroutine leap_sec_corr(r_timein,i_opt,i_leap,r_timeout,i_lps,i_lpf)

c****************************************************************    
c**     
c**   FILE NAME: 
c**     
c**   DATE WRITTEN: 7/24/2009
c**     
c**   PROGRAMMER: Adam Freedman
c**     
c**   FUNCTIONAL DESCRIPTION: 
c**
c**   This subroutine computes the offset between GPS time and UTC time 
c**   (GPS-UTC) in seconds, and adjusts the time for the leapsecond.
c**   It relies on the leap second table stored in an internal table, so 
c**   that the table must be modified and the source recompiled when
c**   leap seconds are made.
c**   The i_opt flag tells what sort of time is being passed in:
c**     i_opt = 1  =>  r_timein is in julian days
c**     i_opt = 2  =>  r_timein is in GPS seconds since epoch
c**   i_leap indicates whether to add or remove leap seconds:
c**             i_leap = 0 => no leap seconds present in r_timein, correct for them (GPS -> UTC)
c**     i_leap = 1 => leap seconds are present in r_timein, remove them (UTC -> GPS)
c**     i_leap = 2 => r_timein in UTC is exact leap second (needs special treatment)
c**   r_timeout is the adjusted time, after leap second corrections
c**   i_lps is the number of leap seconds added or subtracted at this epoch
c**   i_lpf is a leap second flag, similar to i_leap, for r_timeout. If r_timeout is an exact
c**     leap second, i_lpf is set to 2, and the appropriate UTC time of day is 23:59:60 
c**  
c**   ROUTINES CALLED: 
c**     
c**   NOTES: 
c**     
c**   UPDATE LOG:
c**
c**   Date Changed        Reason Changed                  CR # and Version #
c**   ------------       ----------------                 -----------------
c**     
c*****************************************************************

      implicit none

c     PARAMETER STATEMENTS:

      integer   I_MAX
      PARAMETER ( I_MAX = 20 )
      real*8   SECDAY
      PARAMETER(SECDAY=86400.0D0)

c     INPUT VARIABLES:

      real*8 r_timein
      integer i_opt
      integer i_leap

c     OUTPUT VARIABLES:
      
      real*8 r_timeout
      integer i_lps
      integer i_lpf
      
c     LOCAL VARIABLES:
      
      real*8 r_jd, r_jd0
      real*8 r_jdindex(0:I_MAX) ! lower bound of zero to avoid error if i_curr=0 is index
      integer i_gps_utc(I_MAX)
      integer i_curr /0/
      integer j
      integer i_oldleap
      real*8  r_wnd
      logical l_found
      
      save i_curr, i_oldleap
      
c     COMMON BLOCKS:
      
c     EQUIVALENCE STATEMENTS:
      
c     DATA STATEMENTS:
      
      data r_jdindex( 1), i_gps_utc( 1) / 2444239.5 ,  0 / ! 1 jan 1980
      data r_jdindex( 2), i_gps_utc( 2) / 2444786.5 ,  1 / ! 1 jul 1981
      data r_jdindex( 3), i_gps_utc( 3) / 2445151.5 ,  2 / ! 1 jul 1982
      data r_jdindex( 4), i_gps_utc( 4) / 2445516.5 ,  3 / ! 1 jul 1983
      data r_jdindex( 5), i_gps_utc( 5) / 2446247.5 ,  4 / ! 1 jul 1985
      data r_jdindex( 6), i_gps_utc( 6) / 2447161.5 ,  5 / ! 1 jan 1988
      data r_jdindex( 7), i_gps_utc( 7) / 2447892.5 ,  6 / ! 1 jan 1990
      data r_jdindex( 8), i_gps_utc( 8) / 2448257.5 ,  7 / ! 1 jan 1991
      data r_jdindex( 9), i_gps_utc( 9) / 2448804.5 ,  8 / ! 1 jul 1992
      data r_jdindex(10), i_gps_utc(10) / 2449169.5 ,  9 / ! 1 jul 1993
      data r_jdindex(11), i_gps_utc(11) / 2449534.5 , 10 / ! 1 jul 1994           
      data r_jdindex(12), i_gps_utc(12) / 2450083.5 , 11 / ! 1 jan 1996
      data r_jdindex(13), i_gps_utc(13) / 2450630.5 , 12 / ! 1 jul 1997
      data r_jdindex(14), i_gps_utc(14) / 2451179.5 , 13 / ! 1 jan 1999
      data r_jdindex(15), i_gps_utc(15) / 2453736.5 , 14 / ! 1 jan 2006      
      data r_jdindex(16), i_gps_utc(16) / 2454832.5 , 15 / ! 1 jan 2009  
      data r_jdindex(17), i_gps_utc(17) / 2456109.5 , 16 / ! 1 jul 2012  
      data r_jdindex(18), i_gps_utc(18) / 9999999.0 , -1 /      

      save r_jdindex, i_gps_utc

c     FUNCTION STATEMENTS:

      real*8  secgps2jd
      real*8  jd2secgps
      
c     PROCESSING STEPS:

      r_timeout = -1.0  ! error values
      i_lps = -1
      i_lpf = -1
      
c     Convert input time into julian date

      if (i_opt .eq. 1) then    ! Julian date
         r_jd = r_timein         ! r_time already JD
      else if (i_opt .eq. 2) then ! GPS seconds
         r_jd = secgps2jd(r_timein)
      else
         write(*,*) " *** leap_sec_corr => incorrect i_opt called ***"
         return
      end if
      
c         Check if leap seconds have already been computed and corresponds to i_curr index

      r_wnd = I_MAX / 86400.d0
      
      if ((i_curr.gt.0) .and. (r_jdindex(i_curr)+r_wnd.le.r_jd) .and. 
     &   (r_jd.lt.r_jdindex(i_curr+1)-r_wnd) .and. (i_leap.eq.i_oldleap)) then
         if (i_leap.eq.0) then
            i_lps = -i_gps_utc(i_curr)
            r_jd = r_jd + i_lps/SECDAY;
            i_lpf = 1
         else if (i_leap.eq.1) then
            i_lps = i_gps_utc(i_curr)
            r_jd = r_jd + i_lps/SECDAY;
            i_lpf = 0
         else
            write(*,*) " *** leap_sec_corr => invalid i_leap ***"
            return
         endif
         
         if (i_opt .eq. 2) then
            r_timeout = jd2secgps(r_jd)
         else
            r_timeout = r_jd
         endif
         return
      end if      

      i_curr = 0
      
      i_curr = 0
      j = 1
      
      l_found = .false.
      do while ((.not.l_found) .and. (r_jdindex(j).lt.9999999.0))
         if ((r_jdindex(j).le.r_jd) .and. (r_jd.lt.r_jdindex(j+1))) then
            l_found = .true.
            i_curr = j
            i_lps = i_gps_utc(i_curr)
         endif
         j = j + 1
      enddo
      
      if (.not.l_found) then
         write(*,*) " *** leap_sec_corr => no valid leap second found in table ***"
         return
      endif
      
      if (i_leap .eq. 2) i_lps = i_lps - 1   ! exact leap second, use previous table value
      if (i_leap .gt. 0) then ! time contains leap seconds (UTC), which must be added to GPS time
         r_jd = r_jd + i_lps/SECDAY;
         i_lpf = 0
      else   ! time does not contain leap seconds (GPS), which must be subtracted for UTC time
         i_lps = -i_lps
         r_jd0 = r_jd
         r_jd = r_jd + i_lps/SECDAY;
         i_lpf = 1
         
         if ((r_jdindex(i_curr)+r_wnd.le.r_jd) .and. (r_jd.lt.r_jdindex(i_curr+1)-r_wnd)) then
            continue
         else  ! close to leap second boundary
    ! since epoch indicating when leap seconds are added is given in UTC
    ! time, need to readjust number of leapseconds around leap sec boundary
            j = 1
            l_found = .false.
            do while ((.not.l_found) .and. (r_jdindex(j).lt.9999999.0))
               if ((r_jdindex(j).le.r_jd) .and. (r_jd.lt.r_jdindex(j+1))) then
                  l_found = .true.
                  i_curr = j
                  i_lps = i_gps_utc(i_curr)
               endif
               j = j + 1
            enddo
            i_lps = -i_lps
            r_jd = r_jd0 + i_lps/SECDAY;
    ! need to identify exact leap second for flagging on output
            j = 1
            l_found = .false.
            do while ((.not.l_found) .and. (r_jdindex(j).lt.9999999.0))
               if ((r_jdindex(j).le.r_jd) .and. (r_jd.lt.r_jdindex(j+1))) then
                  l_found = .true.  
                  i_curr = j
                  if (i_lps+i_gps_utc(i_curr).ne.0) i_lpf = 2
               endif
               j = j + 1
            enddo
         endif
      endif
      
      if (i_opt .eq. 2) then
         r_timeout = jd2secgps(r_jd)
      else
         r_timeout = r_jd
      endif
      
      return    

      end  
    

c****************************************************************

      real*8 function SECGPS2JD( SEC )

c****************************************************************
c**     
c**   FILE NAME: 
c**     
c**   DATE WRITTEN: 1/1/91
c**     
c**   PROGRAMMER: Frank Webb
c**     
c**   FUNCTIONAL DESCRIPTION: 
c**
c**    This double precision Function sub routine (SEConds GPS 2 Julian Date) takes 
c**    input time in seconds past the Julian reference date (GPSREF) for this 
c**    and returns the corresponding Julian date as its functional value.    
c**
c**   ROUTINES CALLED: 
c**     
c**   NOTES: 
c**
c**   SEC  is the seconds past the GPS reference date for this library.
c**     
c**   UPDATE LOG:
c**
c**   Date Changed        Reason Changed                  CR # and Version #
c**   ------------       ----------------                 -----------------
c**     
c*****************************************************************

      implicit none

c     PARAMETER STATEMENTS:

      real*8   SECDAY
      PARAMETER        ( SECDAY = 86400.0D0 )
      real*8 GPSREF
      PARAMETER        (GPSREF=2444244.5D0)

c     INPUT VARIABLES:

      real*8 SEC
        
c     OUTPUT VARIABLES:

c     LOCAL VARIABLES:

c     COMMON BLOCKS:

c     EQUIVALENCE STATEMENTS:

c     DATA STATEMENTS:

C     FUNCTION STATEMENTS:

c     PROCESSING STEPS:

c     Divide the input seconds past the Julian reference date by the number
c     of seconds in a day and add the Julian reference date.
      
      SECGPS2JD = SEC/SECDAY + GPSREF
        
      end  

c****************************************************************

      real*8  function JD2SECGPS( JD )

c****************************************************************
c**     
c**   FILE NAME: 
c**     
c**   DATE WRITTEN: 1/1/91
c**     
c**   PROGRAMMER: Frank Webb
c**     
c**   FUNCTIONAL DESCRIPTION: 
c**   
c**   This sub routine (Julian Date 2 SECondsGPS) takes an input Julian date an 
c**   returns the corresponding time in seconds past the reference date (GPS JD 
c**   for this library as its functional value.                             
c**     
c**   ROUTINES CALLED: GPSREF
c**     
c**   NOTES: 
c**     
c**   UPDATE LOG:
c**
c**   Date Changed        Reason Changed                  CR # and Version #
c**   ------------       ----------------                 -----------------
c**     
c*****************************************************************

      implicit none

c     PARAMETER STATEMENTS:

      real*8   SECDAY
      PARAMETER(SECDAY=86400.0D0)
      real*8 GPSREF
      PARAMETER        (GPSREF=2444244.5D0)

c     INPUT VARIABLES:

      real*8   JD
        
c     OUTPUT VARIABLES:

c     LOCAL VARIABLES:

c     COMMON BLOCKS:

c     EQUIVALENCE STATEMENTS:

c     DATA STATEMENTS:

C     FUNCTION STATEMENTS:

c     PROCESSING STEPS:

c    Subtract the Julian reference date from the input Julian date and
c    multiply by the number of seconds in a day (86400).

      JD2SECGPS = (JD - GPSREF)*SECDAY
        
      end  

c****************************************************************

      subroutine YMDHMS2gps(i_yyyy,i_mm,i_dd,i_hh,i_mn,r_ss,r_gpssec,i_leap)

c****************************************************************
c**
c**   FILE NAME: 
c**     
c**   DATE WRITTEN: 7/24/2009
c**     
c**   PROGRAMMER: Adam Freedman
c**     
c**   FUNCTIONAL DESCRIPTION: 
c**      This subroutine takes a UTC time in YYYY, MM, DD, HH, MN, SS
c**      format and converts it into GPS time in seconds, accounting for 
c**      leap seconds in the process.If the time is an exact leap second
c**      the i_leap flag must be set.
c**  
c**   ROUTINES CALLED: 
c**     
c**   NOTES: 
c**     
c**   UPDATE LOG:
c**
c**   Date Changed        Reason Changed                  CR # and Version #
c**   ------------       ----------------                 -----------------
c**     
c*****************************************************************

      implicit none

c     PARAMETER STATEMENTS:

c     INPUT VARIABLES:

      integer i_yyyy, i_mm, i_dd
      integer i_hh, i_mn 
      real*8 r_ss
      integer i_leap
      
c     OUTPUT VARIABLES:
      
      real*8 r_gpssec   ! seconds of time since GPS epoch of 1/6/1980
      
c     LOCAL VARIABLES:

      real*8 r_juldate
      real*8 r_juldatelps
      real*8 r_sec
      
      integer i_lps     ! leap seconds at this epoch
      integer i_lpf     ! output leap second flag
      integer i_ss
      real*8  r_frac
      integer i_lpin
      
c     COMMON BLOCKS:
      
c     EQUIVALENCE STATEMENTS:
      
c     DATA STATEMENTS:
      
c     FUNCTION STATEMENTS:

      real*8  jd2secgps
      integer DATE2J

c     PROCESSING STEPS:

c     Convert calendar date to Julian date
      
      r_juldatelps = DATE2J(i_yyyy,i_mm,i_dd) - 0.5d0

      r_sec = r_ss
      
      if (i_leap.ne.0) then   ! leap second
         if (int(r_sec).eq.59) then   ! leap second would be 23:59:59
            r_sec = r_sec + 1  ! turns into 23:59:60
            i_lpin = 2
         elseif (int(r_sec).eq.0) then ! 00:00:00 work as is
            i_lpin = 2
         elseif (int(r_sec).eq.60) then ! 23:59:60 work as is
            i_lpin = 2
         else   
            i_lpin = 1  ! turn back to non-leap second; behavior untested
                        ! Note: not checking for hour or minute either
         endif
      else
         i_lpin = 1
      endif
      
      r_juldatelps = r_juldatelps + i_hh/24.d0 + i_mn/1440.d0 + r_sec/86400.d0

c     Adjust julian date for leap seconds
      
      call leap_sec_corr(r_juldatelps,1,i_lpin,r_juldate,i_lps,i_lpf)
      
c     Convert julian date into GPS seconds

      r_gpssec = jd2secgps(r_juldate)      
      
      return
      
      end
      
c****************************************************************

      subroutine YdoyHMS2gps(i_yyyy,i_doy,i_hh,i_mn,r_ss,r_gpssec,i_leap)    
 
c****************************************************************
c**
c**   FILE NAME: 
c**     
c**   DATE WRITTEN: 7/24/2009
c**     
c**   PROGRAMMER: Adam Freedman
c**     
c**   FUNCTIONAL DESCRIPTION: 
c**      This subroutine takes a UTC time in YYYY, DOY, HH, MN, SS
c**      format and converts it into GPS time in seconds, accounting for 
c**      leap seconds in the process.
c**  
c**   ROUTINES CALLED: 
c**     
c**   NOTES: 
c**     
c**   UPDATE LOG:
c**
c**   Date Changed        Reason Changed                  CR # and Version #
c**   ------------       ----------------                 -----------------
c**     
c*****************************************************************

      implicit none

c     PARAMETER STATEMENTS:

c     INPUT VARIABLES:

      integer i_yyyy, i_doy
      integer i_hh, i_mn 
      real*8 r_ss
      integer i_leap
      
c     OUTPUT VARIABLES:
      
      real*8 r_gpssec   ! seconds of time since GPS epoch of 1/6/1980
      
c     LOCAL VARIABLES:

      integer i_jdYear
      integer i_juldate
      integer i_YEAR, i_MONTH, i_DAY
      integer i_mm, i_dd
      
c     COMMON BLOCKS:
      
c     EQUIVALENCE STATEMENTS:
      
c     DATA STATEMENTS:
      
c     FUNCTION STATEMENTS:

      integer DATE2J

c     PROCESSING STEPS: 

c     Convert DOY to MM, DD
 
      i_jdYear = DATE2J(i_yyyy, 1, 1)
      
      i_juldate = i_jdYear + i_doy - 1
      
      call J2DATE( i_juldate, i_YEAR, i_MONTH, i_DAY )

c     Compute GPS second

      i_mm = i_MONTH
      i_dd = i_DAY
 
      call YMDHMS2gps(i_yyyy,i_mm,i_dd,i_hh,i_mn,r_ss,r_gpssec,i_leap)
      
      return
      end
      
c****************************************************************

      subroutine YMDsod2gps(i_yyyy,i_mm,i_dd,r_sod,r_gpssec,i_leap)

c****************************************************************
c**
c**   FILE NAME: 
c**     
c**   DATE WRITTEN: 7/24/2009
c**     
c**   PROGRAMMER: Adam Freedman
c**     
c**   FUNCTIONAL DESCRIPTION: 
c**      This subroutine takes a UTC time in YYYY, MM, DD, seconds of day
c**      format and converts it into GPS time in seconds, accounting for 
c**      leap seconds in the process.
c**  
c**   ROUTINES CALLED: 
c**     
c**   NOTES: 
c**     
c**   UPDATE LOG:
c**
c**   Date Changed        Reason Changed                  CR # and Version #
c**   ------------       ----------------                 -----------------
c**     
c*****************************************************************

      implicit none

c     PARAMETER STATEMENTS:

c     INPUT VARIABLES:

      integer i_yyyy, i_mm, i_dd
      real*8 r_sod
      integer i_leap
      
c     OUTPUT VARIABLES:
      
      real*8 r_gpssec   ! seconds of time since GPS epoch of 1/6/1980
      
c     LOCAL VARIABLES:

      real*8 r_SEC
      integer i_HOUR, i_MINUTE
      integer i_hh, i_mn
      real*8 r_ss
      
c     COMMON BLOCKS:
      
c     EQUIVALENCE STATEMENTS:
      
c     DATA STATEMENTS:
      
c     FUNCTION STATEMENTS:

c     PROCESSING STEPS: 
 
c     Convert SOD to HH, MIN, SS

      r_SEC = r_sod

c     Check for leap second

          if (r_SEC.ge.86400) then  
             i_leap = 1
             r_SEC = r_SEC - 1
          endif
       
c     Compute HOUR.

      i_HOUR   = int(r_SEC/3600.d0)
      r_SEC  = r_SEC - 3600.d0 * i_HOUR

c     Compute MINUTE.

      i_MINUTE = int(r_SEC/60.d0)
      r_SEC   = r_SEC - 60.d0 * i_MINUTE
            
c     Compute GPS second

      i_hh = i_HOUR
      i_mn = i_MINUTE
      r_ss = r_SEC
      
      call YMDHMS2gps(i_yyyy,i_mm,i_dd,i_hh,i_mn,r_ss,r_gpssec,i_leap)
      
      return
      end
      
c****************************************************************

      subroutine Ydoysod2gps(i_yyyy,i_doy,r_sod,r_gpssec,i_leap)
      
c****************************************************************
c**
c**   FILE NAME: 
c**     
c**   DATE WRITTEN: 7/24/2009
c**     
c**   PROGRAMMER: Adam Freedman
c**     
c**   FUNCTIONAL DESCRIPTION: 
c**      This subroutine takes a UTC time in YYYY, DOY, seconds of day
c**      format and converts it into GPS time in seconds, accounting for 
c**      leap seconds in the process.
c**  
c**   ROUTINES CALLED: 
c**     
c**   NOTES: 
c**     
c**   UPDATE LOG:
c**
c**   Date Changed        Reason Changed                  CR # and Version #
c**   ------------       ----------------                 -----------------
c**     
c*****************************************************************

      implicit none

c     PARAMETER STATEMENTS:

c     INPUT VARIABLES:

      integer i_yyyy, i_doy
      real*8 r_sod
      integer i_leap
      
c     OUTPUT VARIABLES:
      
      real*8 r_gpssec   ! seconds of time since GPS epoch of 1/6/1980
      
c     LOCAL VARIABLES:

      integer i_jdYear
      integer i_juldate
      integer i_YEAR, i_MONTH, i_DAY
      integer i_mm, i_dd
      
      real*8 r_SEC
      integer i_HOUR, i_MINUTE
      integer i_hh, i_mn
      real*8 r_ss
      
c     COMMON BLOCKS:
      
c     EQUIVALENCE STATEMENTS:
      
c     DATA STATEMENTS:
      
c     FUNCTION STATEMENTS:

      integer DATE2J

c     PROCESSING STEPS: 

c     Convert DOY to MM, DD
 
      i_jdYear = DATE2J(i_yyyy, 1, 1)
      
      i_juldate = i_jdYear + i_doy - 1
      
      call J2DATE( i_juldate, i_YEAR, i_MONTH, i_DAY )
      
      i_mm = i_MONTH
      i_dd = i_DAY
      
c     Convert SOD to HH, MIN, SS

      r_SEC = r_sod

c     Check for leap second

          if (r_SEC.ge.86400) then  
             i_leap = 1
             r_SEC = r_SEC - 1
          endif
          
c     Compute HOUR.

      i_HOUR   = int(r_SEC/3600.d0)
      r_SEC  = r_SEC - 3600.d0 * i_HOUR

c     Compute MINUTE.

      i_MINUTE = int(r_SEC/60.d0)
      r_SEC   = r_SEC - 60.d0 * i_MINUTE
      
c     Compute GPS second

      i_hh = i_HOUR
      i_mn = i_MINUTE
      r_ss = r_SEC
 
      call YMDHMS2gps(i_yyyy,i_mm,i_dd,i_hh,i_mn,r_ss,r_gpssec,i_leap)
      
      return
      end
 
c*****************************************************************


c****************************************************************

      subroutine JD2CAL(JD,YEAR,MONTH,DAY,HOUR,MINUTE,SECOND,FRAC)

c****************************************************************
c**     
c**   FILE NAME: time_sub.f
c**     
c**   DATE WRITTEN: 1/1/91
c**     
c**   PROGRAMMER: Frank Webb
c**     
c**   FUNCTIONAL DESCRIPTION:
c**
c**   This sub routine (Julian Date 2 CALendar date) takes an input Julian date  
c**   and returns the various components of the corresponding calendar date.     
c**   The components of the calendar date are all returned as numbers to allow   
c**   for use in computation. For instance, the month is returned as the         
c**   integer month number rather than as a character string.                    
c**     
c**   ROUTINES CALLED: J2DATE
c**     
c**   NOTES: 
c**
c**   JD   is the Julian date.           
c**                                      
c**   YEAR     is the year.              
c**   MONTH    is the month number.      
c**   DAY      is the day.               
c**   HOUR     is the hour.              
c**   MINUTE   is the minute.            
c**   SECOND   is the second.            
c**   FRAC     is the fractional seconds.
c*  
c**   UPDATE LOG:
c**
c**   Date Changed        Reason Changed                  CR # and Version #
c**   ------------       ----------------                 -----------------
c**     
c*****************************************************************

      implicit none

c     PARAMETER STATEMENTS:

      real*8             SECDAY
      PARAMETER        ( SECDAY = 86400.0D0 )

c     INPUT VARIABLES:

      real*8             JD
        
c     OUTPUT VARIABLES:

      INTEGER            YEAR
      INTEGER            MONTH
      INTEGER            DAY
      INTEGER            HOUR
      INTEGER            MINUTE
      INTEGER            SECOND
      real*8             FRAC

c     LOCAL VARIABLES:

      real*8             JDPLUS
      INTEGER            JDINT
      real*8             DSEC
      INTEGER            ISEC

c     COMMON BLOCKS:

c     EQUIVALENCE STATEMENTS:

c     DATA STATEMENTS:

C     FUNCTION STATEMENTS:

c     PROCESSING STEPS:

c    Compute JDINT, the integer Julian date at noon of the current day
c    compute DSEC, the number of seconds elapsed since the start of the
c    current day.

      JDPLUS = JD + 0.5D0
      JDINT  = INT( JDPLUS )
      DSEC   = SECDAY * MOD( JDPLUS, 1.0D0 )
      IF( DSEC .GE. SECDAY )THEN
         JDINT = JDINT + 1
         DSEC  = DSEC  - SECDAY
      END IF
      ISEC = INT( DSEC )

c     Call J2DATE with JDINT to compute the year, month, and day of the
c     calendar date.

      CALL J2DATE( JDINT, YEAR, MONTH, DAY )

c     Compute HOUR.

      HOUR   = ISEC/3600
      ISEC   = ISEC - 3600*HOUR

c     Compute MINUTE.

      MINUTE = ISEC/60
      ISEC   = ISEC - 60*MINUTE

c     Compute SECOND.
      SECOND = ISEC

c     Compute FRAC.
      FRAC = dMOD( DSEC, 1.0D0 )
        if(dabs(1.d0-FRAC)<.001d0) then 
           FRAC=.0d0 
           SECOND = SECOND + 1
           if(second.eq.60) then
              minute = minute + 1
              second = 0
           end if
           if(minute.eq.60) then
              hour = hour + 1
              minute = 0
           end if
        end if
        if(FRAC<.0001d0) then 
           frac = .0d0
        end if

        return
      end  


c****************************************************************

      subroutine J2DATE(JD,YEAR,MONTH,DAY)

c****************************************************************
c**     
c**   FILE NAME: time_sub.f
c**     
c**   DATE WRITTEN: 1/1/91
c**     
c**   PROGRAMMER: Frank Webb
c**     
c**   FUNCTIONAL DESCRIPTION: 
c**
c**   This sub routine (Julian date 2 calendar DATE) converts the input integer    
c**   Julian date to the corresponding Gregorian calendar date. Since the Julian  
c**   date is an integer, this correspondence is exact for noon of the caledar    
c**   date.                                                                       
c**                                                                               
c**   The algorithm for this conversion is taken from the following article       
c**   Tantzen,R.T., "Communications of the ACM", Volume 6, Number 8, August       
c**   Algorithm 199, page 444.                                                    
c**  
c**   ROUTINES CALLED: None
c**     
c**   NOTES: 
c**
c**   JD       is the integer Julian date.  
c**                                         
c**   YEAR     is the year number.          
c**   MONTH    is the month number.         
c**   DAY      is the day number.           
c**  
c**   UPDATE LOG:
c**
c**   Date Changed        Reason Changed                  CR # and Version #
c**   ------------       ----------------                 -----------------
c**     
c*****************************************************************

      implicit none

c     INPUT VARIABLES:

      INTEGER            JD
        
c     OUTPUT VARIABLES:

      INTEGER            YEAR
      INTEGER            MONTH
      INTEGER            DAY

c     LOCAL VARIABLES:

      INTEGER            J
      INTEGER            Y
      INTEGER            M
      INTEGER            D

c     COMMON BLOCKS:

c     EQUIVALENCE STATEMENTS:

c     DATA STATEMENTS:

C     FUNCTION STATEMENTS:

c     PROCESSING STEPS:

      J = JD

      J = J - 1721119
      Y = (4*J-1)/146097
      J = 4*J - 1 - 146097*Y
      D = J/4
      J = (4*D+3)/1461
      D = 4*D + 3 -1461*J
      D = (D+4)/4
      M = (5*D-3)/153
      D = 5*D - 3 - 153*M
      D = (D+5)/5
      Y = 100*Y + J
      IF( M .LT. 10 )THEN
         M = M + 3
      ELSE
         M = M - 9
         Y = Y + 1
      END IF

      YEAR   = Y
      MONTH  = M
      DAY    = D
        return  
      end  
      
c****************************************************************

      integer function DATE2J( YEAR, MONTH, DAY )

c****************************************************************
c**     
c**   FILE NAME: 
c**     
c**   DATE WRITTEN: 1/1/91
c**     
c**   PROGRAMMER: Frank Webb
c**     
c**   FUNCTIONAL DESCRIPTION: 
c**
c**   This integer Function (calendar DATE 2 Julian date) takes the input   
c**   Gregorian calendar date and returns as its functional value the       
c**   corresponding integer Julian date. Since the Julian date is an integer 
c**   this correspondence is exact for noon of the output calendar date.    
c**                                                                         
c**   The algorithm for this conversion is taken from the following article 
c**   Tantzen,R.T., "Communications of the ACM", Volume 6, Number 8, August 
c**   Algorithm 199, page 444.                                              
c**     
c**   ROUTINES CALLED: None
c**     
c**   NOTES: 
c**  
c**    YEAR     is the year number.   
c**    MONTH    is the month number.  
c**    DAY      is the day number.    
c**     
c**   UPDATE LOG:
c**
c**   Date Changed        Reason Changed                  CR # and Version #
c**   ------------       ----------------                 -----------------
c**     
c*****************************************************************

      implicit none

c     INPUT VARIABLES:

      INTEGER            YEAR
      INTEGER            MONTH
      INTEGER            DAY
        
c     OUTPUT VARIABLES:

c     LOCAL VARIABLES:

      INTEGER            Y
      INTEGER            M
      INTEGER            D
      INTEGER            C
      INTEGER            YA

c     COMMON BLOCKS:

c     EQUIVALENCE STATEMENTS:

c     DATA STATEMENTS:

C     FUNCTION STATEMENTS:

c     PROCESSING STEPS:

      Y = YEAR
      M = MONTH
      D = DAY

      IF( M .GT. 2 )THEN
         M = M - 3
      ELSE
         M = M + 9
         Y = Y - 1
      END IF

      C  = Y/100
      YA = Y - 100*C
      DATE2J = (146097*C)/4 + (1461*YA)/4 + (153*M+2)/5 + D + 1721119

      end        


        subroutine find_julian_day(iyear,imonth, iday, julian_day)
        implicit none
        integer iyear, imonth, iday
        real*8 julian_day
        real*8 mm, dd, year
c
c       This subroutine converts any Gregorian calendar date after
c       October 15, 1582 to the corresponding Julian Day number.
c
c       julian_day = julian date at 0h GMT
c       qyear = year (19xx, e.g.), integer
c       qmm   = month, integer
c       qdd   = day, integer
c       -----------------------------------------------------------

          integer*4 idy,imm,iadum,ily
          real*8 m, y, adum , jdn

          mm   = imonth ! convert to real*8 for accuracy
          dd   = iday
          year = iyear

          if (mm .le. 2.d0) then      ! adjust year if necessary
             year = year - 1.0d0
             mm   = mm + 12.0d0
          endif

          m     = mm
          y     = year
          iadum = y/100.d0
          ily   = iadum/4
          adum  = 2.0d0 - iadum + ily
          idy   = 365.25d0*y
          imm   = 30.6001d0*(m + 1.d0)
          jdn   = idy + imm + dd
          julian_day = jdn + 1720994.5d0 + adum      ! real*8 JD


        return
        end

