      subroutine read_sacd_gps(xmaglm,gpsvec,nsig,
     *  igyr,igday,gpsec,secst,secend,ngps)
c
c  Purpose:  This subroutine extracts the valid GPS data from the input
c  		telemetry structure.  Only data points with four or more
c		GPS signals are included.  Also, the time tags for all
c 		input minor frames are checked to determine the time 
c 		range of the data.
c
c  Calling Arguments:
c
c  Name         Type    I/O     Description
c  --------     ----    ---     -----------
c  xmaglm(2)	R*8	 I	Limits on position magnitude
c  gpsvec(6,*)	R*8	 O	Output GPS orbit vectors
c  nsig(*)	I*4	 O	Number of GPS signals for each vector
c  igyr		I*4	 O	Time tag year
c  igday	I*4	 O	Time tag day-of-year
c  gpsec(*)	R*8	 O	Time tag seconds of igday
c  secst       	R*8	 O	Start of interval time in seconds
c  secend	R*8	 O	End of interval time in seconds
c  ngps		I*4	 O	Number of GPS vectors
c
c  By: Frederick S. Patt, GSC, December 21, 1993
c
c  Notes:  
c
c  Modification History:
c
c  Modified to unpack GPS time tag from telemetry (previously used
c  minor frame time tag).  F. S. Patt, GSC, September 14, 1996
c
c  Modified to perform sanity check on time tags (e.g., all tags 
c  fall within +/- 1 day of first valid time tag.
c  F. S. Patt, GSC, October 4, 1996.
c
c  Fixed a bug in validating the first GPS time code.  
c  F. S. Patt, SAIC GSC, September 7, 1997.
c
c  Added additional range checking to the seconds-of-day.
c  F. S. Patt, SAIC GSC, September 10, 1997
c
c  Changed integer*1 to byte for Sun OS compatibility, B. A. Franz,
c  GSC, November 14, 1997.
c
c  Modified to reject GPS samples with time tags later than minor frame.
c  F. S. Patt, SAIC GSC, July 22, 1998.
c
c  Modified to reject GPS samples prior to current time range.
c  F. S. Patt, SAIC GSC, March 3, 2000
c
c  Modified to remove checks on minor frame times (SWl01 takes care of this).
c  F. S. Patt, SAIC GSC, May 20, 2000
c
c  Fixed bug which caused error message to be output for duplicate GPS sample.
c  F. S. Patt, SAIC GSC, June 8, 2000 
c
c  Fixed bug introduced in indexing of input array during generation of 
c  single-source code for SGI and Linux. 
c  F. S. Patt, SAIC GSC, January 19, 2001 
c
c  Created read_sacd_gps from SeaWiFS read_gps routine
c  F. S. Patt, SAIC, July 19, 2011
c
c  Modified to reject samples with 15 satellites tracked (fill value)
c  F. S. Patt, SAIC, July 25, 2011

      implicit none
c#include "nav_cnst.fin"
C#include "input_s.fin"

C      record /input_struct/input(maxlin)

      integer*4 maxlin
      parameter (maxlin = 16000)
      real*8 gpsvec(6,maxlin), gpsec(maxlin), scal_p, secst, secend
      real*8 sec, xmaglm(2), pmag, gsec
      real*4 gvec(6)
      integer*4 nsig(maxlin), nframes, igyr, igday, ngps, csec
      integer*4 iframe, mframe, if1, j, iy2, im2, id2, ms, mp 
      integer*4 jd, jd1, jd2, jdtol
      integer*2 ib2
      byte      b2(2)
      logical gotone,gotime,hdata
      character*80 hfile
      data jdtol/1/
      data hdata/.true./
      equivalence (ib2,b2)

c  Get name and open input file
c      print *,'Enter SAC-D HKT file name'
c      read(5,'(a80)') hfile
      call getarg(1, hfile)
      open (25,file=hfile,status='old',err=990,
     *    access='direct',recl=4000)
c     *     form='binary')
      

c  Find first good GPS data
      call read_gps_from_hkt(25,igyr,igday,gsec,gvec,ms,csec,hdata)
      print *, 'CDH Start Time', csec
      if1 = 1
      ngps = 0
	  gotime = .true.
c          igyr = input(if1).iyear
c	  igday = input(if1).iday
	  jd1 = jd(igyr,1,igday)
	  secst = csec
          secend = secst

c  Find first good vector
      gotone = .false. 
c      dowhile (.not.gotone)
c	mframe = input(if1).sc_id(1)/128

c  Check minor frame time tag against time range
c	  if (sec.lt.secst) secst = sec
c	  if (gsec.gt.secend) secend = gsec
c	if((input(if1).flag.eq.0).and.(mframe.eq.1)) then

c  Unpack number of GPS signals from telemetry data and check for minimum
c	  ms = input(if1).sc_dis(12)
	  if ((ms.ge.4).and.(ms.lt.15)) then
           
            if ( (gsec.ge.secst) .and. (gsec.le.secend) ) then

              ngps = ngps + 1
              gotone = .true.
c	      if (sec.lt.secst) secst = sec

c  Store data in output array
	      gpsec(ngps) = gsec
	      nsig(ngps) = ms
	      do j=1,6
	        gpsvec(j,ngps) = gvec(j)
	      end do
c Perform limit check on position magnitude
	      pmag = sqrt(gpsvec(1,ngps)**2 + gpsvec(2,ngps)**2
     *		+ gpsvec(3,ngps)**2)
	      if ((pmag.lt.xmaglm(1)).or.(pmag.gt.xmaglm(2))) then
	        gotone = .false.
	        ngps = ngps - 1  
	      end if
	    end if
	  end if
c	end if
c	if1 = if1 + 1
c	if (if1.gt.maxlin) go to 999
c      end do
      mp = ms

c  Now unpack remaining GPS vectors; check for >4 signals at previous point.
c      do iframe = if1,nframes
      dowhile (hdata)
        call read_gps_from_hkt(25,igyr,igday,gsec,gvec,ms,csec,hdata)
        iy2 = igyr
	id2 = igday
	jd2 = jd(iy2,1,id2) 
        csec = csec + (jd2-jd1)*864.d2
        gsec = gsec + (jd2-jd1)*864.d2

c  Check time tag against range
        if (csec.lt.secst) secst = csec
        if (csec.gt.secend) secend = csec

c  Check for time tag out of range
          if ((gsec.ge.secst).and.(gsec.le.secend)) then

c  Check for < minimum GPS signals or duplicated time tag
            if ((ms.ge.4).and.(ms.lt.15).and.(gsec.ne.gpsec(ngps))) then
                ngps = ngps + 1
             
c     Store data in output array
                gpsec(ngps) = gsec
                nsig(ngps) = ms
	      do j=1,6
	        gpsvec(j,ngps) = gvec(j)
	      end do
c     Perform limit check on position magnitude
                pmag = sqrt(gpsvec(1,ngps)**2 + gpsvec(2,ngps)**2
     *               + gpsvec(3,ngps)**2)
                if ((pmag.lt.xmaglm(1)).or.(pmag.gt.xmaglm(2))) then
                   ngps = ngps - 1  
                end if
             
c  Save number of GPS signals
                mp = ms
             end if
          else
c             print *,gsec,csec,secst,secend
             print *,'READ_GPS:  Invalid date',iy2,id2
c             stop
	  end if
c        end if
      end do
      
  999 print *,'READ_GPS:',ngps,'  valid GPS vectors'
      print *,' starting at ',igyr,igday,secst

      return

  990 print *,'Error opening file ',hfile
      stop

      end
